@@ -1,34 +1,22 @@
 /*
-
-	cond.c - evaluate conditional expressions
-
-	This file is part of zsh, the Z shell.
-
-	zsh is free software; no one can prevent you from reading the source
-   code, or giving it to someone else.
-
-   This file is copyrighted under the GNU General Public License, which
-   can be found in the file called COPYING.
-
-   Copyright (C) 1990, 1991 Paul Falstad
-
-   zsh is distributed in the hope that it will be useful, but
-   WITHOUT ANY WARRANTY.  No author or distributor accepts
-   responsibility to anyone for the consequences of using it or for
-   whether it serves any particular purpose or works at all, unless he
-   says so in writing.  Refer to the GNU General Public License
-   for full details.
-
-   Everyone is granted permission to copy, modify and redistribute
-   zsh, but only under the conditions described in the GNU General Public
-   License.   A copy of this license is supposed to have been given to you
-   along with zsh so you can know your rights and responsibilities.
-   It should be in a file named COPYING.
-
-   Among other things, the copyright notice and this notice must be
-   preserved on all copies.
-
-*/
+ *
+ * cond.c - evaluate conditional expressions
+ *
+ * This file is part of zsh, the Z shell.
+ *
+ * This software is Copyright 1992 by Paul Falstad
+ *
+ * Permission is hereby granted to copy, reproduce, redistribute or otherwise
+ * use this software as long as: there is no monetary profit gained
+ * specifically from the use or reproduction of this software, it is not
+ * sold, rented, traded or otherwise marketed, and this copyright notice is
+ * included prominently in any copy made. 
+ *
+ * The author make no claims as to the fitness or correctness of this software
+ * for any use whatsoever, and it is provided as is. Any use of this software
+ * is at the user's own risk. 
+ *
+ */
 
 #include "zsh.h"
 
@@ -57,7 +45,7 @@ struct stat *st;
 		case COND_STRNEQ: return !matchpat(c->left,c->right);
 		case COND_STRLT: return strcmp(c->left,c->right) < 0;
 		case COND_STRGTR: return strcmp(c->left,c->right) > 0;
-		case 'a': return(doaccess(c->left,F_OK));
+		case 'e': case 'a': return(doaccess(c->left,F_OK));
 		case 'b': return(S_ISBLK(dostat(c->left)));
 		case 'c': return(S_ISCHR(dostat(c->left)));
 		case 'd': return(S_ISDIR(dostat(c->left)));
@@ -74,7 +62,7 @@ struct stat *st;
 		case 'w': return(doaccess(c->left,W_OK));
 		case 'x': return(doaccess(c->left,X_OK));
 		case 'z': return(!strlen(c->left));
-		case 'L': return(S_ISLNK(dolstat(c->left)));
+		case 'h': case 'L': return(S_ISLNK(dolstat(c->left)));
 		case 'O': return((st = getstat(c->left)) && st->st_uid == geteuid());
 		case 'G': return((st = getstat(c->left)) && st->st_gid == getegid());
 		case 't': return isatty(matheval(c->left));
