@@ -1,34 +1,22 @@
 /*
-
-	exec.c - command execution
-
-	This file is part of zsh, the Z shell.
-
-	zsh is free software; no one can prevent you from reading the source
-   code, or giving it to someone else.
-
-   This file is copyrighted under the GNU General Public License, which
-   can be found in the file called COPYING.
-
-   Copyright (C) 1990, 1991 Paul Falstad
-
-   zsh is distributed in the hope that it will be useful, but
-   WITHOUT ANY WARRANTY.  No author or distributor accepts
-   responsibility to anyone for the consequences of using it or for
-   whether it serves any particular purpose or works at all, unless he
-   says so in writing.  Refer to the GNU General Public License
-   for full details.
-
-   Everyone is granted permission to copy, modify and redistribute
-   zsh, but only under the conditions described in the GNU General Public
-   License.   A copy of this license is supposed to have been given to you
-   along with zsh so you can know your rights and responsibilities.
-   It should be in a file named COPYING.
-
-   Among other things, the copyright notice and this notice must be
-   preserved on all copies.
-
-*/
+ *
+ * exec.c - command execution
+ *
+ * This file is part of zsh, the Z shell.
+ *
+ * This software is Copyright 1992 by Paul Falstad
+ *
+ * Permission is hereby granted to copy, reproduce, redistribute or otherwise
+ * use this software as long as: there is no monetary profit gained
+ * specifically from the use or reproduction of this software, it is not
+ * sold, rented, traded or otherwise marketed, and this copyright notice is
+ * included prominently in any copy made. 
+ *
+ * The author make no claims as to the fitness or correctness of this software
+ * for any use whatsoever, and it is provided as is. Any use of this software
+ * is at the user's own risk. 
+ *
+ */
 
 #include "zsh.h"
 #include <sys/errno.h>
@@ -41,6 +29,9 @@
 #define execerr() { if (forked) exit(1); \
 	closemnodes(mfds); errflag = 1; return; }
 
+static Lklist args;
+static Cmdnam cn;
+
 /* parse list in a string */
 
 List parselstring(s) /**/
@@ -51,13 +42,12 @@ List l;
 	hungets(s);
 	strinbeg();
 	pushheap();
-	if (!(l = parse_list()))
-		{
+	if (!(l = parse_list())) {
 		strinend();
 		hflush();
 		popheap();
 		return NULL;
-		}
+	}
 	strinend();
 	return l;
 }
@@ -69,32 +59,27 @@ char *s;
 {
 List l;
 
-	if (l = parselstring(s))
-		{
+	if (l = parselstring(s)) {
 		execlist(l);
 		popheap();
-		}
-}
-
-/* duplicate a list and run it */
-
-void newrunlist(l) /**/
-List l;
-{
-	List a = (List) dupstruct(l); runlist(a);
+	}
 }
 
 /* fork and set limits */
 
 int phork() /**/
 {
-int pid = fork(),t0;
+int pid,t0;
 
-	if (pid == -1)
-		{
+	if (thisjob >= MAXJOB-1) {
+		zerr("job table full",NULL,0);
+		return -1;
+	}
+	pid = fork();
+	if (pid == -1) {
 		zerr("fork failed: %e",NULL,errno);
 		return -1;
-		}
+	}
 #ifdef RLIM_INFINITY
 	if (!pid)
 		for (t0 = 0; t0 != RLIM_NLIMITS; t0++)
@@ -117,106 +102,81 @@ Cmd cmd;
 
 #define POUNDBANGLIMIT 64
 
-void zexecve(pth,argv,ee,b1,b2) /**/
-char *pth;char **argv;int *ee;char *b1;char *b2;
+int zexecve(pth,argv) /**/
+char *pth;char **argv;
 {
 int eno;
-char buf[MAXPATHLEN*2];
+static char buf[MAXPATHLEN*2];
 char **eep;
 
 	for (eep = environ; *eep; eep++)
-		if (**eep == '_' && (*eep)[1] == '=')
-			break;
+		if (**eep == '_' && (*eep)[1] == '=') break;
 	buf[0] = '_';
 	buf[1] = '=';
-	if (*pth == '/')
-		strcpy(buf+2,pth);
-	else
-		sprintf(buf+2,"%s/%s",cwd,pth);
-	if (!*eep)
-		eep[1] = NULL;
+	if (*pth == '/') strcpy(buf+2,pth);
+	else sprintf(buf+2,"%s/%s",pwd,pth);
+	if (!*eep) eep[1] = NULL;
 	*eep = buf;
 	execve(pth,argv,environ);
-	if ((eno = errno) == ENOEXEC)
-		{
+	if ((eno = errno) == ENOEXEC) {
 		char buf[POUNDBANGLIMIT+1],*ptr,*ptr2,*argv0;
 		int fd,ct,t0;
 
-		if ((fd = open(pth,O_RDONLY)) >= 0)
-			{
+		if ((fd = open(pth,O_RDONLY)) >= 0) {
 			argv0 = *argv;
 			*argv = pth;
 			ct = read(fd,buf,POUNDBANGLIMIT);
 			close(fd);
-			if (ct > 0)
-				{
+			if (ct > 0) {
 				if (buf[0] == '#')
-					if (buf[1] == '!')
-						{
+					if (buf[1] == '!') {
 						for (t0 = 0; t0 != ct; t0++)
 							if (buf[t0] == '\n')
 								buf[t0] = '\0';
 						buf[POUNDBANGLIMIT] = '\0';
 						for (ptr = buf+2; *ptr && *ptr == ' '; ptr++);
 						for (ptr2 = ptr; *ptr && *ptr != ' '; ptr++);
-						if (*ptr)
-							{
+						if (*ptr) {
 							*ptr = '\0';
 							argv[-2] = ptr2;
 							argv[-1] = ptr+1;
 							execve(ptr2,argv-2,environ);
-							}
-						else
-							{
+						} else {
 							argv[-1] = ptr2;
 							execve(ptr2,argv-1,environ);
-							}
 						}
-					else
-						{
+					} else {
 						argv[-1] = "sh";
 						execve("/bin/sh",argv-1,environ);
-						}
-				else
-					{
+					}
+				else {
 					for (t0 = 0; t0 != ct; t0++)
-						if (!buf[t0])
-							break;
-					if (t0 == ct)
-						{
+						if (!buf[t0]) break;
+					if (t0 == ct) {
 						argv[-1] = "sh";
 						execve("/bin/sh",argv-1,environ);
-						}
 					}
 				}
-			else
-				eno = errno;
+			} else eno = errno;
 			*argv = argv0;
-			}
-		else
-			eno = errno;
-		}
-	if (ee && eno != ENOENT)
-		{
-		*ee = eno;
-		strcpy(b1,b2);
-		}
+		} else eno = errno;
+	}
+	return eno;
 }
 
 #define MAXCMDLEN (MAXPATHLEN*4)
 
 /* execute an external command */
 
-void execute(args,dash) /**/
-Lklist args;int dash;
+void execute(dash) /**/
+int dash;
 {
-char **argv,*arg0;
+char **argv,*arg0,**pp;
 char *z,*s,buf[MAXCMDLEN],buf2[MAXCMDLEN];
-Cmdnam cn;
-int tl,ee = 0;
+int ee = 0,eno = 0;
 
 	if (!full(args)) {
-		zerr("no command");
+		zerr("no command",NULL,0);
 		_exit(1);
 	}
 	cn = (Cmdnam) gethnode(peekfirst(args),cmdnamtab);
@@ -225,119 +185,179 @@ int tl,ee = 0;
 	if (s = zgetenv("STTY"))
 		zyztem("stty",s);
 	arg0 = peekfirst(args);
-	if (z = zgetenv("ARGV0"))
-		{
+	if (z = zgetenv("ARGV0")) {
 		setdata(firstnode(args),ztrdup(z));
 		delenv(z-6);
-		}
-	else if (dash)
-		{
+	} else if (dash) {
 		sprintf(buf2,"-%s",arg0);
 		setdata(firstnode(args),ztrdup(buf2));
-		}
+	}
 	argv = makecline(args);
 	fixsigs();
-	if (cn && ISEXCMD(cn->type))
-		{
-		if (cn->type == EXCMD_POSTDOT)
-			zexecve(arg0,argv,&ee,buf2,buf);
-		zexecve(cn->u.nam,argv,&ee,buf2,buf);
-		}
+	if (strlen(arg0) > MAXPATHLEN) {
+		zerr("command too long: %s",arg0,0);
+		_exit(1);
+	}
 	for (s = arg0; *s; s++)
-		if (*s == '/')
-			{
-			zexecve(arg0,argv,NULL,NULL,NULL);
-			if (arg0 == s || unset(PATHDIRS))
-				{
+		if (*s == '/') {
+			errno = zexecve(arg0,argv);
+			if (arg0 == s || unset(PATHDIRS)) {
 				zerr("%e: %s",arg0,errno);
 				_exit(1);
-				}
+			}
 			break;
-			}
-	for (; *path; path++)
-		if ((*path)[0] == '.' && !(*path)[1])
-			zexecve(arg0,argv,&ee,buf2,buf);
-		else
-			{
-			tl = strlen(*path);
-			strcpy(buf,*path);
-			buf[tl] = '/';
-			if (strlen(arg0)+strlen(buf)+1 >= MAXCMDLEN)
-				{
-				zerr("command too long: %s",arg0,0);
-				_exit(1);
-				}
-			strcpy(buf+tl+1,arg0);
-			zexecve(buf,argv,&ee,buf2,buf);
-			}
-	if (ee)
-		{
-		zerr("%e: %s",arg0,ee);
-		_exit(1);
 		}
-	zerr("command not found: %s",arg0,0);
+	if (cn && ISEXCMD(cn->type)) {
+		for (pp = path; pp < cn->pcomp; pp++)
+			if (**pp == '.' && (*pp)[1] == '\0') {
+				ee = zexecve(arg0,argv);
+				if (ee != ENOENT) eno = ee;
+			} else if (**pp != '/') {
+				z = buf;
+				strucpy(&z,*pp);
+				*z++ = '/';
+				strcpy(z,arg0);
+				ee = zexecve(buf,argv);
+				if (ee != ENOENT) eno = ee;
+			}
+		ee = zexecve(cn->u.nam,argv);
+		if (ee != ENOENT) eno = ee;
+	}
+	for (pp = path; *pp; pp++)
+		if ((*pp)[0] == '.' && !(*pp)[1]) {
+			ee = zexecve(arg0,argv);
+			if (ee != ENOENT) eno = ee;
+		} else {
+			z = buf;
+			strucpy(&z,*pp);
+			*z++ = '/';
+			strcpy(z,arg0);
+			ee = zexecve(buf,argv);
+			if (ee != ENOENT) eno = ee;
+		}
+	if (eno) zerr("%e: %s",arg0,eno);
+	else zerr("command not found: %s",arg0,0);
 	_exit(1);
 }
 
-#define try(X) { if (!access(X,X_OK)) return ztrdup(X); }
+#define try(X) { if (iscom(X)) return ztrdup(X); }
 
-/* get the pathname of a command */
+/* get the full pathname of an external command */
 
 char *findcmd(arg0) /**/
 char *arg0;
 {
-char *s,buf[MAXPATHLEN];
-int tl;
-struct cmdnam *cn = (Cmdnam) gethnode(arg0,cmdnamtab);
-char **pp = path;
+char **pp;
+char *z,*s,buf[MAXCMDLEN];
 
-	if (cn && cn->type == DISABLED)
-		cn = NULL;
-	if (cn && ISEXCMD(cn->type))
-		{
-		if (cn->type == EXCMD_POSTDOT)
-			{
-			strcpy(buf,"./");
-			strcat(buf,arg0);
-			try(buf);
+	cn = (Cmdnam) gethnode(arg0,cmdnamtab);
+	if (!cn && isset(HASHCMDS)) hashcmd(arg0,path);
+	if (cn && cn->type == DISABLED) cn = NULL;
+	if (strlen(arg0) > MAXPATHLEN) return NULL;
+	for (s = arg0; *s; s++)
+		if (*s == '/') {
+			try(arg0);
+			if (arg0 == s || unset(PATHDIRS)) {
+				return NULL;
+			}
+			break;
+		}
+	if (cn && ISEXCMD(cn->type)) {
+		for (pp = path; pp < cn->pcomp; pp++)
+			if (**pp != '/') {
+				z = buf;
+				strucpy(&z,*pp);
+				*z++ = '/';
+				strcpy(z,arg0);
+				try(buf);
 			}
 		try(cn->u.nam);
-		}
-	for (s = arg0; *s; s++)
-		if (*s == '/')
-			{
-			try(arg0);
-			if (s == arg0 || unset(PATHDIRS))
-				goto failed;
-			break;
-			}
-	for (; *pp; pp++)
-		if (**pp == '.')
-			{
-			strcpy(buf,"./");
-			strcat(buf,arg0);
-			try(buf);
-			}
-		else
-			{
-			tl = strlen(*pp);
-			strcpy(buf,*pp);
-			buf[tl] = '/';
-			strcpy(buf+tl+1,arg0);
-			try(buf);
-			}
-failed:
+	}
+	for (pp = path; *pp; pp++) {
+		z = buf;
+		strucpy(&z,*pp);
+		*z++ = '/';
+		strcpy(z,arg0);
+		try(buf);
+	}
 	return NULL;
 }
 
+int iscom(s) /**/
+char *s;
+{
+struct stat statbuf;
+
+	return (access(s,X_OK) == 0 && stat(s,&statbuf) >= 0 &&
+			S_ISREG(statbuf.st_mode));
+}
+
+int isrelative(s) /**/
+char *s;
+{
+	if (*s != '/') return 1;
+	for (; *s; s++)
+		if (*s == '.' && s[-1] == '/' &&
+			(s[1] == '/' || s[1] == '\0' ||
+				(s[1] == '.' && (s[2] == '/' || s[2] == '\0')))) return 1;
+	return 0;
+}
+
+int hashcmd(arg0,pp) /**/
+char *arg0;char **pp;
+{
+char *s,buf[MAXPATHLEN];
+char **pq;
+DIR *dir;
+struct direct *de;
+
+	for (; *pp; pp++)
+		if (**pp == '/') {
+			s = buf;
+			strucpy(&s,*pp);
+			*s++ = '/';
+			strcpy(s,arg0);
+			if (iscom(buf)) break;
+		}
+	if (!*pp || isrelative(*pp)) return 0;
+	cn = (Cmdnam) zcalloc(sizeof *cn);
+	cn->type = EXCMD;
+	cn->pcomp = pp;
+	cn->u.nam = ztrdup(buf);
+	addhnode(ztrdup(arg0),cn,cmdnamtab,freecmdnam);
+	if (unset(HASHDIRS)) return 1;
+	for (pq = pathchecked; pq <= pp; pq++) {
+		if (isrelative(*pq) || !(dir = opendir(*pq))) continue;
+		readdir(dir); readdir(dir);
+		while (de = readdir(dir)) addhcmdnode(de->d_name,pq);
+		closedir(dir);
+	}
+	pathchecked = pp+1;
+	return 1;
+}
+
+void fullhash() /**/
+{
+char **pq;
+DIR *dir;
+struct direct *de;
+
+	for (pq = pathchecked; *pq; pq++) {
+		if (isrelative(*pq) || !(dir = opendir(*pq))) continue;
+		readdir(dir); readdir(dir);
+		while (de = readdir(dir)) addhcmdnode(de->d_name,pq);
+		closedir(dir);
+	}
+	pathchecked = pq;
+}
+
 void execlist(list) /**/
 List list;
 {
-	if (breaks)
-		return;
+	if (breaks) return;
+	if (!list) return;
 	simplifyright(list);
-	switch(list->type)
-		{
+	switch(list->type) {
 		case SYNC:
 		case ASYNC:
 			execlist2(list->left,list->type,!list->right);
@@ -348,26 +368,34 @@ List list;
 			if (list->right && !retflag)
 				execlist(list->right);
 			break;
-		}
+	}
 }
 
 void execlist2(list,type,last1) /**/
 Sublist list;int type;int last1;
 {
-	switch(list->type)
-		{
+	if (!list) return;
+	switch(list->type) {
 		case END:
 			execpline(list,type,last1);
 			break;
 		case ORNEXT:
-			if (!execpline(list,SYNC,0))
-				execlist2(list->right,type,last1);
+			if (!execpline(list,SYNC,0)) execlist2(list->right,type,last1);
+			else while (list = list->right)
+				if (list->type == ANDNEXT) {
+					execlist2(list->right,type,last1);
+					return;
+				}
 			break;
 		case ANDNEXT:
-			if (execpline(list,SYNC,0))
-				execlist2(list->right,type,last1);
+			if (execpline(list,SYNC,0)) execlist2(list->right,type,last1);
+			else while (list = list->right)
+				if (list->type == ORNEXT) {
+					execlist2(list->right,type,last1);
+					return;
+				}
 			break;
-		}
+	}
 }
 
 int execpline(l,how,last1) /**/
@@ -375,44 +403,39 @@ Sublist l;int how;int last1;
 {
 int ipipe[2],opipe[2];
 
+	if (!l) return 0;
 	ipipe[0] = ipipe[1] = opipe[0] = opipe[1] = 0;
 	blockchld();
 	if ((thisjob = getfreejob()) == -1)
 		return 1;
 	initjob();
-	if (how == TIMED)
-		{
+	if (how == TIMED) {
 		jobtab[thisjob].stat |= STAT_TIMED;
 		how = SYNC;
-		}
-	if (l->flags & PFLAG_COPROC)
-		{
+	}
+	if (l->flags & PFLAG_COPROC) {
 		how = ASYNC;
-		mpipe(ipipe);
-		mpipe(opipe);
-		if (coprocin)
-			{
+		if (coprocin >= 0) {
 			close(coprocin);
 			close(coprocout);
-			}
+		}
+		mpipe(ipipe);
+		mpipe(opipe);
 		coprocin = ipipe[0];
 		coprocout = opipe[1];
-		}
+	}
 	execpline2(l->left,how,opipe[0],ipipe[1],last1);
-	if (how == ASYNC)
-		{
+	if (how == ASYNC) {
+		if (l->flags & PFLAG_COPROC) close(ipipe[1]);
 		spawnjob();
 		unblockchld();
 		return 1;
-		}
-	else
-		{
+	} else {
 		waitjobs();
 		unblockchld();
-		if (l->flags & PFLAG_NOT)
-			lastval = !lastval;
+		if (l->flags & PFLAG_NOT) lastval = !lastval;
 		return !lastval;
-		}
+	}
 }
 
 void execpline2(pline,how,input,output,last1) /**/
@@ -425,53 +448,40 @@ int pipes[2];
 		return;
 	if (!pline)
 		return;
-	if (pline->type == END)
-		{
+	if (pline->type == END) {
 		execcmd(pline->left,input,output,how==ASYNC,last1);
 		pline->left = NULL;
-		}
-	else
-		{
+	} else {
 		mpipe(pipes);
+		if (pline->left->type >= CURSH && how == SYNC) {
 
-		/* if we are doing "foo | bar" where foo is a current
-			shell command, do foo in the current shell and do
-			the rest of the pipeline in a subshell. */
+			/* if we are doing "foo | bar" where foo is a current
+				shell command, do foo in a subshell and do
+				the rest of the pipeline in the current shell. */
 
-		if (pline->left->type >= CURSH && how == SYNC)
-			{
-			if (!(pid = fork()))
-				{
-				close(pipes[1]);
+			if (!(pid = fork())) {
+				close(pipes[0]);
 				entersubsh(how==ASYNC);
 				exiting = 1;
-				execpline2(pline->right,ASYNC,pipes[0],output,1);
+				execcmd(pline->left,input,pipes[1],how==ASYNC,0);
 				_exit(lastval);
-				}
-			else if (pid == -1)
+			} else if (pid == -1)
 				zerr("fork failed: %e",NULL,errno);
-			else
-				{
-				char *s,*text;
-
-				close(pipes[0]);
-				text = s = gettext((vptr) pline->right,0);
-				addproc(pid,text)->lastfg = 1;
-				pline->right = NULL;
-				}
+			else {
+				char *text = getjobtext((vptr) pline->left);
+				addproc(pid,text);
 			}
-
-		/* otherwise just do the pipeline normally. */
-
-		execcmd(pline->left,input,pipes[1],how==ASYNC,0);
+		} else {
+			/* otherwise just do the pipeline normally. */
+			execcmd(pline->left,input,pipes[1],how==ASYNC,0);
+		}
 		pline->left = NULL;
 		close(pipes[1]);
-		if (pline->right)
-			{
+		if (pline->right) {
 			execpline2(pline->right,how,pipes[0],output,last1);
 			close(pipes[0]);
-			}
 		}
+	}
 }
 
 /* make the argv array */
@@ -483,38 +493,33 @@ int ct = 0;
 Lknode node;
 char **argv,**ptr;
 
-	if (isset(XTRACE))
-		{
+	if (isset(XTRACE)) {
 		fprintf(stderr,"%s",(prompt4) ? prompt4 : "");
 		for (node = firstnode(list); node; incnode(node),ct++);
 		ptr = argv = 2+(char **) ncalloc((ct+4)*sizeof(char *));
 		for (node = firstnode(list); node; incnode(node))
-			if (*(char *) getdata(node))
-				{
+			if (*(char *) getdata(node)) {
 				*ptr++ = getdata(node);
 				untokenize(getdata(node));
 				fputs(getdata(node),stderr);
 				if (nextnode(node))
 					fputc(' ',stderr);
-				}
+			}
 		*ptr = NULL;
 		fputc('\n',stderr);
 		fflush(stderr);
 		return(argv);
-		}
-	else
-		{
+	} else {
 		for (node = firstnode(list); node; incnode(node),ct++);
 		ptr = argv = 2+(char **) ncalloc((ct+4)*sizeof(char *));
 		for (node = firstnode(list); node; incnode(node))
-			if (*(char *) getdata(node))
-				{
+			if (*(char *) getdata(node)) {
 				*ptr++ = getdata(node);
 				untokenize(getdata(node));
-				}
+			}
 		*ptr = NULL;
 		return(argv);
-		}
+	}
 }
 
 /* untokenize the command line and remove null arguments */
@@ -524,14 +529,11 @@ Lklist l;
 {
 Lknode node,next;
 
-	for (node = firstnode(l); node; node = next)
-		{
+	for (node = firstnode(l); node; node = next) {
 		next = nextnode(node);
-		if (!*(char *) getdata(node))
-			uremnode(l,node);
-		else
-			untokenize(getdata(node));
-		}
+		if (!*(char *) getdata(node)) uremnode(l,node);
+		else untokenize(getdata(node));
+	}
 }
 
 void untokenize(s) /**/
@@ -539,10 +541,8 @@ char *s;
 {
 	for (; *s; s++)
 		if (itok(*s))
-			if (*s == Nularg)
-				chuck(s--);
-			else
-				*s = ztokens[*s-Pound];
+			if (*s == Nularg) chuck(s--);
+			else *s = ztokens[*s-Pound];
 }
 
 /* nonzero if we shouldn't clobber a file */
@@ -552,10 +552,8 @@ struct redir *f;
 {
 struct stat buf;
 
-	if (unset(NOCLOBBER) || f->type & 1)
-		return 0;
-	if (stat(f->name,&buf) == -1)
-		return 1;
+	if (unset(NOCLOBBER) || f->type & 1) return 0;
+	if (stat(f->name,&buf) == -1) return 1;
 	return S_ISREG(buf.st_mode);
 }
 
@@ -564,15 +562,14 @@ struct stat buf;
 void closemn(mfds,fd) /**/
 struct multio **mfds;int fd;
 {
-	if (mfds[fd])
-		{
+	if (mfds[fd]) {
 		if (mfds[fd]->ct > 1)
 			if (mfds[fd]->rflag == 0)
 				catproc(mfds[fd]);
 			else
 				teeproc(mfds[fd]);
 		mfds[fd] = NULL;
-		}
+	}
 }
 
 /* close all the mnodes (failure) */
@@ -583,12 +580,11 @@ struct multio **mfds;
 int t0,t1;
 
 	for (t0 = 0; t0 != 10; t0++)
-		if (mfds[t0])
-			{
+		if (mfds[t0]) {
 			for (t1 = 0; t1 != mfds[t0]->ct; t1++)
 				close(mfds[t0]->fds[t1]);
 			mfds[t0] = NULL;
-			}
+		}
 }
 
 /* add a fd to an multio */
@@ -602,8 +598,7 @@ int forked;int *save;struct multio **mfds;int fd1;int fd2;int rflag;
 {
 int pipes[2];
 
-	if (!mfds[fd1])	/* starting a new multio */
-		{
+	if (!mfds[fd1]) {	/* starting a new multio */
 		mfds[fd1] = (struct multio *) alloc(sizeof(struct multio));
 		if (!forked && fd1 != fd2 && fd1 < 10)
 			save[fd1] = movefd(fd1);
@@ -611,26 +606,21 @@ int pipes[2];
 		mfds[fd1]->ct = 1;
 		mfds[fd1]->fds[0] = fd1;
 		mfds[fd1]->rflag = rflag;
-		}
-	else
-		{
-		if (mfds[fd1]->rflag != rflag)
-			{
+	} else {
+		if (mfds[fd1]->rflag != rflag) {
 			zerr("file mode mismatch on fd %d",NULL,fd1);
 			return;
-			}
-		if (mfds[fd1]->ct == 1)		/* split the stream */
-			{
+		}
+		if (mfds[fd1]->ct == 1) {		/* split the stream */
 			mfds[fd1]->fds[0] = movefd(fd1);
 			mfds[fd1]->fds[1] = movefd(fd2);
 			mpipe(pipes);
 			mfds[fd1]->pipe = pipes[1-rflag];
 			redup(pipes[rflag],fd1);
 			mfds[fd1]->ct = 2;
-			}
-		else		/* add another fd to an already split stream */
+		} else		/* add another fd to an already split stream */
 			mfds[fd1]->fds[mfds[fd1]->ct++] = movefd(fd2);
-		}
+	}
 }
 
 void addvars(l,export) /**/
@@ -639,8 +629,7 @@ Lklist l;int export;
 struct varasg *v;
 Lklist vl;
 
-	while (full(l))
-		{
+	while (full(l)) {
 		char **arr,**ptr;
 
 		v = (struct varasg *) ugetnode(l);
@@ -648,12 +637,10 @@ Lklist vl;
 		if (errflag)
 			return;
 		untokenize(v->name);
-		if (v->type == PMFLAG_s)
-			{
+		if (v->type == PMFLAG_s) {
 			vl = newlist();
 			addnode(vl,v->str);
-			}
-		else
+		} else
 			vl = v->arr;
 		prefork(vl);
 		if (errflag)
@@ -661,31 +648,32 @@ Lklist vl;
 		postfork(vl,1);
 		if (errflag)
 			return;
-		if (v->type == PMFLAG_s && (!full(vl) || !nextnode(firstnode(vl))))
-			{
+		if (v->type == PMFLAG_s && (!full(vl) || !nextnode(firstnode(vl)))) {
 			Param pm;
 			char *val;
 
 			if (!full(vl))
 				pm = setsparam(v->name,val = ztrdup(""));
-			else
-				{
+			else {
 				untokenize(peekfirst(vl));
 				pm = setsparam(v->name,val = ztrdup(ugetnode(vl)));
-				}
+			}
 			if (export && !(pm->flags & PMFLAG_x))
 				addenv(v->name,val);
 			continue;
-			}
+		}
 		ptr = arr = (char **) zalloc(sizeof(char **)*(countnodes(vl)+1));
-		while (full(vl))
-			{
-			*ptr = ztrdup(ugetnode(vl));
-			untokenize(*ptr++);
+		while (full(vl)) {
+			char *pp;
+			pp = ugetnode(vl);
+			if (*pp) {
+				*ptr = ztrdup(pp);
+				untokenize(*ptr++);
 			}
+		}
 		*ptr = NULL;
 		setaparam(v->name,arr);
-		}
+	}
 }
 
 void execcmd(cmd,input,output,bkg,last1) /**/
@@ -693,84 +681,92 @@ Cmd cmd;int input;int output;int bkg;int last1;
 {
 int type;
 long pid = 0;
-Lklist args = cmd->args;
 int save[10],t0;
 struct redir *fn;
 struct multio *mfds[10];
 int fil,forked = 0,iscursh = 0,nullexec = 0;
-Cmdnam chn = NULL;
 char *text;
 
-	for (t0 = 0; t0 != 10; t0++)
-		{
+	args = cmd->args;
+	cn = NULL;
+	for (t0 = 0; t0 != 10; t0++) {
 		save[t0] = 0;
 		mfds[t0] = NULL;
-		}
+	}
 	if ((type = cmd->type) == SIMPLE && !full(args))
 		if (full(cmd->redir))
-			if (cmd->flags & CFLAG_EXEC)
+			if (cmd->flags & CFLAG_EXEC) {
 				nullexec = 1;
-			else if (!*nullcmd)
-				{
+			} else if (!*nullcmd) {
 				zerr("redirection with no command",NULL,0);
 				errflag = lastval = 1;
 				return;
-				}
-			else
+			} else if (*readnullcmd &&
+					((Redir)peekfirst(cmd->redir))->type == READ &&
+					!nextnode(firstnode(cmd->redir))) {
+				addnode(args,strdup(readnullcmd));
+			} else
 				addnode(args,strdup(nullcmd));
-		else
-			{
+		else {
 			addvars(cmd->vars,0);
 			return;
-			}
-	if (full(args) && *(char *) peekfirst(args) == '%')
-		{
+		}
+	if (full(args) && *(char *) peekfirst(args) == '%') {
 		insnode(args,(Lknode) args,strdup((bkg) ? "bg" : "fg"));
 		bkg = 0;
-		}
+	}
 	if (isset(AUTORESUME) && !bkg && !full(cmd->redir) && full(args) &&
-			!input && type == SIMPLE && !nextnode(firstnode(args)) &&
-			findjobnam(peekfirst(args)) != -1)
-		pushnode(args,strdup("fg"));
+			!input && type == SIMPLE && !nextnode(firstnode(args))) {
+		if (unset(NOTIFY)) scanjobs();
+		if (findjobnam(peekfirst(args)) != -1)
+			pushnode(args,strdup("fg"));
+	}
 	if (unset(RMSTARSILENT) && interact && isset(SHINSTDIN) &&
 				type == SIMPLE && full(args) && nextnode(firstnode(args)) &&
-				!strcmp(peekfirst(args),"rm"))
-			{
+				!strcmp(peekfirst(args),"rm")) {
 			char *s = getdata(nextnode(firstnode(args)));
+			int l = strlen(s);
 
 			if (s[0] == Star && !s[1])
-				checkrmall();
+				checkrmall(pwd);
+			else if (l > 2 && s[l-2] == '/' && s[l-1] == Star) {
+				char t = s[l-2];
+				s[l-2] = 0;
+				checkrmall(s);
+				s[l-2] = t;
 			}
-	if (jobbing)	/* get the text associated with this command */
-		{
+	}
+	if (jobbing) {	/* get the text associated with this command */
 		char *s;
-		s = text = gettext((vptr) cmd,0);
-		}
-	else
-		text = NULL;
+		s = text = getjobtext((vptr) cmd);
+	} else text = NULL;
 	prefork(args);	/* do prefork substitutions */
-	if (errflag)
-		{
+	if (errflag) {
 		lastval = 1;
 		return;
-		}
+	}
 	if (full(args) && ((char*)peekfirst(args))[0] == Inbrack &&
 			((char*)peekfirst(args))[1] == '\0')
 		((char*)peekfirst(args))[0] = '[';
-	if (full(args) && !(cmd->flags & CFLAG_COMMAND))
-		chn = (Cmdnam) gethnode(peekfirst(args),cmdnamtab);
-	if (type == SIMPLE && !pathsuppress && !chn && isset(AUTOCD) &&
-			full(args) && !full(cmd->redir) &&
-			!nextnode(firstnode(args)) && cancd(peekfirst(args)))
-		{
-		pushnode(args,strdup("cd"));
-		chn = (Cmdnam) gethnode("cd",cmdnamtab);
+	if (type == SIMPLE && full(args) && !(cmd->flags & CFLAG_COMMAND)) {
+		char *s,*t;
+		cn = (Cmdnam) gethnode(t = s = peekfirst(args),cmdnamtab);
+		if (!cn && isset(HASHCMDS) && strcmp(t,"..")) {
+			while (*t && *t != '/') t++;
+			if (!*t) hashcmd(s,pathchecked);
 		}
+	}
+	if (type == SIMPLE && !cn && isset(AUTOCD) && isset(SHINSTDIN) &&
+			full(args) && !full(cmd->redir) &&
+			!nextnode(firstnode(args)) && cancd(peekfirst(args))) {
+		pushnode(args,strdup("cd"));
+		cn = (Cmdnam) gethnode("cd",cmdnamtab);
+	}
 
 	/* this is nonzero if cmd is a current shell procedure */
 
-	iscursh = (type >= CURSH) || (type == SIMPLE && chn &&
-		(chn->type == BUILTIN || chn->type == SHFUNC));
+	iscursh = (type >= CURSH) || (type == SIMPLE && cn &&
+		(cn->type == BUILTIN || cn->type == SHFUNC));
 
 	/* if this command is backgrounded or (this is an external
 		command and we are not exec'ing it) or this is a builtin
@@ -779,121 +775,110 @@ char *text;
 		the rest of the function think we forked. */
 
 	if (bkg || !(iscursh || (cmd->flags & CFLAG_EXEC)) ||
-			(chn && (chn->type == BUILTIN || chn->type == SHFUNC) && output))
-		{
+			(cn && (cn->type == BUILTIN || cn->type == SHFUNC) && output)) {
 		int synch[2];
 
 		pipe(synch);
 		pid = (last1 && execok()) ? 0 : phork();
-		if (pid == -1)
-			{
+		if (pid == -1) {
 			close(synch[0]);
 			close(synch[1]);
 			return;
-			}
-		if (pid)
-			{
+		}
+		if (pid) {
 			close(synch[1]);
 			read(synch[0],"foo",1);
 			close(synch[0]);
 			if (pid == -1)
 				zerr("%e",NULL,errno);
-			else
+			else {
+				if (bkg) lastpid = pid;
 				( void ) addproc(pid,text);
-			return;
 			}
+			return;
+		}
 		close(synch[0]);
 		entersubsh(bkg);
 		close(synch[1]);
 		forked = 1;
-		}
-	if (bkg && isset(BGNICE))	/* stupid */
+	}
+	if (bkg && isset(BGNICE))
 		nice(5);
+
+	/* perform postfork substitutions */
+	postfork(args,!(cmd->flags & CFLAG_NOGLOB));
+	if (errflag) {
+		lastval = 1;
+		goto err;
+	} else {
+		char *s;
+		while (full(args) && (s = peekfirst(args)) && !*s) ugetnode(args);
+	}
+
 	if (input)		/* add pipeline input/output to mnodes */
 		addfd(forked,save,mfds,0,input,0);
 	if (output)
 		addfd(forked,save,mfds,1,output,1);
 	spawnpipes(cmd->redir);		/* do process substitutions */
 	while (full(cmd->redir))
-		if ((fn = (struct redir*) ugetnode(cmd->redir))->type == INPIPE)
-			{
+		if ((fn = (struct redir*) ugetnode(cmd->redir))->type == INPIPE) {
 			if (fn->fd2 == -1)
 				execerr();
 			addfd(forked,save,mfds,fn->fd1,fn->fd2,0);
-			}
-		else if (fn->type == OUTPIPE)
-			{
+		} else if (fn->type == OUTPIPE) {
 			if (fn->fd2 == -1)
 				execerr();
 			addfd(forked,save,mfds,fn->fd1,fn->fd2,1);
-			}
-		else
-			{
+		} else {
 			if (!(fn->type == HERESTR || fn->type == CLOSE || fn->type ==
 					MERGE || fn->type == MERGEOUT))
 				if (xpandredir(fn,cmd->redir))
 					continue;
-			if (fn->type == HERESTR)
-				{
+			if (errflag) execerr();
+			if (fn->type == HERESTR) {
 				fil = getherestr(fn);
-				if (fil == -1)
-					{
+				if (fil == -1) {
 					if (errno != EINTR)
 						zerr("%e",NULL,errno);
 					execerr();
-					}
-				addfd(forked,save,mfds,fn->fd1,fil,0);
 				}
-			else if (fn->type == READ)
-				{
+				addfd(forked,save,mfds,fn->fd1,fil,0);
+			} else if (fn->type == READ) {
 				fil = open(fn->name,O_RDONLY);
-				if (fil == -1)
-					{
+				if (fil == -1) {
 					if (errno != EINTR)
 						zerr("%e: %s",fn->name,errno);
 					execerr();
-					}
-				addfd(forked,save,mfds,fn->fd1,fil,0);
 				}
-			else if (fn->type == CLOSE)
-				{
+				addfd(forked,save,mfds,fn->fd1,fil,0);
+			} else if (fn->type == CLOSE) {
 				if (!forked && fn->fd1 < 10)
 					save[fn->fd1] = movefd(fn->fd1);
 				closemn(mfds,fn->fd1);
 				close(fn->fd1);
-				}
-			else if (fn->type == MERGE || fn->type == MERGEOUT)
-				{
+			} else if (fn->type == MERGE || fn->type == MERGEOUT) {
+				if (fn->fd2 == FD_COPROC)
+					fn->fd2 = (fn->type == MERGEOUT) ? coprocout : coprocin;
+				if (!forked && fn->fd1 < 10)
+					save[fn->fd1] = movefd(fn->fd1);
+				closemn(mfds,fn->fd1);
 				fil = dup(fn->fd2);
-				if (mfds[fn->fd1])
-					redup(fil,fn->fd1);
-				else
-					addfd(forked,save,mfds,fn->fd1,fil,fn->type == MERGEOUT);
-				}
-			else
-				{
+				addfd(forked,save,mfds,fn->fd1,fil,fn->type == MERGEOUT);
+			} else {
 				if (fn->type >= APP)
-					fil = open(fn->name,isset(NOCLOBBER) ?
+					fil = open(fn->name,
+						(isset(NOCLOBBER) && !(fn->type & 1)) ?
 						O_WRONLY|O_APPEND : O_WRONLY|O_APPEND|O_CREAT,0666);
 				else
 					fil = open(fn->name,dontclob(fn) ? 
 						O_WRONLY|O_CREAT|O_EXCL : O_WRONLY|O_CREAT|O_TRUNC,0666);
-				if (fil == -1)
-					{
+				if (fil == -1) {
 					if (errno != EINTR)
 						zerr("%e: %s",fn->name,errno);
 					execerr();
-					}
-				addfd(forked,save,mfds,fn->fd1,fil,1);
 				}
+				addfd(forked,save,mfds,fn->fd1,fil,1);
 			}
-	/* perform postfork substitutions */
-	/* you put this here so that ls *.sdfoij >& /dev/null works, remember? */
-	postfork(args,!(cmd->flags & CFLAG_NOGLOB));
-	if (errflag)
-		{
-		lastval = 1;
-		goto err;
 		}
 	
 	/* we are done with redirection.  close the mnodes, spawning
@@ -904,27 +889,30 @@ char *text;
 	if (nullexec)
 		return;
 	if (unset(NOEXEC))
-		if (type >= ZCTIME)
+		if (type >= CURSH)
 			{
 			static int (*func[]) DCLPROTO((Cmd)) = {
-				exectime,execcursh,execfuncdef,execfor,execwhile,
+				execcursh,exectime,execfuncdef,execfor,execwhile,
 				execrepeat,execif,execcase,execselect,execcond };
 	
 			fixcline(args);
-			lastval = (func[type-ZCTIME])(cmd);
+			lastval = (func[type-CURSH])(cmd);
 			}
 		else if (iscursh)		/* builtin or shell function */
 			{
 			if (cmd->vars)
 				addvars(cmd->vars,0);
 			fixcline(args);
-			if (chn && chn->type == SHFUNC)
-				execshfunc(cmd,chn);
+			if (cn && cn->type == SHFUNC)
+				execshfunc(cmd,cn);
 			else
 				{
-				lastval = execbin(args,chn);
-				if (isset(PRINTEXITVALUE) && lastval)
-					zerr("exit %d",NULL,lastval);
+				if (forked) closem();
+				lastval = execbin(args,cn);
+				if (isset(PRINTEXITVALUE) && isset(SHINSTDIN) &&
+						lastval && !subsh) {
+					fprintf(stderr,"zsh: exit %d\n",lastval);
+				}
 				fflush(stdout);
 				if (ferror(stdout))
 					{
@@ -940,7 +928,7 @@ char *text;
 			if (type == SIMPLE)
 				{
 				closem();
-				execute(args,cmd->flags & CFLAG_DASH);
+				execute(cmd->flags & CFLAG_DASH);
 				}
 			else	/* ( ... ) */
 				execlist(cmd->u.list);
@@ -1046,7 +1034,7 @@ char *s,*t,*bptr;
 		{
 		char *u,*v;
 
-		if (!fgets(pbuf,256,bshin))
+		if (!hgets(pbuf,256))
 			break;
 		bptr = pbuf;
 		if (strip)
@@ -1072,16 +1060,16 @@ char *s,*t,*bptr;
 		t[siz-1] = '\0';
 	if (!qt)
 		for (s = t; *s; s++)
-			if (*s == '$')
+			if (*s == '$') {
 				*s = Qstring;
-			else if (*s == '`')
+			} else if (*s == '`') {
 				*s = Qtick;
-			else if (*s == '\\')
-				{
-				s++;
-				if (!*s)
-					break;
-				}
+			} else if (*s == '(') {
+				*s = Inpar;
+			} else if (*s == ')') {
+				*s = Outpar;
+			} else if (*s == '\\' &&
+				(s[1] == '$' || s[1] == '`')) chuck(s);
 	s = strdup(t);
 	free(t);
 	return s;
@@ -1202,8 +1190,7 @@ char *cmd;int qt;
 List list;
 int pipes[2];
 
-	if (*cmd == '<')
-		{
+	if (*cmd == '<') {
 		int stream;
 		char *fi,*s,x;
 
@@ -1223,13 +1210,12 @@ int pipes[2];
 		if (errflag)
 			return NULL;
 		stream = open(fi,O_RDONLY);
-		if (stream == -1)
-			{
-			zerr("%e: %s",cmd+1,errno);
+		if (stream == -1) {
+			zerr("%e: %s",fi,errno);
 			return NULL;
-			}
-		return readoutput(stream,qt);
 		}
+		return readoutput(stream,qt);
+	}
 	if (!(list = parselstring(cmd)))
 		return NULL;
 	mpipe(pipes);
@@ -1263,37 +1249,30 @@ FILE *fin;
 
 	fin = fdopen(in,"r");
 	ret = newlist();
-	ptr = buf = zalloc(bsiz = 256);
+	ptr = buf = ncalloc(bsiz = 64);
 	while ((c = fgetc(fin)) != EOF)
-		if (!qt && isep(c))
-			{
-			if (cnt)
-				{
+		if (!qt && isep(c)) {
+			if (cnt) {
 				*ptr = '\0';
-				addnode(ret,ztrdup(buf));
+				addnode(ret,buf);
+				ptr = buf = ncalloc(bsiz = 64);
 				cnt = 0;
 				ptr = buf;
-				}
 			}
-		else
-			{
+		} else {
 			*ptr++ = c;
-			if (++cnt == bsiz)
-				{
-				char *pp = zalloc(bsiz *= 2);
+			if (++cnt == bsiz) {
+				char *pp = ncalloc(bsiz *= 2);
 				
 				memcpy(pp,buf,cnt);
-				free(buf);
 				ptr = (buf = pp)+cnt;
-				}
 			}
+		}
 	if (ptr != buf && ptr[-1] == '\n')
 		ptr[-1] = '\0';
 	else
 		*ptr = '\0';
-	if (cnt)
-		addnode(ret,ztrdup(buf));
-	free(buf);
+	if (cnt) addnode(ret,buf);
 	fclose(fin);
 	return ret;
 }
@@ -1331,8 +1310,10 @@ List list;
 		return nam;
 		}
 #else
-	if (waitfork())
+	if (waitfork()) {
+		popheap();
 		return nam;
+	}
 #endif
 	subsh = 1;
 	close(1);
@@ -1348,10 +1329,14 @@ List list;
 
 char *namedpipe() /**/
 {
+#ifndef NO_FIFOS
 char *tnam = gettemp();
 
-	mknod(tnam,0010666,0);
+	if (mknod(tnam,0010666,0) < 0) return NULL;
 	return tnam;
+#else
+	return NULL;
+#endif
 }
 
 /* <(...) */
@@ -1359,6 +1344,10 @@ char *tnam = gettemp();
 char *getoutproc(cmd) /**/
 char *cmd;
 {
+#ifdef NO_FIFOS
+	zerr("doesn't look like your system supports FIFOs.",NULL,0);
+	return NULL;
+#else
 List list;
 int fd;
 char *pnam,*str;
@@ -1370,6 +1359,7 @@ char *pnam,*str;
 		zerr("oops.",NULL,0);
 	*str = '\0';
 	pnam = namedpipe();
+	if (!pnam) return NULL;
 	permalloc();
 	if (!jobtab[thisjob].filelist)
 		jobtab[thisjob].filelist = newlist();
@@ -1383,6 +1373,7 @@ char *pnam,*str;
 		return pnam;
 		}
 	entersubsh(1);
+	closem();
 	fd = open(pnam,O_WRONLY);
 	if (fd == -1)
 		{
@@ -1396,6 +1387,7 @@ char *pnam,*str;
 	execlist(list);
 	close(1);
 	_exit(0);  return NULL;
+#endif
 }
 
 /* >(...) */
@@ -1403,6 +1395,10 @@ char *pnam,*str;
 char *getinproc(cmd) /**/
 char *cmd;
 {
+#ifdef NO_FIFOS
+	zerr("doesn't look like your system supports FIFOs.",NULL,0);
+	return NULL;
+#else
 List list;
 int pid,fd;
 char *pnam,*str;
@@ -1414,6 +1410,7 @@ char *pnam,*str;
 		zerr("oops.",NULL,0);
 	*str = '\0';
 	pnam = namedpipe();
+	if (!pnam) return NULL;
 	permalloc();
 	if (!jobtab[thisjob].filelist)
 		jobtab[thisjob].filelist = newlist();
@@ -1427,11 +1424,13 @@ char *pnam,*str;
 		return pnam;
 		}
 	entersubsh(1);
+	closem();
 	fd = open(pnam,O_RDONLY);
 	redup(fd,0);
 	exiting = 1;
 	execlist(list);
 	_exit(0);  return NULL;
+#endif
 }
 
 /* > >(...) (does not use named pipes) */
@@ -1457,6 +1456,7 @@ char *str = cmd;
 		return pipes[0];
 		}
 	close(pipes[0]);
+	closem();
 	entersubsh(1);
 	redup(pipes[1],1);
 	exiting = 1;
@@ -1488,6 +1488,7 @@ char *str;
 		return pipes[1];
 		}
 	close(pipes[1]);
+	closem();
 	entersubsh(1);
 	redup(pipes[0],0);
 	exiting = 1;
@@ -1626,6 +1627,7 @@ Cmd cmd;
 {
 int jb = thisjob;
 
+	if (!cmd->u.pline) { shelltime(); return 0; }
 	execpline(cmd->u.pline,TIMED,0);
 	thisjob = jb;
 	return lastval;
@@ -1673,66 +1675,82 @@ Cmd cmd;
 void execshfunc(cmd,cn) /**/
 Cmd cmd;Cmdnam cn;
 {
-char **tab,**x,*oargv0;
-int oxtr = opts[XTRACE],flags,xexittr;
 List l;
-Lklist olist;
-char *s;
-List xexitfn;
 
-	if (errflag)
-		return;
+	if (errflag) return;
 	l = cn->u.list;
-	if (!l)
-		{
+	if (!l) {
 		char *nam;
 
-		if (!(l = getfpfunc(nam = peekfirst(cmd->args))))
-			{
+		if (!(cn->flags & PMFLAG_u)) return;
+		if (!(l = getfpfunc(nam = peekfirst(cmd->args)))) {
 			zerr("function not found: %s",nam,0);
 			lastval = 1;
 			return;
-			}
+		}
 		cn->flags &= ~PMFLAG_u;
 		permalloc();
 		cn->u.list = (List) dupstruct(l);
 		heapalloc();
-		}
-	flags = cn->flags;
+	}
+	doshfunc(l,cmd->args,cn->flags);
+}
+
+void doshfuncnoval(list,args,flags) /**/
+List list; Lklist args; int flags;
+{
+int val = lastval;
+
+	doshfunc(list,args,flags);
+	lastval = val;
+}
+
+void doshfunc(list,args,flags) /**/
+List list; Lklist args; int flags;
+{
+char **tab,**x,*oargv0;
+int oxtr = opts[XTRACE],opev = opts[PRINTEXITVALUE],xexittr;
+Lklist olist;
+char *s;
+List xexitfn;
+
 	xexittr = sigtrapped[SIGEXIT];
 	xexitfn = sigfuncs[SIGEXIT];
 	tab = pparams;
 	oargv0 = argzero;
-	optind = 0;
-	if (flags & PMFLAG_t)
-		opts[XTRACE] = OPT_SET;
-	pparams = x = (char **) zcalloc(((sizeof *x)*(1+countnodes(cmd->args))));
-	argzero = ztrdup(ugetnode(cmd->args));
-	while (*x = ugetnode(cmd->args))
-		*x = ztrdup(*x), x++;
+	zoptind = 1;
+	if (flags & PMFLAG_t) opts[XTRACE] = OPT_SET;
+	opts[PRINTEXITVALUE] = OPT_UNSET;
+	if (args) {
+		pparams = x = (char **) zcalloc(((sizeof *x)*(1+countnodes(args))));
+		argzero = ztrdup(ugetnode(args));
+		while (*x = ugetnode(args))
+			*x = ztrdup(*x), x++;
+	} else {
+		pparams = zcalloc(sizeof *pparams);
+		argzero = ztrdup(argzero);
+	}
 	permalloc();
 	olist = locallist;
 	locallist = newlist();
 	heapalloc();
-	newrunlist(l);
-	while (s = getnode(locallist))
-		unsetparam(s);
+	runlist(dupstruct(list));
+	while (s = getnode(locallist)) unsetparam(s);
 	free(locallist);
 	locallist = olist;
 	retflag = 0;
-	cmd->u.list = NULL;
 	freearray(pparams);
 	free(argzero);
 	argzero = oargv0;
 	pparams = tab;
-	if (sigfuncs[SIGEXIT] && sigfuncs[SIGEXIT] != xexitfn)
-		{
+	if (sigfuncs[SIGEXIT] && sigfuncs[SIGEXIT] != xexitfn) {
 		dotrap(SIGEXIT);
 		freestruct(sigfuncs[SIGEXIT]);
-		}
+	}
 	sigtrapped[SIGEXIT] = xexittr;
 	sigfuncs[SIGEXIT] = xexitfn;
 	opts[XTRACE] = oxtr;
+	opts[PRINTEXITVALUE] = opev;
 }
 
 /* search fpath for an undefined function */
@@ -1779,6 +1797,9 @@ int fd;
 int cancd(s)
 char *s;
 {
+char *t;
+
+	if ((t = getsparam(s)) && *t == '/') return 1;
 	if (*s != '/')
 		{
 		char sbuf[MAXPATHLEN],**cp;
