@@ -30,11 +30,11 @@ struct utmp;
 #include "zle_word.pro"
 
 char *mktemp DCLPROTO((char *));
+#ifndef HAS_STDLIB
 char *malloc DCLPROTO((int));
-#ifndef __STDC__
 char *realloc DCLPROTO((char *,int));
 char *calloc DCLPROTO((int,int));
-#endif /* __STDC__ */
+#endif
 char *ttyname DCLPROTO((int));
 
 extern char PC, *BC, *UP;
