@@ -1,34 +1,22 @@
 /*
-
-	glob.c - filename generation
-
-	This file is part of zsh, the Z shell.
-
-	zsh is free software; no one can prevent you from reading the source
-   code, or giving it to someone else.
-
-   This file is copyrighted under the GNU General Public License, which
-   can be found in the file called COPYING.
-
-   Copyright (C) 1990, 1991 Paul Falstad
-
-   zsh is distributed in the hope that it will be useful, but
-   WITHOUT ANY WARRANTY.  No author or distributor accepts
-   responsibility to anyone for the consequences of using it or for
-   whether it serves any particular purpose or works at all, unless he
-   says so in writing.  Refer to the GNU General Public License
-   for full details.
-
-   Everyone is granted permission to copy, modify and redistribute
-   zsh, but only under the conditions described in the GNU General Public
-   License.   A copy of this license is supposed to have been given to you
-   along with zsh so you can know your rights and responsibilities.
-   It should be in a file named COPYING.
-
-   Among other things, the copyright notice and this notice must be
-   preserved on all copies.
-
-*/
+ *
+ * glob.c - filename generation
+ *
+ * This file is part of zsh, the Z shell.
+ *
+ * This software is Copyright 1992 by Paul Falstad
+ *
+ * Permission is hereby granted to copy, reproduce, redistribute or otherwise
+ * use this software as long as: there is no monetary profit gained
+ * specifically from the use or reproduction of this software, it is not
+ * sold, rented, traded or otherwise marketed, and this copyright notice is
+ * included prominently in any copy made. 
+ *
+ * The author make no claims as to the fitness or correctness of this software
+ * for any use whatsoever, and it is provided as is. Any use of this software
+ * is at the user's own risk. 
+ *
+ */
 
 #include "zsh.h"
 
@@ -43,7 +31,7 @@
 #endif
 #include <sys/errno.h>
 
-#define exists(X) (access(X,0) == 0)
+#define exists(X) (access(X,0) == 0 || readlink(X,NULL,0) == 0)
 
 static int mode;				/* != 0 if we are parsing glob patterns */
 static int pathpos;			/* position in pathbuf */
@@ -52,6 +40,7 @@ static int matchct;			/* number of matches found */
 static char pathbuf[MAXPATHLEN];	/* pathname buffer */
 static char **matchbuf;		/* array of matches */
 static char **matchptr;		/* &matchbuf[matchct] */
+static Comp exclude;			/* pattern to exclude */
 
 /* max # of qualifiers */
 
@@ -61,6 +50,7 @@ static int (*qualfuncs[QUALCT])DCLPROTO((struct stat *,long));
 static long qualdata[QUALCT];
 static int qualsense[QUALCT];
 static int qualct;
+static int gf_nullglob,gf_markdirs,gf_noglobdots;
 
 /* pathname component in filename patterns */
 
@@ -89,6 +79,9 @@ char *str = getdata(*np);	/* the pattern */
 	ostr = strdup(str);
 	uremnode(list,*np);
 	qualct = 0;
+	gf_nullglob = isset(NULLGLOB);
+	gf_markdirs = isset(MARKDIRS);
+	gf_noglobdots = unset(GLOBDOTS);
 	if (str[sl-1] == Outpar)	/* check for qualifiers */
 		{
 		char *s;
@@ -113,9 +106,9 @@ char *str = getdata(*np);	/* the pattern */
 					while (idigit(*s))
 						data = data*010+(*s++-'0');
 					}
-				else switch (*s++)
+				else switch ((int)(unsigned char)(*s++))
 					{
-					case Hat: case '^': sense = 1-sense; break;
+					case (int)(unsigned char)Hat: case '^': sense = 1-sense; break;
 #ifdef S_IFLNK
 					case '@': func = qualmode; data = S_IFLNK; break;
 #endif
@@ -128,7 +121,7 @@ char *str = getdata(*np);	/* the pattern */
 					case '/': func = qualmode; data = S_IFDIR; break;
 					case '.': func = qualmode; data = S_IFREG; break;
 					case '%': func = qualisdev; break;
-					case Star:  func = qualiscom; break;
+					case (int)(unsigned char)Star:  func = qualiscom; break;
 					case 'R': func = qualflags; data = 0004; break;
 					case 'W': func = qualflags; data = 0002; break;
 					case 'X': func = qualflags; data = 0001; break;
@@ -143,6 +136,9 @@ char *str = getdata(*np);	/* the pattern */
 					case 'G': func = qualgid; data = getegid(); break;
 					case 'u': func = qualuid; data = qgetnum(&s); break;
 					case 'g': func = qualgid; data = qgetnum(&s); break;
+					case 'M': gf_markdirs = !sense; break;
+					case 'N': gf_nullglob = !sense; break;
+					case 'D': gf_noglobdots = sense; break;
 					default: zerr("unknown file attribute",NULL,0); return;
 					}
 				if (func)
@@ -162,7 +158,9 @@ char *str = getdata(*np);	/* the pattern */
 				}
 			}
 		}
-	else if (str[sl-1] == '/')		/* foo/ == foo(/) */
+	else if ((str[sl-1] == '/') && !((str[sl-2] == Star)&&
+				(str[sl-3] == Star)&&(str[sl-4] == Star)&&
+				(str[sl-5]==Star)))		/* foo/ == foo(/) */
 		{
 		str[sl-1] = '\0';
 		qualfuncs[0] = qualmode;
@@ -177,7 +175,7 @@ char *str = getdata(*np);	/* the pattern */
 		pathbuf[0] = '/';
 		pathbuf[pathpos = 1] = '\0';
 		}
-	else		/* pattern is relative to cwd */
+	else		/* pattern is relative to pwd */
 		pathbuf[pathpos = 0] = '\0';
 	q = parsepat(str);
 	if (!q || errflag)	/* if parsing failed */
@@ -196,7 +194,7 @@ char *str = getdata(*np);	/* the pattern */
 	scanner(q);		/* do the globbing */
 	if (matchct)
 		badcshglob |= 2;
-	else if (unset(NULLGLOB))
+	else if (!gf_nullglob)
 		if (isset(CSHNULLGLOB)) {
 			badcshglob |= 1;
 		} else if (unset(NONOMATCH)) {
@@ -240,7 +238,7 @@ int x1,x2;
 
 	for (; *c == *d && *c; c++,d++);
 	x1 = atoi(c); x2 = atoi(d);
-	if (x1==x2)
+	if (x1==x2 || unset(NUMERICGLOBSORT))
 		return ((int) (unsigned char) *c-(int) (unsigned char) *d);
 	return x1-x2;
 }
@@ -260,9 +258,10 @@ void insert(s) /**/
 char *s;
 {
 struct stat buf;
+int statted = 0;
 
-	if (isset(MARKDIRS) && !lstat(s,&buf) && S_ISDIR(buf.st_mode)) /* grrr */
-		{
+	if (exclude && domatch(s,exclude,gf_noglobdots)) return;
+	if (gf_markdirs && !lstat(s,&buf) && S_ISDIR(buf.st_mode)) {
 		char *t;
 		int ll = strlen(s);
 
@@ -271,14 +270,23 @@ struct stat buf;
 		t[ll] = '/';
 		t[ll+1] = '\0';
 		s = t;
-		}
+		statted = 1;
+	}
+	if (qualct)	{ /* do the (X) (^X) stuff */
+		int (**fptr)DCLPROTO((struct stat *,long)) = qualfuncs;
+		int *sptr = qualsense;
+		long *lptr = qualdata;
+		struct stat buf;
+
+		if (statted || lstat(s,&buf) >= 0)
+			while (*fptr) if (!(!!((*fptr++)(&buf,*lptr++)) ^ *sptr++)) return;
+	}
 	*matchptr++ = s;
-	if (++matchct == matchsz)
-		{
+	if (++matchct == matchsz) {
 		matchbuf = (char **) realloc((char *) matchbuf,
 			sizeof(char **)*(matchsz *= 2));
 		matchptr = matchbuf+matchct;
-		}
+	}
 }
 
 /* check to see if str is eligible for filename generation */
@@ -286,18 +294,12 @@ struct stat buf;
 int haswilds(str) /**/
 char *str;
 {
-	if (!str[1] && (*str == Inbrack || *str == Outbrack))
-		return 0;
-	if (str[0] == '%')
-		return 0;
+	if ((*str == Inbrack || *str == Outbrack) && !str[1]) return 0;
+	if (str[0] == '%') return 0;
 	for (; *str; str++)
-		if (str[0] == Star && str[1] == Star &&
-			 str[2] == Star && str[3] == Star && str[4] == '/')
-			return 1;
-		else if (*str == Pound || *str == Hat || *str == Star ||
+		if (*str == Pound || *str == Hat || *str == Star ||
 				*str == Bar || *str == Inbrack || *str == Inang ||
-				*str == Quest)
-			return 1;
+				*str == Quest) return 1;
 	return 0;
 }
 
@@ -311,6 +313,17 @@ char *lbr = NULL;
 
 	if (str[0] == Inbrace && str[1] == Outbrace)
 		return 0;
+	if (isset(BRACECCL)) {
+		for (mb = bc = 0; *str; ++str)
+			if (*str == Inbrace) {
+				if (++bc > mb)
+					mb = bc;
+			}
+			else if (*str == Outbrace)
+				if (--bc < 0)
+					return(0);
+		return(mb && bc == 0);
+	}
 	for (mb = bc = cmct1 = cmct2 = 0; *str; str++)
 		{
 		if (*str == Inbrace)
@@ -367,22 +380,18 @@ int ret = 0;
 	prefork(fake);
 	if (!errflag)
 		postfork(fake,1);
-	if (errflag)
-		return 0;
-	if (full(fake) && !nextnode(firstnode(fake)))
-		{
+	if (errflag) return 0;
+	if (full(fake) && !nextnode(firstnode(fake))) {
 		fn->name = peekfirst(fake);
 		untokenize(fn->name);
-		}
-	else
-		while (nam = ugetnode(fake))
-			{
+	} else
+		while (nam = ugetnode(fake)) {
 			ff = alloc(sizeof *ff);
 			*ff = *fn;
 			ff->name = nam;
 			addnode(tab,ff);
 			ret = 1;
-			}
+		}
 	return ret;
 }
 
@@ -420,9 +429,46 @@ Lklist list;Lknode *np;
 {
 Lknode node = (*np),last = prevnode(node);
 char *str = getdata(node),*str3 = str,*str2;
-int prev;
+int prev, bc, comma;
 
 	for (; *str != Inbrace; str++);
+	for (str2 = str, bc = comma = 0; *str2; ++str2)
+		if (*str2 == Inbrace)
+			++bc;
+		else if (*str2 == Outbrace) {
+			if (--bc == 0)
+				break;
+		}
+		else if (bc == 1 && *str2 == Comma)
+			++comma;
+	if (!comma && !bc && isset(BRACECCL)) {			/* {a-mnop} */
+		char	ccl[256], *p;
+		unsigned char c1,c2,lastch;
+
+		uremnode(list,node);
+		memset(ccl, 0, sizeof(ccl) / sizeof(ccl[0]));
+		for (p = str + 1, lastch = 0; p < str2; ) {
+			if (itok(c1 = *p++))
+				c1 = ztokens[c1 - (unsigned char)Pound];
+			if (itok(c2 = *p))
+				c2 = ztokens[c2 - (unsigned char)Pound];
+			if (c1 == '-' && lastch && p < str2 && lastch <= c2) {
+				while (lastch < c2)
+					ccl[lastch++] = 1;
+				lastch = 0;
+			}
+			else
+				ccl[lastch = c1] = 1;
+		}
+		strcpy(str + 1, str2 + 1);
+		for (p = ccl+255; p-- > ccl; )
+			if (*p) {
+				*str = p - ccl;
+				insnode(list, last, strdup(str3));
+			}
+		*np = nextnode(last);
+		return;
+	}
  	if (str[2] == '-' && str[4] == Outbrace)	 /* {a-z} */
 		{
 		char c1,c2;
@@ -675,7 +721,7 @@ int closure;
 				if (errflag)
 					break;
 				fn = &de->d_name[0];
-				if (domatch(fn,c,unset(GLOBDOTS)))
+				if (domatch(fn,c,gf_noglobdots))
 					{
 					int oppos = pathpos;
 
@@ -704,35 +750,7 @@ int closure;
 							scanner((q->closure) ? q : q->next); /* scan next level */
 						pathbuf[pathpos = oppos] = '\0';
 						}
-					else
-						{
-						if (qualct)	/* do the (X) (^X) stuff */
-							{
-							int (**fptr)DCLPROTO((struct stat *,long)) = qualfuncs;
-							int *sptr = qualsense;
-							long *lptr = qualdata;
-							struct stat buf;
-	
-							if (lstat(getfullpath(fn),&buf) == -1)
-								{
-								if (errno != ENOENT && errno != EINTR)
-									{
-									zerr("%e: %s",fn,errno);
-									errflag = 0;
-									}
-								continue;
-								}
-							while (*fptr)
-								if (!(!!((*fptr++)(&buf,*lptr++)) ^ *sptr++))
-									{
-									fptr = NULL;
-									break;
-									}
-							if (!fptr)
-								continue;
-							}
-						insert(dyncat(pathbuf,fn));
-						}
+					else insert(dyncat(pathbuf,fn));
 					}
 				}
 			closedir(lock);
@@ -778,16 +796,14 @@ Comp c;
 {
 char *pat = c->str;
 
-	if (c->closure == 1)
-		{
+	if (c->closure == 1) {
 		char *saves = pptr;
-		int savei = first;
 
-		if (doesmatch(c->next))
-			return 1;
+		if (first && *pptr == '.') return 0;
+		if (doesmatch(c->next)) return 1;
 		pptr = saves;
-		first = savei;
-		}
+		first = 0;
+	}
 	for(;;)
 		{
 		if (!pat || !*pat)
@@ -917,6 +933,20 @@ char *pat = c->str;
 Complist parsepat(str) /**/
 char *str;
 {
+char *s;
+
+	exclude = NULL;
+	if (isset(EXTENDEDGLOB)) {
+		s = str+strlen(str);
+		while (s-- > str) {
+			if (*s == Tilde && s[1]) {
+				*s++ = '\0';
+				exclude = parsereg(s);
+				if (!exclude) return NULL;
+				break;
+			}
+		}
+	}
 	mode = 0;
 	pptr = str;
 	return parsecomplist();
@@ -936,15 +966,19 @@ Comp c1;
 Complist p1;
 
 	if (pptr[0] == Star && pptr[1] == Star &&
-		 pptr[2] == Star && pptr[3] == Star &&
-		 pptr[4] == '/')
-		{
-		pptr[0] = Inpar;
-		pptr[1] = Star;
-		pptr[2] = '/';
-		pptr[3] = Outpar;
-		pptr[4] = Pound;	/* "####/" -> "( * /)#" */
-		}
+			(pptr[2] == '/' ||
+			(pptr[2] == Star && pptr[3] == Star && pptr[4] == '/'))) {
+		pptr += 3;
+		if (pptr[-1] == Star) pptr += 2;
+		p1 = (Complist) alloc(sizeof *p1);
+		p1->next = parsecomplist();
+		p1->comp = (Comp) alloc(sizeof *p1->comp);
+		p1->comp->last = 1;
+		p1->comp->str = strdup("*");
+		*p1->comp->str = Star;
+		p1->closure = 1;
+		return p1;
+	}
 	if (*pptr == Inpar)
 		{
 		char *str;
