@@ -1,37 +1,36 @@
 /*
-
-	hist.c - history expansion
-
-	This file is part of zsh, the Z shell.
-
-	zsh is free software; no one can prevent you from reading the source
-   code, or giving it to someone else.
-
-   This file is copyrighted under the GNU General Public License, which
-   can be found in the file called COPYING.
-
-   Copyright (C) 1990, 1991 Paul Falstad
-
-   zsh is distributed in the hope that it will be useful, but
-   WITHOUT ANY WARRANTY.  No author or distributor accepts
-   responsibility to anyone for the consequences of using it or for
-   whether it serves any particular purpose or works at all, unless he
-   says so in writing.  Refer to the GNU General Public License
-   for full details.
-
-   Everyone is granted permission to copy, modify and redistribute
-   zsh, but only under the conditions described in the GNU General Public
-   License.   A copy of this license is supposed to have been given to you
-   along with zsh so you can know your rights and responsibilities.
-   It should be in a file named COPYING.
-
-   Among other things, the copyright notice and this notice must be
-   preserved on all copies.
-
-*/
+ *
+ * hist.c - history expansion
+ *
+ * This file is part of zsh, the Z shell.
+ *
+ * This software is Copyright 1992 by Paul Falstad
+ *
+ * Permission is hereby granted to copy, reproduce, redistribute or otherwise
+ * use this software as long as: there is no monetary profit gained
+ * specifically from the use or reproduction of this software, it is not
+ * sold, rented, traded or otherwise marketed, and this copyright notice is
+ * included prominently in any copy made. 
+ *
+ * The author make no claims as to the fitness or correctness of this software
+ * for any use whatsoever, and it is provided as is. Any use of this software
+ * is at the user's own risk. 
+ *
+ */
 
 #include "zsh.h"
 
+#define HEAPSIZE 4096
+
+struct hp {
+	Hp next;
+	char *pool,*ptr;
+	int free,histno;
+};
+
+static Hp hp_lit, hp_lex;
+static Histent curhistent;
+
 static int lastc;
 
 /* add a character to the current history word */
@@ -39,23 +38,20 @@ static int lastc;
 void hwaddc(c) /**/
 int c;
 {
-	if (hlastw && hline && (!(errflag || lexstop) || c == HISTSPACE))
-		{
+	if (hlastw && hline && (!(errflag || lexstop) || c == HISTSPACE)) {
+		if (c == '!' && unset(NOBANGHIST)) hwaddc('\\');
 		*hptr++ = c;
-		if (hptr-hline >= hlinesz)
-			{
-			int ll,flag = 0;
+		if (hptr-hline >= hlinesz) {
+			int ll,flag = 0,oldsiz = hlinesz;
 
 			ll = hptr-hlastw;
-			if (full(histlist) && getdata(lastnode(histlist)) == hline)
-				flag = 1;
-			hline = realloc(hline,hlinesz *= 4);
-			if (flag)
-				setdata(lastnode(histlist),hline);
-			hptr = hline+(hlinesz/4);
+			if (curhistent->lex == hline) flag = 1;
+			hline = hp_realloc(&hp_lex,hline,oldsiz,hlinesz = oldsiz+16);
+			if (flag) curhistent->lex = hline;
+			hptr = hline+oldsiz;
 			hlastw = hptr-ll;
-			}
 		}
+	}
 }
 
 #define habort() { errflag = lexstop = 1; return ' '; }
@@ -85,18 +81,16 @@ tailrec:
 		}
 	if (c != ' ')
 		isfirstch = 0;
-	if (c == '\\')
-		{
+	if (c == '\\') {
 		int g = hgetch();
 		
 		if (g != bangchar)
 			hungetch(g);
-		else
-			{
+		else {
 			hwaddc(bangchar);
 			return bangchar;
-			}
 		}
+	}
 	if (c != bangchar)
 		{
 		hwaddc(c);
@@ -160,7 +154,7 @@ hatskip:
 				break;
 			if (ptr != buf) {
 				if (c == '-') break;
-				if (idigit(buf[0]) && !idigit(c)) break;
+				if ((idigit(buf[0]) || buf[0] == '-') && !idigit(c)) break;
 			}
 			*ptr++ = c;
 			if (c == '#' || c == bangchar)
@@ -362,12 +356,12 @@ hatskip:
 		hungets(HISTMARK);
 		alstack[alstackind++] = NULL;
 		}
+	for (ptr = sline; *ptr; ptr++) {
+		if (ptr[0] == '\\' && ptr[1] == '!') chuck(ptr);
+	}
 	hungets(sline);
 	histdone |= HISTFLAG_DONE;
-	if (isset(HISTVERIFY)) {
-		histdone |= HISTFLAG_NOEXEC;
-		zrecall = 1;
-	}
+	if (isset(HISTVERIFY)) histdone |= HISTFLAG_NOEXEC|HISTFLAG_RECALL;
 	goto tailrec;
 }
 
@@ -388,7 +382,7 @@ Alias ix;
 
 int hgetch() /**/
 {
-char *line,*pmpt,*pmpt2 = NULL;
+unsigned char *line,*pmpt,*pmpt2 = NULL;
 int plen;
 
 start:
@@ -429,66 +423,49 @@ start:
 		}
 	if (interact && isset(SHINSTDIN))
 		if (!isfirstln)
-			pmpt = putprompt(prompt2,&plen);
+			pmpt = (unsigned char *)putprompt(prompt2,&plen);
 		else
 			{
 			int foo;
 
-			pmpt = putprompt(prompt,&plen);
-			pmpt2 = (rprompt) ? putprompt(rprompt,&foo) : NULL;
+			pmpt = (unsigned char *)putprompt(prompt,&plen);
+			pmpt2 = (unsigned char *)((rprompt) ? putprompt(rprompt,&foo) : NULL);
 			}
-	if (!(interact && isset(SHINSTDIN) && SHTTY != -1 && isset(USEZLE)))
-		{
+	if (!(interact && isset(SHINSTDIN) && SHTTY != -1 && isset(USEZLE))) {
+		char *lbuf;
 		if (interact && isset(SHINSTDIN))
-			write(2,pmpt,strlen(pmpt));
-		line = fgets(zalloc(256),256,bshin);
-		}
-	else
+			write(2,pmpt,strlen((char *) pmpt));
+		line = (unsigned char *)fgets(lbuf = zalloc(256),256,bshin);
+		if (!line) free(lbuf);
+	} else
 		line = zleread(pmpt,pmpt2,plen);
-	if (!line)
-		{
+	if (!line) {
 		lexstop = 1;
 		return lastc = ' ';
-		}
-	if (errflag)
-		{
+	}
+	if (errflag) {
+		free(line);
 		lexstop = errflag = 1;
 		return lastc = ' ';
-		}
-	if (interact && isset(SHINSTDIN))
-		{
-		char *s = getdata(lastnode(lithistlist));
-
-		if (!*s)
-			{
-			free(s);
-			setdata(lastnode(lithistlist),ztrdup(line));
-			}
-		else
-			{
-			char *t = zalloc(strlen(s)+strlen(line)+3);
-
-			strcpy(t,s);
-			strcat(t,line);
-			free(s);
-			setdata(lastnode(lithistlist),t);
-			}
-		}
-	if (isfirstln)
-		spaceflag = *line == ' ';
-	if (isset(VERBOSE))
-		{
-		fputs(line,stderr);
+	}
+	if (interact && isset(SHINSTDIN)) {
+		char *s = curhistent->lit;
+		curhistent->lit = hp_concat(s,(char*)line);
+	}
+	if (isfirstln) spaceflag = *line == ' ';
+	if (isset(VERBOSE)) {
+		fputs((char *) line,stderr);
 		fflush(stderr);
-		}
-	if (line[strlen(line)-1] == '\n')
+	}
+	if (*line && line[strlen((char *) line)-1] == '\n')
 		{
 		lineno++;
 		if (interact && isset(SUNKEYBOARDHACK) && isset(SHINSTDIN) && 
-				SHTTY != -1 && *line && line[1] && line[strlen(line)-2] == '`')
+				SHTTY != -1 && *line && line[1] &&
+				line[strlen((char *) line)-2] == '`')
 			{
 			int ct;
-			char *ptr;
+			unsigned char *ptr;
 
 			for (ct = 0, ptr = line; *ptr; ptr++)
 				if (*ptr == '`')
@@ -501,11 +478,26 @@ start:
 			}
 		}
 	isfirstch = 1;
-	hungets(line);
+	hungets((char*)line);
 	free(line);
 	goto start;
 }
 
+/* Read one line of at most n-1 chars from the input queue */
+
+char *hgets(buf, n) /**/
+char *buf;int n;
+{
+int l;
+
+	for (l = 0; l < n-1; l++)
+		if ((buf[l] = hgetch()) == '\n' || lexstop)
+			break;
+	buf[l+(lexstop?0:1)] = 0;
+
+	return (!lexstop || l) ? buf : NULL;
+}
+
 /* put a string in the input queue */
 
 void hungets(str) /**/
@@ -545,13 +537,14 @@ int c;
 {
 	if (lexstop)
 		return;
-	if (hlastw)
-		{
+	if (hlastw) {
 		if (hlastw == hptr)
 			zerr("hungetc attempted at buffer start",NULL,0);
-		else
+		else {
 			hptr--;
+			if (*hptr == '!' && unset(NOBANGHIST)) hptr--;
 		}
+	}
 	hungetch(c);
 }
 
@@ -637,24 +630,21 @@ void hbegin() /**/
 {
 	isfirstln = isfirstch = 1;
 	histremmed = errflag = histdone = spaceflag = 0;
-	stophist = isset(NOBANGHIST);
+	stophist = isset(NOBANGHIST) || unset(SHINSTDIN);
 	lithist = isset(HISTLIT);
-	hline = hptr = zalloc(hlinesz = 32);
+	hline = hptr = hp_alloc(&hp_lex,hlinesz = 16);
+	curhistent = gethistent(curhist);
+	if (!curhistent->ftim) curhistent->ftim = time(NULL);
 	if (interact && isset(SHINSTDIN) && !strin) {
 		inittty();
 		defev = curhist++;
-		while (curhist-firsthist >= histsiz) {
-			free(getnode(histlist));
-			firsthist++;
-		}
-		while (curhist-firstlithist >= lithistsiz) {
-			free(getnode(lithistlist));
-			firstlithist++;
-		}
-		permalloc();
-		addnode(histlist,hline);
-		addnode(lithistlist,ztrdup(""));
-		heapalloc();
+		if (curhist-histsiz >= 0) gethistent(curhist-histsiz)->lex = NULL;
+		if (curhist-lithistsiz >= 0) gethistent(curhist-lithistsiz)->lit = NULL;
+		curhistent = gethistent(curhist);
+		hp_purge(hp_lex,curhist-histsiz);
+		hp_purge(hp_lit,curhist-lithistsiz);
+		curhistent->lex = hline;
+		*(curhistent->lit = hp_alloc(&hp_lit,1)) = '\0';
 	} else
 		histremmed = 1;
 }
@@ -669,11 +659,12 @@ void inittty() /**/
 int hend() /**/
 {
 int flag,save = 1;
+Histent he;
 
 	if (!hline)
 		return 1;
 	if (!interact || strin || unset(SHINSTDIN)) {
-		free(hline);
+		hp_free(hp_lex,hline,hlinesz);
 		return 1;
 	}
 	flag = histdone;
@@ -681,42 +672,40 @@ int flag,save = 1;
 	if (hptr < hline+2)
 		save = 0;
 	else {
-		char *s = getdata(lastnode(lithistlist));
+		char *s,*t;
 
-		if (*s)
-			s[strlen(s)-1] = '\0';
+		s = curhistent->lit;
+		if (*s && *(t = s+strlen(s)-1) == HISTSPACE) *t = '\0';
 		hptr[-1] = '\0';
 		if (hptr[-2] == '\n')
-			if (hline[1])
-				hptr[-3] = '\0';
-			else
-				save = 0;
+			if (hline[1]) {
+				if (hptr[-3] == HISTSPACE) hptr[-3] = '\0';
+			} else save = 0;
+		he = gethistent(curhist-1);
 		if (!strcmp(hline,"\n") ||
-				(isset(HISTIGNOREDUPS) && firstnode(histlist) &&
-				nextnode(firstnode(histlist)) &&
-				!strcmp(hline,getdata(prevnode(lastnode(histlist))))) ||
-				(isset(HISTIGNORESPACE) && spaceflag) )
+				(isset(HISTIGNOREDUPS) && he->lex && !strcmp(he->lex,hline)) ||
+				(isset(HISTIGNORESPACE) && spaceflag))
 			save = 0;
 	}
-	if (flag & HISTFLAG_DONE) {
+	if (flag & (HISTFLAG_DONE|HISTFLAG_RECALL)) {
 		char *ptr,*p;
-
 		p = ptr = ztrdup(hline);
-		for (;*p;p++)
-			if (*p == HISTSPACE)
-				*p = ' ';
-		fprintf(stderr,"%s\n",ptr);
-		fflush(stderr);
-		free(ptr);
-	}
-	if (!save) {
-		free(hline);
-		if (!histremmed) {
-			remnode(histlist,lastnode(histlist));
-			free(remnode(lithistlist,lastnode(lithistlist)));
-			curhist--;
+		for (;*p;p++) if (*p == HISTSPACE) *p = ' ';
+		if ((flag & (HISTFLAG_DONE|HISTFLAG_RECALL)) == HISTFLAG_DONE) {
+			fprintf(stderr,"%s\n",ptr);
+			fflush(stderr);
 		}
+		if (flag & HISTFLAG_RECALL) {
+			permalloc();
+			pushnode(bufstack,ptr);
+			lastalloc();
+			save = 0;
+		} else free(ptr);
 	}
+	curhistent->stim = time(NULL);
+	curhistent->ftim = 0L;
+	if (!save) remhist();
+	if (hline && !curhistent->lex) hp_free(hp_lex,hline,hlinesz);
 	hline = NULL;
 	return !(flag & HISTFLAG_NOEXEC || errflag);
 }
@@ -725,13 +714,7 @@ int flag,save = 1;
 
 void remhist() /**/
 {
-	if (!histremmed)
-		{
-		histremmed = 1;
-		free(remnode(histlist,lastnode(histlist)));
-		free(remnode(lithistlist,lastnode(lithistlist)));
-		curhist--;
-		}
+	if (!histremmed) { histremmed = 1; curhist--; }
 }
 
 /* begin a word */
@@ -803,21 +786,14 @@ int hconsearch(str,marg) /**/
 char *str;int *marg;
 {
 int t0,t1 = 0;
-Lknode node;
-char *s;
- 
-	if (curhist-firsthist < 1)
-		return -1;
-	for (t0 = curhist-1,node = prevnode(lastnode(histlist));
-			t0 >= firsthist; t0--,node = prevnode(node))
-		if (s = ztrstr(getdata(node),str))
-			{
-			while (s != (char *) getdata(node))
-				if (*s-- == HISTSPACE)
-					t1++;
+char *s,*hs;
+
+	for (t0 = curhist-1; hs = quietgetevent(t0); t0--)
+		if (s = ztrstr(hs,str)) {
+			while (s != hs) if (*s-- == HISTSPACE) t1++;
 			*marg = t1;
 			return t0;
-			}
+		}
 	return -1;
 }
 
@@ -827,14 +803,10 @@ int hcomsearch(str) /**/
 char *str;
 {
 int t0;
-Lknode node;
+char *hs;
 
-	if (curhist-firsthist < 1)
-		return -1;
-	for (t0 = curhist-1,node = prevnode(lastnode(histlist)); t0 >= firsthist;
-			t0--,node = prevnode(node))
-		if (!strncmp(getdata(node),str,strlen(str)))
-			return t0;
+	for (t0 = curhist-1; hs = quietgetevent(t0); t0--)
+		if (!strncmp(hs,str,strlen(str))) return t0;
 	return -1;
 }
 
@@ -921,11 +893,12 @@ int off;
 
 	while (cut = (char *) ztrstr(str,in)) {
 		*cut = '\0';
+		sptr = convamps(out,in);
+		off = cut-*strptr+strlen(sptr);
 		cut += strlen(in);
-		off = cut-*strptr;
-		*strptr = tricat(*strptr,sptr = convamps(out,in),cut);
+		*strptr = tricat(*strptr,sptr,cut);
 		if (gbal) {
-			str = (char *) *strptr+off+strlen(sptr);
+			str = (char *) *strptr+off;
 			continue;
 		}
 		break;
@@ -973,33 +946,24 @@ char *t;
 char *quietgetevent(ev) /**/
 int ev;
 {
-Lknode node;
- 
-	ev -= (lithist) ? firstlithist : firsthist;
-	if (ev < 0)
-		return NULL;
-	for (node = firstnode((lithist) ? lithistlist : histlist);
-			ev && node; incnode(node), ev--);
-	if (!node)
-		return NULL;
-	return getdata(node);
+Histent ent;
+
+	if (ev < firsthist()) return NULL;
+	ent = gethistent(ev);
+	return (lithist) ? ent->lit : ent->lex;
 }
 
 char *getevent(ev) /**/
 int ev;
 {
-Lknode node;
-int oev = ev;
- 
-	ev -= firsthist;
-	for (node = firstnode(histlist); ev && node; incnode(node), ev--);
-	if (!node)
-		{
+char *ret;
+
+	ret = quietgetevent(ev);
+	if (!ret) {
 		herrflush();
-		zerr("no such event: %d",NULL,oev);
-		return NULL;
-		}
-	return getdata(node);
+		zerr("no such event: %d",NULL,ev);
+	}
+	return ret;
 }
  
 int getargc(list) /**/
@@ -1007,9 +971,7 @@ char *list;
 {
 int argc = 0;
 
-	for (; *list; list++)
-		if (*list == HISTSPACE)
-			argc++;
+	for (; *list; list++) if (*list == HISTSPACE) argc++;
 	return argc;
 }
  
@@ -1070,21 +1032,21 @@ int quote(tr) /**/
 char **tr;
 {
 char *ptr,*rptr,**str = (char **) tr;
-int len = 1;
+int len = 3;
  
 	for (ptr = *str; *ptr; ptr++,len++)
-		if (*ptr == '\'')
-			len += 3;
+		if (*ptr == '\'') len += 3;
 	ptr = *str;
-	*str = rptr = zalloc(len);
-	for (ptr = *str; *ptr; )
-		if (*ptr == '\'')
-			{
+	*str = rptr = alloc(len);
+	*rptr++ = '\'';
+	for (; *ptr; ptr++)
+		if (*ptr == '\'') {
 			*rptr++ = '\''; *rptr++ = '\\'; *rptr++ = '\''; *rptr++ = '\'';
-			ptr++;
-			}
-		else
-			*rptr++ = *ptr++;
+		} else
+			*rptr++ = *ptr;
+	*rptr++ = '\'';
+	*rptr++ = 0;
+	str[1] = NULL;
 	return 0;
 }
  
@@ -1092,7 +1054,7 @@ int quotebreak(tr) /**/
 char **tr;
 {
 char *ptr,*rptr,**str = (char **) tr;
-int len = 1;
+int len = 3;
  
 	for (ptr = *str; *ptr; ptr++,len++)
 		if (*ptr == '\'')
@@ -1100,229 +1062,21 @@ int len = 1;
 		else if (inblank(*ptr))
 			len += 2;
 	ptr = *str;
-	*str = rptr = zalloc(len);
-	for (ptr = *str; *ptr; )
-		if (*ptr == '\'')
-			{
+	*str = rptr = alloc(len);
+	*rptr++ = '\'';
+	for (; *ptr; )
+		if (*ptr == '\'') {
 			*rptr++ = '\''; *rptr++ = '\\'; *rptr++ = '\''; *rptr++ = '\'';
 			ptr++;
-			}
-		else if (inblank(*ptr))
-			{
+		} else if (inblank(*ptr)) {
 			*rptr++ = '\''; *rptr++ = *ptr++; *rptr++ = '\'';
-			}
-		else
+		} else
 			*rptr++ = *ptr++;
+	*rptr++ = '\'';
+	*rptr++ = '\0';
 	return 0;
 }
 
-static char *bp;
-static int lensb,countp;
-
-void stradd(d) /**/
-char *d;
-{
-	while (*bp++ = *d++);
-	bp--;
-}
-
-int putstr(d) /**/
-int d;
-{
-	*bp++ = d;
-	if (countp)
-		lensb++;
-	return 0;
-}
-
-#define tstradd(X) \
-	if (termok && unset(SINGLELINEZLE)) { \
-		char tbuf[2048],*tptr = tbuf; \
-		if (tgetstr(X,&tptr)) \
-			tputs(tbuf,1,putstr); \
-	} \
-	break
-
-/* get a prompt string */
-
-char *putprompt(fm,lenp) /**/
-char *fm;int *lenp;
-{
-char *ss,*ttyname DCLPROTO((int)),*bl0;
-static char buf1[256],buf2[256],*buf;
-char buf3[MAXPATHLEN];
-int t0,bracepos = 0;
-struct tm *tm = NULL;
-time_t timet;
-
-	lensb = 0; countp = 1;
-	if (!fm)
-		{
-		*lenp = 0;
-		return "";
-		}
-	/* kludge alert! */
-	buf = (buf == buf1) ? buf2 : buf1;
-	bp = bl0 = buf;
-	if (!columns)
-		columns = 80;
-	clearerr(stdin);
-	for(;*fm;fm++)
-		{
-		if (bp-buf >= 220)
-			break;
-		if (*fm == '%')
-			switch (*++fm)
-				{
-				case '~':
-					t0 = finddir(cwd);
-					if (t0 != -1) {
-						*bp++ = '~';
-						stradd(usernames[t0]);
-						stradd(cwd+strlen(userdirs[t0]));
-						break;
-					}
-					if (!strncmp(cwd,home,t0 = strlen(home)) && t0 > 1) {
-						*bp++ = '~';
-						stradd(cwd+t0);
-						break;
-					}
-				case 'd': case '/': stradd(cwd); break;
-				case 'c': case '.':
-					t0 = finddir(cwd);
-					if (t0 != -1) {
-						sprintf(buf3,"~%s%s",usernames[t0],
-							cwd+strlen(userdirs[t0]));
-					} else if (!strncmp(cwd,home,t0 = strlen(home)) && t0 > 1) {
-						sprintf(buf3,"~%s",cwd+t0);
-					} else {
-						strcpy(buf3,cwd);
-					}
-					t0 = 1;
-					if (idigit(fm[1])) { t0 = fm[1]-'0'; fm++; }
-					for (ss = buf3+strlen(buf3); ss > buf3; ss--)
-						if (*ss == '/' && !--t0) {
-							ss++;
-							break;
-						}
-					if (*ss == '/' && ss[1]) ss++;
-					stradd(ss);
-					break;
-				case 'C':
-					strcpy(buf3,cwd);
-					t0 = 1;
-					if (idigit(fm[1])) { t0 = fm[1]-'0'; fm++; }
-					for (ss = buf3+strlen(buf3); ss > buf3; ss--)
-						if (*ss == '/' && !--t0) {
-							ss++;
-							break;
-						}
-					if (*ss == '/' && ss[1]) ss++;
-					stradd(ss);
-					break;
-				case 'h': case '!':
-					sprintf(bp,"%d",curhist);
-					bp += strlen(bp);
-					break;
-				case 'M': stradd(hostnam); break;
-				case 'm':
-					if (idigit(fm[1]))
-						t0 = (*++fm)-'0';
-					else
-						t0 = 1;
-					for (ss = hostnam; *ss; ss++)
-						if (*ss == '.' && !--t0)
-							break;
-					t0 = *ss;
-					*ss = '\0';
-					stradd(hostnam);
-					*ss = t0;
-					break;
-				case 'S': tstradd("so"); /* <- this is a macro */
-				case 's': tstradd("se");
-				case 'B': tstradd("md");
-				case 'b': tstradd("me");
-				case 'U': tstradd("us");
-				case 'u': tstradd("ue");
-				case '{': bracepos = bp-buf; countp = 0; break;
-				case '}': lensb += (bp-buf)-bracepos; countp = 1; break;
-				case 't': case '@':
-					timet = time(NULL);
-					tm = localtime(&timet);
-					ztrftime(bp,16,"%l:%M%p",tm);
-					if (*bp == ' ')
-						chuck(bp);
-					bp += strlen(bp);
-					break;
-				case 'T':
-					timet = time(NULL);
-					tm = localtime(&timet);
-					ztrftime(bp,16,"%k:%M",tm);
-					bp += strlen(bp);
-					break;
-				case '*':
-					timet = time(NULL);
-					tm = localtime(&timet);
-					ztrftime(bp,16,"%k:%M:%S",tm);
-					bp += strlen(bp);
-					break;
-				case 'n': stradd(username); break;
-				case 'w':
-					timet = time(NULL);
-					tm = localtime(&timet);
-					ztrftime(bp,16,"%a %e",tm);
-					bp += strlen(bp);
-					break;
-				case 'W':
-					timet = time(NULL);
-					tm = localtime(&timet);
-					ztrftime(bp,16,"%m/%d/%y",tm);
-					bp += strlen(bp);
-					break;
-				case 'D':
-					timet = time(NULL);
-					tm = localtime(&timet);
-					ztrftime(bp,16,"%y-%m-%d",tm);
-					bp += strlen(bp);
-					break;
-				case 'l':
-					if (ss = ttyname(SHTTY))
-						stradd((strncmp(ss,"/dev/tty",8) ? ss : ss+8));
-					else
-						stradd("()");
-					break;
-				case '?':
-					sprintf(bp,"%d",lastval);
-					bp += strlen(bp);
-					break;
-				case '%': *bp++ = '%'; break;
-				case '#': *bp++ = (geteuid()) ? '%' : '#'; break;
-				default: *bp++ = '%'; *bp++ = *fm; break;
-				}
-		else if (*fm == '!')
-			{
-			sprintf(bp,"%d",curhist);
-			bp += strlen(bp);
-			}
-		else
-			{
-			if (fm[0] == '\\' && fm[1])
-				fm++;
-			if ((*bp++ = *fm) == '\n')
-				bl0 = bp;
-			}
-		}
-	*lenp = (bp-bl0)-lensb;
-	*lenp %= columns;
-	if (*lenp == columns-1)
-		{
-		*lenp = 0;
-		*bp++ = ' ';
-		}
-	*bp = '\0';
-	return buf;
-}
-
 void herrflush() /**/
 {
 	if (strin)
@@ -1391,66 +1145,199 @@ char *buf = zalloc(bsiz),*ptr;
 	return buf;
 }
 
-/* set cbreak mode, or the equivalent */
-
-void setcbreak() /**/
+void inithist() /**/
 {
-struct ttyinfo ti;
-
-	ti = shttyinfo;
-#ifdef TERMIOS
-	ti.termios.c_lflag &= ~ICANON;
-	ti.termios.c_cc[VMIN] = 1;
-	ti.termios.c_cc[VTIME] = 0;
-#else
-#ifdef TERMIO
-	ti.termio.c_lflag &= ~ICANON;
-	ti.termio.c_cc[VMIN] = 1;
-	ti.termio.c_cc[VTIME] = 0;
-#else
-	ti.sgttyb.sg_flags |= CBREAK;
-#endif
-#endif
-	settyinfo(&ti);
+	hp_lit = zalloc(sizeof *hp_lit);
+	hp_lit->next = NULL;
+	hp_lit->ptr = hp_lit->pool = zalloc(HEAPSIZE);
+	hp_lit->free = HEAPSIZE;
+	hp_lex = zalloc(sizeof *hp_lex);
+	hp_lex->next = NULL;
+	hp_lex->ptr = hp_lex->pool = zalloc(HEAPSIZE);
+	hp_lex->free = HEAPSIZE;
+	histentct = (lithistsiz > histsiz) ? lithistsiz : histsiz;
+	histentarr = zcalloc(histentct*sizeof *histentarr);
 }
 
-int getlineleng() /**/
+void resizehistents() /**/
 {
-int z;
+int newentct,t0,t1,firstlit,firstlex;
+Histent newarr;
 
-#ifdef TIOCSWINSZ
-	z = shttyinfo.winsize.ws_col;
-	return (z) ? z : 80;
-#else
-	return 80;
-#endif
+	newentct = (lithistsiz > histsiz) ? lithistsiz : histsiz;
+	newarr = zcalloc(newentct*sizeof *newarr);
+	firstlex = curhist-histsiz+1;
+	firstlit = curhist-lithistsiz+1;
+	t0 = firsthist();
+	if (t0 < curhist-newentct) t0 = curhist-newentct;
+	t1 = t0 % newentct;
+	for (; t0 <= curhist; t0++) {
+		newarr[t1] = *gethistent(t0);
+		if (t0 < firstlex) newarr[t1].lex = NULL;
+		if (t0 < firstlit) newarr[t1].lit = NULL;
+		t1++; if (t1 == newentct) t1 = 0;
+	}
+	free(histentarr);
+	histentarr = newarr;
+	histentct = newentct;
 }
 
-void unsetcbreak() /**/
+char *hp_alloc(hp,siz) /**/
+Hp *hp; int siz;
 {
-	settyinfo(&shttyinfo);
+char *ret;
+Hp h = *hp;
+
+	if (h->free >= siz) {
+		ret = h->ptr;
+		h->ptr += siz;
+		h->free -= siz;
+		return ret;
+	}
+#ifdef MEMDEBUG
+	fprintf(stderr,"new heap (siz = %d, curhist = %d)\n",siz,curhist);
+#endif
+	permalloc();
+	h = zalloc(sizeof *h);
+	h->next = *hp;
+	h->free = (siz > HEAPSIZE) ? siz : HEAPSIZE;
+	h->ptr = h->pool = zalloc(h->free);
+	h->histno = curhist;
+	*hp = h;
+	heapalloc();
+	return hp_alloc(hp,siz);
 }
 
-/* give the tty to some process */
-
-void attachtty(pgrp) /**/
-long pgrp;
+char *hp_realloc(hp,ptr,oldsiz,newsiz) /**/
+Hp *hp; char *ptr; int oldsiz; int newsiz;
 {
-static int ep = 0;
-int arg = pgrp;
+Hp h = *hp;
+int delta = newsiz-oldsiz;
+char *ret;
 
-	if (jobbing)
-#ifndef TIOCSPGRP
-		if (SHTTY != -1 && tcsetpgrp(SHTTY,pgrp) == -1 && !ep)
-#else
-		if (SHTTY != -1 && ioctl(SHTTY,TIOCSPGRP,&arg) == -1 && !ep)
+	if (h->ptr-oldsiz == ptr && h->free >= delta) {
+		h->free -= delta;
+		h->ptr += delta;
+		return ptr;
+	}
+#ifdef MEMDEBUG
+	fprintf(stderr,"realloc copy\n");
 #endif
-			{
-			zerr("can't set tty pgrp: %e",NULL,errno);
-			fflush(stderr);
-			opts[MONITOR] = OPT_UNSET;
-			ep =1;
-			errflag = 0;
+	memcpy(ret = hp_alloc(hp,newsiz),ptr,oldsiz);
+	return ret;
+}
+
+void hp_free(h,ptr,siz) /**/
+Hp h; char *ptr; int siz;
+{
+	if (h->ptr-siz == ptr) {
+		h->free += siz;
+		h->ptr -= siz;
+	}
+}
+
+char *hp_concat(old,new) /**/
+char *old; char *new;
+{
+int oldlen,newlen;
+
+	oldlen = strlen(old); newlen = strlen(new);
+	old = hp_realloc(&hp_lit,old,oldlen+1,oldlen+newlen+1);
+	strcpy(old+oldlen,new);
+	return old;
+}
+
+void hp_purge(h,lim) /**/
+Hp h; int lim;
+{
+Hp hlast;
+
+	if (!h->next) return;
+	while (h->next) { hlast = h; h = h->next; }
+	if (h->histno <= lim || h->histno == 0) {
+#ifdef MEMDEBUG
+		fprintf(stderr,"purging %d\n",lim);
+#endif
+		free(h->pool);
+		free(h);
+		hlast->next = NULL;
+	}
+}
+
+void readhistfile(s,err) /**/
+char *s;int err;
+{
+char buf[1024];
+FILE *in;
+Histent ent;
+time_t tim = time(NULL);
+
+	if (!s) return;
+	if (in = fopen(s,"r")) {
+		while (fgets(buf,1024,in)) {
+			int l = strlen(buf);
+			char *pt = buf;
+
+			while (l && buf[l-1] == '\n') {
+				buf[l-1] = '\0';
+				if (l > 1 && buf[l-2] == '\\') {
+					buf[l-2] = '\n';
+					fgets(buf+l-1,1024-(l-1),in);
+					l = strlen(buf);
+				} else break;
 			}
+			for (;*pt;pt++) if (*pt == ' ') *pt = HISTSPACE;
+
+			ent = gethistent(++curhist);
+			ent->lex = hp_alloc(&hp_lex,strlen(buf)+1);
+			strcpy(ent->lex,buf);
+			ent->lit = hp_alloc(&hp_lit,strlen(buf)+1);
+			strcpy(ent->lit,buf);
+			ent->ftim = ent->stim = tim;
+		}
+		fclose(in);
+	} else if (err)
+		zerr("can't read history file",s,0);
+}
+
+void savehistfile(s,err,app) /**/
+char *s;int err;int app;
+{
+char *t;
+FILE *out;
+int ev,flag;
+
+	if (!s || !interact) return;
+	ev = curhist-savehist+1;
+	flag = (app) ? O_APPEND : O_TRUNC;
+	if (ev < firsthist()) ev = firsthist();
+	if (out = fdopen(open(s,O_CREAT|O_WRONLY|flag,0600),"w")) {
+		for (; ev <= curhist; ev++) {
+			t = quietgetevent(ev);
+			for (; *t; t++)
+				if (*t == HISTSPACE) fputc(' ',out);
+				else {
+					if (*t == '\n') fputc('\\',out);
+					fputc(*t,out);
+				}
+			fputc('\n',out);
+		}
+		fclose(out);
+	} else if (err) zerr("can't write history file: %s",s,0);
+}
+
+int firsthist() /**/
+{
+int ev;
+Histent ent;
+
+	ev = curhist-histentct+1;
+	if (ev < 1) ev = 1;
+	do {
+		ent = gethistent(ev);
+		if ((lithist) ? ent->lit : ent->lex) break;
+		ev++;
+	} while (ev < curhist);
+	return ev;
 }
 
