@@ -1,48 +1,36 @@
 /*
-
-	init.c - main loop and initialization routines
-
-	This file is part of zsh, the Z shell.
-
-	zsh is free software; no one can prevent you from reading the source
-   code, or giving it to someone else.
-
-   This file is copyrighted under the GNU General Public License, which
-   can be found in the file called COPYING.
-
-   Copyright (C) 1990, 1991 Paul Falstad
-
-   zsh is distributed in the hope that it will be useful, but
-   WITHOUT ANY WARRANTY.  No author or distributor accepts
-   responsibility to anyone for the consequences of using it or for
-   whether it serves any particular purpose or works at all, unless he
-   says so in writing.  Refer to the GNU General Public License
-   for full details.
-
-   Everyone is granted permission to copy, modify and redistribute
-   zsh, but only under the conditions described in the GNU General Public
-   License.   A copy of this license is supposed to have been given to you
-   along with zsh so you can know your rights and responsibilities.
-   It should be in a file named COPYING.
-
-   Among other things, the copyright notice and this notice must be
-   preserved on all copies.
-
-*/
+ *
+ * init.c - main loop and initialization routines
+ *
+ * This file is part of zsh, the Z shell.
+ *
+ * This software is Copyright 1992 by Paul Falstad
+ *
+ * Permission is hereby granted to copy, reproduce, redistribute or otherwise
+ * use this software as long as: there is no monetary profit gained
+ * specifically from the use or reproduction of this software, it is not
+ * sold, rented, traded or otherwise marketed, and this copyright notice is
+ * included prominently in any copy made. 
+ *
+ * The author make no claims as to the fitness or correctness of this software
+ * for any use whatsoever, and it is provided as is. Any use of this software
+ * is at the user's own risk. 
+ *
+ */
 
 #define GLOBALS
 #include "zsh.h"
 #include <pwd.h>
 
-extern int yydebug;
-
 void main(argc,argv,envp) /**/
 int argc; char **argv; char **envp;
 {
 int notect = 0;
 
+#ifdef LC_ALL
+	setlocale(LC_ALL, "");
+#endif
 	environ = envp;
-	pathsuppress = 1;
 	meminit();
 	setflags();
 	parseargs(argv);
@@ -51,11 +39,6 @@ int notect = 0;
 	initialize();
 	heapalloc();
 	runscripts();
-	if (interact)
-		{
-		pathsuppress = 0;
-		newcmdnamtab();
-		}
 	for(;;)
 		{
 		do
@@ -83,6 +66,7 @@ int notect = 0;
 void loop() /**/
 {
 List list;
+Heap h = (Heap) peekfirst(heaplist);
 
 	pushheap();
 	for(;;)
@@ -124,25 +108,23 @@ List list;
 			exit(lastval);
 			}
 		}
-	popheap();
+	while ((Heap) peekfirst(heaplist) != h)
+		popheap();
 }
 
 void setflags() /**/
 {
 int c;
 
-	for (c = 0; c != 32; c++)
-		opts[c] = OPT_UNSET;
-	for (c = 32; c != 128; c++)
-		opts[c] = OPT_INVALID;
-	for (c = 'a'; c <= 'z'; c++)
-		opts[c] = opts[c-'a'+'A'] = OPT_UNSET;
-	for (c = '0'; c <= '9'; c++)
-		opts[c] = OPT_UNSET;
+	for (c = 0; c != 32; c++)		opts[c] = OPT_UNSET;
+	for (c = 32; c != 128; c++)	opts[c] = OPT_INVALID;
+	for (c = 'a'; c <= 'z'; c++)	opts[c] = opts[c-'a'+'A'] = OPT_UNSET;
+	for (c = '0'; c <= '9'; c++)	opts[c] = OPT_UNSET;
 	opts['A'] = OPT_INVALID;
 	opts['i'] = (isatty(0)) ? OPT_SET : OPT_UNSET;
 	opts[BGNICE] = opts[NOTIFY] = OPT_SET;
 	opts[USEZLE] = (interact && SHTTY != -1) ? OPT_SET : OPT_UNSET;
+	opts[HASHCMDS] = opts[HASHLISTALL] = opts[HASHDIRS] = OPT_SET;
 }
 
 static char *cmd;
@@ -160,49 +142,41 @@ Lklist paramlist;
 	while (!bk && *argv && (**argv == '-' || **argv == '+'))
 		{
 		action = (**argv == '-') ? OPT_SET : OPT_UNSET;
-		while (*++*argv)
-			{
-			if (opts[**argv] == OPT_INVALID)
-				{
+		while (*++*argv) {
+			if (opts[**argv] == OPT_INVALID) {
 				zerr("bad option: -%c",NULL,**argv);
 				exit(1);
-				}
-			opts[**argv] = action;
-			if (bk = **argv == 'b')
-				break;
-			if (**argv == 'c') /* -c command */
-				{
+			}
+			if (bk = **argv == 'b') break;
+			if (**argv == 'c') { /* -c command */
 				argv++;
-				if (!*argv)
-					{
+				if (!*argv) {
 					zerr("string expected after -c",NULL,0);
 					exit(1);
-					}
+				}
 				cmd = *argv;
 				opts[INTERACTIVE] = OPT_UNSET;
+				opts['c'] = OPT_SET;
 				break;
-				}
-			else if (**argv == 'o')
-				{
+			} else if (**argv == 'o') {
 				int c;
 
 				if (!*++*argv)
 					argv++;
-				if (!*argv)
-					{
+				if (!*argv) {
 					zerr("string expected after -o",NULL,0);
 					exit(1);
-					}
+				}
 				c = optlookup(*argv);
 				if (c == -1)
 					zerr("no such option: %s",*argv,0);
 				else
 					opts[c] = action;
 				break;
-				}
-			}
-		argv++;
+			} else opts[**argv] = action;
 		}
+		argv++;
+	}
 	paramlist = newlist();
 	if (*argv)
 		{
@@ -231,6 +205,7 @@ Lklist paramlist;
 void setmoreflags() /**/
 {
 int t0;
+long ttpgrp;
 
 	/* stdout,stderr fully buffered */
 #ifdef _IOFBF
@@ -248,53 +223,64 @@ int t0;
 #endif
 #ifdef JOB_CONTROL
 	opts[MONITOR] = (interact) ? OPT_SET : OPT_UNSET;
-	if (jobbing)
-		{
+	if (jobbing) {
 		SHTTY = movefd((isatty(0)) ? dup(0) : open("/dev/tty",O_RDWR));
 		if (SHTTY == -1)
 			opts[MONITOR] = OPT_UNSET;
-		else
-			{
+		else {
 #ifdef TIOCSETD
+#ifdef NTTYDISC
 			int ldisc = NTTYDISC;
 			ioctl(SHTTY, TIOCSETD, &ldisc);
+#endif
 #endif
 			gettyinfo(&shttyinfo);	/* get tty state */
+#ifdef sgi
+			if (shttyinfo.tio.c_cc[VSWTCH] <= 0) /* hack for irises */
+				shttyinfo.tio.c_cc[VSWTCH] = CSWTCH;
+#endif
 			savedttyinfo = shttyinfo;
-			}
+		}
 #ifdef sgi
 		setpgrp(0,getpgrp(0));
 #endif
 		if ((mypgrp = getpgrp(0)) <= 0)
 			opts[MONITOR] = OPT_UNSET;
+		else while ((ttpgrp = gettygrp()) != -1 && ttpgrp != mypgrp) {
+			sleep(1);
+			mypgrp = getpgrp(0);
+			if (mypgrp == gettygrp()) break;
+			killpg(mypgrp,SIGTTIN);
+			mypgrp = getpgrp(0);
 		}
-	else
+	} else
 		SHTTY = -1;
 #else
 	opts[MONITOR] = OPT_UNSET;
 	SHTTY = movefd((isatty(0)) ? dup(0) : open("/dev/tty",O_RDWR));
-	if (SHTTY != -1)
-		{
+	if (SHTTY != -1) {
 		gettyinfo(&shttyinfo);
 		savedttyinfo = shttyinfo;
-		}
+	}
 #endif
 }
 
 void setupvals() /**/
 {
-struct passwd *pwd;
-char *ptr;
+struct passwd *pswd;
+char *ptr,*s;
 static long bauds[] = {
 	0,50,75,110,134,150,200,300,600,1200,1800,2400,4800,9600,19200,38400
 	};
 
 	curhist = 0;
-	histsiz = 20;
+	histsiz = DEFAULT_HISTSIZE;
 	lithistsiz = 5;
-	logcheck = 60;
+	inithist();
+	mailcheck = logcheck = 60;
 	dirstacksize = -1;
 	listmax = 100;
+	reporttime = -1;
 	bangchar = '!';
 	hashchar = '#';
 	hatchar = '^';
@@ -306,24 +292,28 @@ static long bauds[] = {
 	aliastab = newhtable(37);
 	addreswords();
 	paramtab = newhtable(151);
-	cmdnamtab = newhtable(13);
+	cmdnamtab = newhtable(37);
+	compctltab = newhtable(13);
 	initxbindtab();
 	nullcmd = ztrdup("cat");
+	readnullcmd = ztrdup("more");
 	prompt = ztrdup("%m%# ");
 	prompt2 = ztrdup("> ");
 	prompt3 = ztrdup("?# ");
 	prompt4 = ztrdup("+ ");
-	sprompt = ztrdup("zsh: correct `%s' to `%r' [nyae]? ");
+	sprompt = ztrdup("zsh: correct `%R' to `%r' [nyae]? ");
+	term = ztrdup("");
 	ppid = getppid();
-#ifdef TERMIOS
-	baud = bauds[shttyinfo.termios.c_cflag & CBAUD];
+#ifdef TIO
+#ifdef HAS_TCCRAP
+	baud = cfgetospeed(&shttyinfo.tio);
+	if (baud < 100) baud = bauds[baud]; /* aren't "standards" great?? */
 #else
-#ifdef TERMIO
-	baud = bauds[shttyinfo.termio.c_cflag & CBAUD];
+	baud = bauds[shttyinfo.tio.c_cflag & CBAUD];
+#endif
 #else
 	baud = bauds[shttyinfo.sgttyb.sg_ospeed];
 #endif
-#endif
 #ifdef TIOCGWINSZ
 	if (!(columns = shttyinfo.winsize.ws_col))
 		columns = 80;
@@ -334,13 +324,17 @@ static long bauds[] = {
 	lines = 24;
 #endif
 	ifs = ztrdup(" \t\n");
-	if (pwd = getpwuid(getuid())) {
-		username = ztrdup(pwd->pw_name);
-		home = ztrdup(pwd->pw_dir);
+	if (pswd = getpwuid(getuid())) {
+		username = ztrdup(pswd->pw_name);
+		home = ztrdup(pswd->pw_dir);
 	} else {
 		username = ztrdup("");
 		home = ztrdup("/");
 	}
+	if (ptr = zgetenv("LOGNAME"))
+		logname = ztrdup(ptr);
+	else
+		logname = ztrdup(username);
 	timefmt = ztrdup(DEFTIMEFMT);
 	watchfmt = ztrdup(DEFWATCHFMT);
 	if (!(ttystrname = ztrdup(ttyname(SHTTY))))
@@ -348,40 +342,29 @@ static long bauds[] = {
 	wordchars = ztrdup(DEFWORDCHARS);
 	fceditparam = ztrdup(DEFFCEDIT);
 	tmpprefix = ztrdup(DEFTMPPREFIX);
-	if (ispwd(home))
-		cwd = ztrdup(home);
-	else if ((ptr = zgetenv("PWD")) && ispwd(ptr))
-		cwd = ztrdup(ptr);
-	else
-		cwd = zgetwd();
-	oldpwd = ztrdup(cwd);
+	postedit = ztrdup("");
+	if (ispwd(home)) pwd = ztrdup(home);
+	else if ((ptr = zgetenv("PWD")) && ispwd(ptr)) pwd = ztrdup(ptr);
+	else pwd = zgetwd();
+	oldpwd = ztrdup(pwd);
 	hostnam = zalloc(256);
 	underscore = ztrdup("");
 	gethostname(hostnam,256);
 	mypid = getpid();
 	cdpath = mkarray(NULL);
+	manpath = mkarray(NULL);
 	fignore = mkarray(NULL);
 	fpath = mkarray(NULL);
 	mailpath = mkarray(NULL);
 	watch = mkarray(NULL);
 	hosts = mkarray(NULL);
-	hostcmds = (char **) zcalloc(sizeof(char *)*7);
-	hostcmds[0] = ztrdup("telnet"); hostcmds[1] = ztrdup("rlogin");
-	hostcmds[2] = ztrdup("ftp"); hostcmds[3] = ztrdup("rup");
-	hostcmds[4] = ztrdup("rusers"); hostcmds[5] = ztrdup("rsh");
-	optcmds = (char **) zcalloc(sizeof(char *)*3);
-	optcmds[0] = ztrdup("setopt"); optcmds[1] = ztrdup("unsetopt");
-	bindcmds = (char **) zcalloc(sizeof(char *)*2);
-	bindcmds[0] = ztrdup("bindkey");
-	varcmds = (char **) zcalloc(sizeof(char *)*5);
-	varcmds[0] = ztrdup("export"); varcmds[1] = ztrdup("typeset");
-	varcmds[2] = ztrdup("vared"); varcmds[3] = ztrdup("unset");
+	compctlsetup();
 	userdirs = (char **) zcalloc(sizeof(char *)*2);
 	usernames = (char **) zcalloc(sizeof(char *)*2);
 	userdirsz = 2;
 	userdirct = 0;
 	optarg = ztrdup("");
-	optind = 0;
+	zoptind = 1;
 	schedcmds = NULL;
 	path = (char **) zalloc(4*sizeof *path);
 	path[0] = ztrdup("/bin"); path[1] = ztrdup("/usr/bin");
@@ -391,6 +374,29 @@ static long bauds[] = {
 	setupparams();
 	setparams();
 	inittyptab();
+	if ((s = zgetenv("EMACS")) && !strcmp(s,"t") && !strcmp(term,"emacs"))
+		opts[USEZLE] = OPT_UNSET;
+#ifndef HAS_RUSAGE
+	times(&shtms);
+#endif
+}
+
+void compctlsetup() /**/
+{
+static char
+	*hs[] = {"telnet","rlogin","ftp","rup","rusers","rsh",NULL},
+	*os[] = {"setopt","unsetopt",NULL},
+	*vs[] = {"export","typeset","vared","unset",NULL},
+	*cs[] = {"which","builtin",NULL},
+	*bs[] = {"bindkey",NULL};
+
+	compctl_process(hs,CC_HOSTS|CC_FILES,NULL);
+	compctl_process(os,CC_OPTIONS,NULL);
+	compctl_process(vs,CC_VARS,NULL);
+	compctl_process(bs,CC_BINDINGS,NULL);
+	compctl_process(cs,CC_COMMPATH,NULL);
+	cc_compos.mask   = CC_COMMPATH;
+	cc_default.mask  = CC_FILES;
 }
 
 void initialize() /**/
@@ -399,10 +405,6 @@ int t0;
 
 	breaks = loops = 0;
 	lastmailcheck = time(NULL);
-	firsthist = firstlithist = 1;
-	histsiz = DEFAULT_HISTSIZE;
-	histlist = newlist();
-	lithistlist = newlist();
 	locallist = NULL;
 	dirstack = newlist();
 	bufstack = newlist();
@@ -410,7 +412,7 @@ int t0;
 	inbuf = zalloc(inbufsz = 256);
 	inbufptr = inbuf+inbufsz;
 	inbufct = 0;
-#ifdef QDEBUG
+#ifndef QDEBUG
 	signal(SIGQUIT,SIG_IGN);
 #endif
 #ifdef RLIM_INFINITY
@@ -423,6 +425,18 @@ int t0;
 	signal(SIGCHLD,handler);
 	if (jobbing)
 		{
+		int ttypgrp;
+		for (;;)
+			{
+#ifdef TIOCGPGRP
+			ioctl(SHTTY, TIOCGPGRP, &ttypgrp);
+#else
+			ttypgrp = tcgetpgrp(SHTTY);
+#endif
+			if (ttypgrp == mypgrp)
+				break;
+			kill(0,SIGTTIN);
+			}
 		signal(SIGTTOU,SIG_IGN);
 		signal(SIGTSTP,SIG_IGN);
 		signal(SIGTTIN,SIG_IGN);
@@ -455,9 +469,12 @@ int t0;
 
 void runscripts() /**/
 {
+#ifdef GLOBALZSHENV
+	source(GLOBALZSHENV);
+#endif
 	if (opts[NORCS] == OPT_SET) {
 #ifdef GLOBALZPROFILE
-		source(GLOBALZPROFILE);
+		if (islogin) source(GLOBALZPROFILE);
 #endif
 #ifdef GLOBALZSHRC
 		source(GLOBALZSHRC);
@@ -465,27 +482,27 @@ void runscripts() /**/
 #ifdef GLOBALZLOGIN
 		if (islogin) source(GLOBALZLOGIN);
 #endif
-		return;
-	}
-	sourcehome(".zshenv");
-	if (opts[NORCS] == OPT_SET)
-		return;
-	if (interact) {
-		if (islogin) {
-			sourcehome(".zprofile");
+	} else {
+		sourcehome(".zshenv");
+		if (opts[NORCS] == OPT_SET)
+			return;
+		if (interact) {
+			if (islogin) {
+				sourcehome(".zprofile");
 #ifdef GLOBALZPROFILE
-			source(GLOBALZPROFILE);
+				source(GLOBALZPROFILE);
 #endif
-		}
+			}
 #ifdef GLOBALZSHRC
-		source(GLOBALZSHRC);
+			source(GLOBALZSHRC);
 #endif
-		sourcehome(".zshrc");
-		if (islogin) {
+			sourcehome(".zshrc");
+			if (islogin) {
 #ifdef GLOBALZLOGIN
-			source(GLOBALZLOGIN);
+				source(GLOBALZLOGIN);
 #endif
-			sourcehome(".zlogin");
+				sourcehome(".zlogin");
+			}
 		}
 	}
 	if (interact)
@@ -495,8 +512,9 @@ void runscripts() /**/
 		if (SHIN >= 10)
 			close(SHIN);
 		SHIN = movefd(open("/dev/null",O_RDONLY));
-		hungets(cmd);
-		strinbeg();
+		execstring(cmd);
+		stopmsg = 1;
+		zexit(NULL);
 		}
 #ifdef TIOCSWINSZ
 	if (!(columns = shttyinfo.winsize.ws_col))
@@ -513,69 +531,3 @@ void ainit() /**/
 	isfirstln = 1;
 }
 
-void readhistfile(s,err) /**/
-char *s;int err;
-{
-char buf[1024];
-FILE *in;
-
-	if (!s)
-		return;
-	if (in = fopen(s,"r"))
-		{
-		permalloc();
-		while (fgets(buf,1024,in))
-			{
-			int l = strlen(buf);
-			char *pt = buf;
-
-			if (l && buf[l-1] == '\n')
-				buf[l-1] = '\0';
-			for (;*pt;pt++)
-			    if (*pt == ' ')
-				*pt = HISTSPACE;
-			addnode(histlist,ztrdup(buf));
-			addnode(lithistlist,ztrdup(buf));
-			curhist++;
-			}
-		fclose(in);
-		lastalloc();
-		}
-	else if (err)
-		zerr("can't read history file",s,0);
-}
-
-void savehistfile(s,err) /**/
-char *s;int err;
-{
-char *t;
-Lknode n;
-Lklist l;
-FILE *out;
-
-	if (!s || !interact)
-		return;
-	if (out = fdopen(open(s,O_CREAT|O_WRONLY|O_TRUNC,0600),"w"))
-		{
-		n = lastnode(l = (isset(HISTLIT) ? lithistlist : histlist));
-		if (n == (Lknode) l)
-			{
-			fclose(out);
-			return;
-			}
-		while (--savehist && prevnode(n) != (Lknode) l)
-			n = prevnode(n);
-		for (; n; incnode(n))
-			{
-			for (s = t = getdata(n); *s; s++)
-				if (*s == HISTSPACE)
-					*s = ' ';
-			fputs(t,out);
-			fputc('\n',out);
-			}
-		fclose(out);
-		}
-	else if (err)
-		zerr("can't write history file: %s",s,0);
-}
-
