@@ -1,34 +1,22 @@
 /*
-
-	jobs.c - job control
-
-	This file is part of zsh, the Z shell.
-
-	zsh is free software; no one can prevent you from reading the source
-   code, or giving it to someone else.
-
-   This file is copyrighted under the GNU General Public License, which
-   can be found in the file called COPYING.
-
-   Copyright (C) 1990, 1991 Paul Falstad
-
-   zsh is distributed in the hope that it will be useful, but
-   WITHOUT ANY WARRANTY.  No author or distributor accepts
-   responsibility to anyone for the consequences of using it or for
-   whether it serves any particular purpose or works at all, unless he
-   says so in writing.  Refer to the GNU General Public License
-   for full details.
-
-   Everyone is granted permission to copy, modify and redistribute
-   zsh, but only under the conditions described in the GNU General Public
-   License.   A copy of this license is supposed to have been given to you
-   along with zsh so you can know your rights and responsibilities.
-   It should be in a file named COPYING.
-
-   Among other things, the copyright notice and this notice must be
-   preserved on all copies.
-
-*/
+ *
+ * jobs.c - job control
+ *
+ * This file is part of zsh, the Z shell.
+ *
+ * This software is Copyright 1992 by Paul Falstad
+ *
+ * Permission is hereby granted to copy, reproduce, redistribute or otherwise
+ * use this software as long as: there is no monetary profit gained
+ * specifically from the use or reproduction of this software, it is not
+ * sold, rented, traded or otherwise marketed, and this copyright notice is
+ * included prominently in any copy made. 
+ *
+ * The author make no claims as to the fitness or correctness of this software
+ * for any use whatsoever, and it is provided as is. Any use of this software
+ * is at the user's own risk. 
+ *
+ */
 
 #include "zsh.h"
 #include <sys/errno.h>
@@ -37,10 +25,6 @@
 
 static int handling = 0;
 
-/* != 0 means the shell is waiting for a job to complete */
-
-static int waiting = 0;
-
 #ifdef INTHANDTYPE
 #define RETURN return 0
 #else
@@ -56,7 +40,11 @@ long pid;
 int statusp;
 Job jn;
 struct process *pn;
+#ifdef HAS_RUSAGE
 struct rusage ru;
+#else
+long chlds,chldu;
+#endif
 
 #ifdef RESETHANDNEEDED
 	signal(sig,handler);
@@ -85,14 +73,17 @@ struct rusage ru;
 		}
 	for (;;)
 		{
-#ifdef SYSV
+#ifdef HAS_RUSAGE
+		pid = wait3((vptr) &statusp,WNOHANG|WUNTRACED,&ru);
+#else
+#ifndef WNOHANG
 		pid = wait(&statusp);
 #else
-#if defined(RLIM_INFINITY)  & ! defined(__hpux)
-		pid = wait3(&statusp,WNOHANG|WUNTRACED,&ru);
-#else
-		pid = wait3(&statusp,WNOHANG|WUNTRACED,NULL);
+		pid = wait3((vptr) &statusp,WNOHANG|WUNTRACED,NULL);
 #endif
+		chlds = shtms.tms_cstime;
+		chldu = shtms.tms_cutime;
+		times(&shtms);
 #endif
 		if (pid == -1)
 			{
@@ -107,17 +98,21 @@ struct rusage ru;
 			{
 			pn->statusp = statusp;
 			handling = 1;
-			pn->ru = ru;
+#ifdef HAS_RUSAGE
+			pn->ti.ru = ru;
+#else
+			pn->ti.st = shtms.tms_cstime-chlds;
+			pn->ti.ut = shtms.tms_cutime-chldu;
+#endif
 			pn->endtime = time(NULL);
 			updatestatus(jn);
 			handling = 0;
-			if (zleactive)
-				refresh();
 			}
-		else if (WIFSTOPPED(SP(statusp)))
+#if 0
+		else if (WIFSTOPPED(statusp))
 			kill(pid,SIGKILL);	/* kill stopped untraced children */
+#endif
 		}
-	RETURN;
 }
 
 /* change job table entry from stopped to running */
@@ -129,7 +124,7 @@ struct process *pn;
 
 	jn->stat &= ~STAT_STOPPED;
 	for (pn = jn->procs; pn; pn = pn->next)
-		if (WIFSTOPPED(SP(pn->statusp)))
+		if (WIFSTOPPED(pn->statusp))
 			pn->statusp = SP_RUNNING;
 }
 
@@ -145,13 +140,13 @@ int notrunning = 1,alldone = 1,val,job = jn-jobtab,somestopped = 0;
 		{
 		if (pn->statusp == SP_RUNNING)
 			notrunning = 0;
-		if (pn->statusp == SP_RUNNING || WIFSTOPPED(SP(pn->statusp)))
+		if (pn->statusp == SP_RUNNING || WIFSTOPPED(pn->statusp))
 			alldone = 0;
-		if (WIFSTOPPED(SP(pn->statusp)))
+		if (WIFSTOPPED(pn->statusp))
 			somestopped = 1;
 		if (!pn->next && jn)
-			val = (WIFSIGNALED(SP(pn->statusp))) ?
-				0200 | WTERMSIG(SP(pn->statusp)) : WEXITSTATUS(SP(pn->statusp));
+			val = (WIFSIGNALED(pn->statusp)) ?
+				0200 | WTERMSIG(pn->statusp) : WEXITSTATUS(pn->statusp);
 		}
 	if (!notrunning)
 		return;
@@ -161,7 +156,7 @@ int notrunning = 1,alldone = 1,val,job = jn-jobtab,somestopped = 0;
 		STAT_CHANGED|STAT_STOPPED;
 	if (alldone && job == thisjob)
 		{
-		if (!val) {
+		if (!ttyfrozen && !val) {
 			gettyinfo(&shttyinfo);
 			if (interact && isset(SHINSTDIN) && SHTTY != -1 && isset(USEZLE))
 				sanetty(&shttyinfo);
@@ -178,8 +173,10 @@ int notrunning = 1,alldone = 1,val,job = jn-jobtab,somestopped = 0;
 		prevjob = curjob;
 		curjob = job;
 	}
-	if ((isset(NOTIFY) || job == thisjob) && jn->stat & STAT_LOCKED)
+	if ((isset(NOTIFY) || job == thisjob) && jn->stat & STAT_LOCKED) {
 		printjob(jn,!!isset(LONGLISTJOBS));
+		if (zleactive) refresh();
+	}
 	if (sigtrapped[SIGCHLD] && job != thisjob)
 		dotrap(SIGCHLD);
 }
@@ -229,9 +226,9 @@ struct process *pn;
 	for (pn = jn->procs; pn; pn = pn->next)
 		{
 		if (pn->statusp != SP_RUNNING)
-			if (WIFSIGNALED(SP(pn->statusp)))
+			if (WIFSIGNALED(pn->statusp))
 				{
-				sig = WTERMSIG(SP(pn->statusp));
+				sig = WTERMSIG(pn->statusp);
 				llen = strlen(sigmsg[sig]);
 				if (WCOREDUMPED(pn->statusp))
 					llen += 14;
@@ -244,15 +241,16 @@ struct process *pn;
 				if (job == thisjob && sig == SIGINT)
 					doputnl = 1;
 				}
-			else if (WIFSTOPPED(SP(pn->statusp)))
+			else if (WIFSTOPPED(pn->statusp))
 				{
-				sig = WSTOPSIG(SP(pn->statusp));
+				sig = WSTOPSIG(pn->statusp);
 				if (strlen(sigmsg[sig]) > len)
 					len = strlen(sigmsg[sig]);
 				if (job == thisjob && sig == SIGTSTP)
 					doputnl = 1;
 				}
-			else if (isset(PRINTEXITVALUE) && WEXITSTATUS(SP(pn->statusp)))
+			else if (isset(PRINTEXITVALUE) && isset(SHINSTDIN) &&
+					WEXITSTATUS(pn->statusp))
 				sflag = 1;
 		}
 
@@ -307,20 +305,20 @@ struct process *pn;
 					fprintf(stderr,"running%*s",len-7+2,"");
 				else
 					fprintf(stderr,"continued%*s",len-9+2,"");
-			else if (WIFEXITED(SP(pn->statusp)))
-				if (WEXITSTATUS(SP(pn->statusp)))
-					fprintf(stderr,"exit %-4d%*s",WEXITSTATUS(SP(pn->statusp)),
+			else if (WIFEXITED(pn->statusp))
+				if (WEXITSTATUS(pn->statusp))
+					fprintf(stderr,"exit %-4d%*s",WEXITSTATUS(pn->statusp),
 						len-9+2,"");
 				else
 					fprintf(stderr,"done%*s",len-4+2,"");
-			else if (WIFSTOPPED(SP(pn->statusp)))
-				fprintf(stderr,"%-*s",len+2,sigmsg[WSTOPSIG(SP(pn->statusp))]);
+			else if (WIFSTOPPED(pn->statusp))
+				fprintf(stderr,"%-*s",len+2,sigmsg[WSTOPSIG(pn->statusp)]);
 			else if (WCOREDUMPED(pn->statusp))
 				fprintf(stderr,"%s (core dumped)%*s",
-					sigmsg[WTERMSIG(SP(pn->statusp))],
-					len-14+2-strlen(sigmsg[WTERMSIG(SP(pn->statusp))]),"");
+					sigmsg[WTERMSIG(pn->statusp)],
+					len-14+2-strlen(sigmsg[WTERMSIG(pn->statusp)]),"");
 			else
-				fprintf(stderr,"%-*s",len+2,sigmsg[WTERMSIG(SP(pn->statusp))]);
+				fprintf(stderr,"%-*s",len+2,sigmsg[WTERMSIG(pn->statusp)]);
 			for (; pn != qn; pn = pn->next)
 				fprintf(stderr,(pn->next) ? "%s | " : "%s",pn->text);
 			putc('\n',stderr);
@@ -334,10 +332,10 @@ struct process *pn;
 
 	/* print "(pwd now: foo)" messages */
 
-	if (interact && job==thisjob && strcmp(jn->cwd,cwd))
+	if (interact && job==thisjob && strcmp(jn->pwd,pwd))
 		{
 		printf("(pwd now: ");
-		printdir(cwd);
+		printdir(pwd);
 		printf(")\n");
 		fflush(stdout);
 		}
@@ -350,19 +348,16 @@ struct process *pn;
 		struct process *nx;
 		char *s;
 
-		if (jn->stat & STAT_TIMED)
-			{
+		if ((jn->stat & STAT_TIMED) || (reporttime != -1 && report(jn))) {
 			dumptime(jn);
 			printed = 1;
-			}
+		}
 		for (pn = jn->procs; pn; pn = nx)
 			{
 			nx = pn->next;
-			if (pn->text)
-				free(pn->text);
 			free(pn);
 			}
-		free(jn->cwd);
+		free(jn->pwd);
 		if (jn->filelist)
 			{
 			while (s = getnode(jn->filelist))
@@ -406,7 +401,7 @@ int t0;
 
 void initjob() /**/
 {
-	jobtab[thisjob].cwd = ztrdup(cwd);
+	jobtab[thisjob].pwd = ztrdup(pwd);
 	jobtab[thisjob].stat = STAT_INUSE;
 	jobtab[thisjob].gleader = 0;
 }
@@ -418,25 +413,20 @@ long pid;char *text;
 {
 struct process *process;
 
-	if (!jobtab[thisjob].gleader)
-		jobtab[thisjob].gleader = lastpid = pid;
-	lastpid = pid;
+	if (!jobtab[thisjob].gleader) jobtab[thisjob].gleader = pid;
 	process = zcalloc(sizeof *process);
 	process->pid = pid;
-	process->text = text;
+	if (text) strcpy(process->text,text);
+	else *process->text = '\0';
 	process->next = NULL;
 	process->statusp = SP_RUNNING;
 	process->bgtime = time(NULL);
-	if (jobtab[thisjob].procs)
-		{
+	if (jobtab[thisjob].procs) {
 		struct process *n;
-
-		for (n = jobtab[thisjob].procs; n->next && !n->next->lastfg; n = n->next);
-		process->next = n->next;
+		for (n = jobtab[thisjob].procs; n->next; n = n->next);
+		process->next = NULL;
 		n->next = process;
-		}
-	else
-		jobtab[thisjob].procs = process;
+	} else jobtab[thisjob].procs = process;
 	return process;
 }
 
@@ -454,6 +444,13 @@ Job jn;
 		if (jn->stat && jn->filelist)
 			return 0;
 	return 1;
+
+}
+
+void waitforpid(pid) /**/
+long pid;
+{
+	while (!errflag && (kill(pid,0) >= 0 || errno != ESRCH)) chldsuspend();
 }
 
 /* wait for a job to finish */
@@ -467,20 +464,18 @@ Job jn;
 	if (jobtab[job].procs)	/* if any forks were done */
 		{
 		jobtab[job].stat |= STAT_LOCKED;
-		waiting = 1;
 		if (jobtab[job].stat & STAT_CHANGED)
 			printjob(jobtab+job,!!isset(LONGLISTJOBS));
 		while (jobtab[job].stat &&
 				!(jobtab[job].stat & (STAT_DONE|STAT_STOPPED)))
 			chldsuspend();
-		waiting = 0;
 		}
 	else	/* else do what printjob() usually does */
 		{
 		char *s;
 
 		jn = jobtab+job;
-		free(jn->cwd);
+		free(jn->pwd);
 		if (jn->filelist)
 			{
 			while (s = getnode(jn->filelist))
@@ -559,7 +554,7 @@ struct process *pn;
 		struct job *jn;
 
 		jn = jobtab+thisjob;
-		free(jn->cwd);
+		free(jn->pwd);
 		if (jn->filelist)
 			{
 			while (s = getnode(jn->filelist))
@@ -581,18 +576,59 @@ void fixsigs() /**/
 	unblockchld();
 }
 
-void printtime(real,ru,desc) /**/
-time_t real;struct rusage *ru;char *desc;
+int report(j) /**/
+Job j;
+{
+	if (!j->procs) return 0;
+#ifdef HAS_RUSAGE
+	return (j->procs->ti.ru.ru_utime.tv_sec+j->procs->ti.ru.ru_stime.tv_sec)
+				 >= reporttime;
+#else
+	return (j->procs->ti.ut+j->procs->ti.st)/HZ >= reporttime;
+#endif
+}
+
+void printtime(real,ti,desc) /**/
+time_t real;struct timeinfo *ti;char *desc;
 {
 char *s;
+#ifdef sun
+long ticks = 1;
+int pk = getpagesize()/1024;
+#else
+long sec;
+#endif
+#ifdef HAS_RUSAGE
+struct rusage *ru = &ti->ru;
+#endif
 
-	if (!desc)
-		desc = "";
-	for (s = (timefmt) ? timefmt : DEFTIMEFMT; *s; s++)
+	if (!desc) desc = "";
+#ifdef HAS_RUSAGE
+#ifdef sun
+	ticks = (ru->ru_utime.tv_sec+ru->ru_stime.tv_sec)*HZ+
+			  (ru->ru_utime.tv_usec+ru->ru_stime.tv_usec)*HZ/1000000;
+	if (!ticks) ticks = 1;
+#else
+	sec = ru->ru_utime.tv_sec + ru->ru_stime.tv_sec;
+	if (!sec) sec = 1;
+#endif
+#endif
+	for (s = timefmt; *s; s++)
 		if (*s == '%')
 			switch(s++,*s)
 				{
 				case 'E': fprintf(stderr,"%lds",real); break;
+#ifndef HAS_RUSAGE
+				case 'U': fprintf(stderr,"%ld.%03lds",
+					ti->ut/HZ,ti->ut*1000/60%1000); break;
+				case 'S': fprintf(stderr,"%ld.%03lds",
+					ti->st/HZ,ti->st*1000/60%1000); break;
+				case 'P':
+					if (real)
+						fprintf(stderr,"%d%%",
+							(int) (100*((ti->ut+ti->st)/HZ))/real);
+					break;
+#else
 				case 'U': fprintf(stderr,"%ld.%03lds",
 					ru->ru_utime.tv_sec,ru->ru_utime.tv_usec/1000); break;
 				case 'S': fprintf(stderr,"%ld.%03lds",
@@ -600,14 +636,22 @@ char *s;
 				case 'P':
 					if (real)
 						fprintf(stderr,"%d%%",
-							(int) (100*ru->ru_utime.tv_sec+ru->ru_stime.tv_sec)
+							(int) (100*(ru->ru_utime.tv_sec+ru->ru_stime.tv_sec))
 								/ real);
 					break;
 				case 'W': fprintf(stderr,"%ld",ru->ru_nswap); break;
-				case 'X': fprintf(stderr,"%ld",ru->ru_ixrss); break;
-				case 'D': fprintf(stderr,"%ld",ru->ru_idrss); break;
-				case 'K': fprintf(stderr,"%ld",ru->ru_ixrss+ru->ru_idrss); break;
-				case 'M': fprintf(stderr,"%ld",ru->ru_maxrss); break;
+#ifdef sun
+				case 'K': case 'D':
+					fprintf(stderr,"%ld",ru->ru_idrss/ticks*pk); break;
+				case 'M': fprintf(stderr,"%ld",ru->ru_maxrss*pk); break;
+#else
+				case 'X': fprintf(stderr,"%ld",ru->ru_ixrss/sec); break;
+				case 'D': fprintf(stderr,"%ld",
+					(ru->ru_idrss+ru->ru_isrss)/sec); break;
+				case 'K': fprintf(stderr,"%ld",
+					(ru->ru_ixrss+ru->ru_idrss+ru->ru_isrss)/sec); break;
+				case 'M': fprintf(stderr,"%ld",ru->ru_maxrss/1024); break;
+#endif
 				case 'F': fprintf(stderr,"%ld",ru->ru_majflt); break;
 				case 'R': fprintf(stderr,"%ld",ru->ru_minflt); break;
 				case 'I': fprintf(stderr,"%ld",ru->ru_inblock); break;
@@ -617,7 +661,9 @@ char *s;
 				case 'k': fprintf(stderr,"%ld",ru->ru_nsignals); break;
 				case 'w': fprintf(stderr,"%ld",ru->ru_nvcsw); break;
 				case 'c': fprintf(stderr,"%ld",ru->ru_nivcsw); break;
-				default: fprintf(stderr,"%%%c",s[-1]); break;
+#endif
+				case 'J': fprintf(stderr,"%s",desc); break;
+				default: fprintf(stderr,"%%%c",*s); break;
 				}
 		else
 			putc(*s,stderr);
@@ -633,7 +679,33 @@ struct process *pn = jn->procs;
 	if (!jn->procs)
 		return;
 	for (pn = jn->procs; pn; pn = pn->next)
-		printtime(pn->endtime-pn->bgtime,&pn->ru,pn->text);
+		printtime(pn->endtime-pn->bgtime,&pn->ti,pn->text);
+}
+
+void shelltime() /**/
+{
+struct timeinfo ti;
+#ifdef HAS_RUSAGE
+struct rusage ru;
+
+	getrusage(RUSAGE_SELF,&ru);
+	memcpy(&ti.ru,&ru,sizeof(ru));
+	printtime(time(NULL)-shtimer,&ti,"shell");
+
+	getrusage(RUSAGE_CHILDREN,&ru);
+	memcpy(&ti.ru,&ru,sizeof(ru));
+	printtime(time(NULL)-shtimer,&ti,"children");
+#else
+struct tms buf;
+
+	times(&buf);
+	ti.ut = buf.tms_utime;
+	ti.st = buf.tms_stime;
+	printtime(time(NULL)-shtimer,&ti,"shell");
+	ti.ut = buf.tms_cutime;
+	ti.st = buf.tms_cstime;
+	printtime(time(NULL)-shtimer,&ti,"children");
+#endif
 }
 
 /* SIGHUP any jobs left running */
@@ -644,14 +716,11 @@ int t0,killed = 0;
 
 	if (isset(NOHUP)) return;
 	for (t0 = 1; t0 != MAXJOB; t0++)
-		if (t0 != thisjob && jobtab[t0].stat &&
-				!(jobtab[t0].stat & STAT_STOPPED))
-			{
-			kill(-jobtab[t0].gleader,SIGHUP);
-			killed++;
-			}
-	if (killed)
-		zerr("warning: %d jobs SIGHUPed",NULL,killed);
+		if (t0 != thisjob && (jobtab[t0].stat & STAT_LOCKED) &&
+				!(jobtab[t0].stat & STAT_STOPPED)) {
+			if (killpg(jobtab[t0].gleader,SIGHUP) != -1) killed++;
+		}
+	if (killed) zerr("warning: %d jobs SIGHUPed",NULL,killed);
 }
 
 /* check to see if user has jobs running/stopped */
@@ -660,23 +729,21 @@ void checkjobs() /**/
 {
 int t0;
 
+	scanjobs();
 	for (t0 = 1; t0 != MAXJOB; t0++)
-		if (t0 != thisjob && jobtab[t0].stat)
+		if (t0 != thisjob && jobtab[t0].stat & STAT_LOCKED)
 			break;
-	if (t0 != MAXJOB)
-		{
-		if (jobtab[t0].stat & STAT_STOPPED)
-			{
+	if (t0 != MAXJOB) {
+		if (jobtab[t0].stat & STAT_STOPPED) {
 #ifdef USE_SUSPENDED
 			zerr("you have suspended jobs.",NULL,0);
 #else
 			zerr("you have stopped jobs.",NULL,0);
 #endif
-			}
-		else
+		} else
 			zerr("you have running jobs.",NULL,0);
 		stopmsg = 1;
-		}
+	}
 }
 
 /* send a signal to a job (simply involves kill if monitoring is on) */
@@ -688,7 +755,7 @@ struct process *pn;
 int err;
 
 	if (jobbing)
-		return(kill(-jn->gleader,sig));
+		return(killpg(jn->gleader,sig));
 	for (pn = jn->procs; pn; pn = pn->next)
 		if ((err = kill(pn->pid,sig)) == -1 && errno != ESRCH)
 			return -1;
