@@ -1,33 +1,22 @@
 /*
-
-	lex.c - lexical analysis
-
-	This file is part of zsh, the Z shell.
-
-   zsh is free software; no one can prevent you from reading the source
-   code, or giving it to someone else.
-   This file is copyrighted under the GNU General Public License, which
-   can be found in the file called COPYING.
-
-   Copyright (C) 1990, 1991 Paul Falstad
-
-   zsh is distributed in the hope that it will be useful, but
-   WITHOUT ANY WARRANTY.  No author or distributor accepts
-   responsibility to anyone for the consequences of using it or for
-   whether it serves any particular purpose or works at all, unless he
-   says so in writing.  Refer to the GNU General Public License
-   for full details.
-
-   Everyone is granted permission to copy, modify and redistribute
-   zsh, but only under the conditions described in the GNU General Public
-   License.   A copy of this license is supposed to have been given to you
-   along with zsh so you can know your rights and responsibilities.
-   It should be in a file named COPYING.
-
-   Among other things, the copyright notice and this notice must be
-   preserved on all copies.
-
-*/
+ *
+ * lex.c - lexical analysis
+ *
+ * This file is part of zsh, the Z shell.
+ *
+ * This software is Copyright 1992 by Paul Falstad
+ *
+ * Permission is hereby granted to copy, reproduce, redistribute or otherwise
+ * use this software as long as: there is no monetary profit gained
+ * specifically from the use or reproduction of this software, it is not
+ * sold, rented, traded or otherwise marketed, and this copyright notice is
+ * included prominently in any copy made. 
+ *
+ * The author make no claims as to the fitness or correctness of this software
+ * for any use whatsoever, and it is provided as is. Any use of this software
+ * is at the user's own risk. 
+ *
+ */
 
 #include "zsh.h"
 
@@ -63,6 +52,7 @@ void lexsave() /**/
 	xhptr = hptr;
 	xhlastw = hlastw;
 	xhlinesz = hlinesz;
+	inredir = 0;
 }
 
 /* restore lexical state */
@@ -96,14 +86,14 @@ void yylex() /**/
 	do
 		tok = gettok();
 	while (tok != ENDINPUT && exalias());
-	isnewlin = (tok == NEWLIN && !inbufct);
-	if (tok == SEMI || tok == NEWLIN)
-		tok = SEPER;
+	if (tok != NEWLIN) isnewlin = 0;
+	else isnewlin = (inbufct) ? -1 : 1;
+	if (tok == SEMI || tok == NEWLIN) tok = SEPER;
 }
 
 void ctxtlex() /**/
 {
-static int oldpos,inredir = 0;
+static int oldpos;
 
 	yylex();
 	switch (tok) {
@@ -111,16 +101,16 @@ static int oldpos,inredir = 0;
 	case INPAR: case INBRACE: case DBAR: case DAMPER: case BAR:
 	case BARAMP: case INOUTPAR: case DO: case THEN: case ELIF:
 	case ELSE: incmdpos = 1; break;
-	case STRING: case ENVSTRING: case ENVARRAY: case OUTPAR:
-	case FOR: case SELECT: case FOREACH: case CASE: incmdpos = 0; break;
+	case STRING: /* case ENVSTRING: */ case ENVARRAY: case OUTPAR:
+	case CASE: incmdpos = 0; break;
 	}
-	if (IS_REDIROP(tok)) {
-		inredir = 2;
+	if (IS_REDIROP(tok) || tok == FOR || tok == FOREACH || tok == SELECT) {
+		inredir = 1;
 		oldpos = incmdpos;
 		incmdpos = 0;
 	} else if (inredir) {
-		if (!--inredir) incmdpos = oldpos;
-		else incmdpos = 0;
+		incmdpos = oldpos;
+		inredir = 0;
 	}
 }
 
@@ -232,14 +222,14 @@ int c;
 
 static void unadd()
 {
-	bptr--; bsiz--;
+	bptr--; len--;
 }
 
 int gettok() /**/
 {
 int bct = 0,pct = 0,brct = 0;
 int c,d,intpos = 1;
-int peekfd = -1,peek;
+int peekfd = -1,peek,ninbracks;
 
 beginning:
 	hlastw = NULL;
@@ -511,8 +501,12 @@ beginning:
 					{
 					add(String);
 					add(Inbrack);
-					while ((c = hgetc()) != ']' && !lexstop)
-						add(c);
+					ninbracks = 1;
+					while (ninbracks && (c = hgetc()) && !lexstop) {
+						if (c == '[') ninbracks++;
+						else if (c == ']') ninbracks--;
+						if (ninbracks) add(c);
+					}
 					c = Outbrack;
 					}
 				else if (d == '(')
@@ -535,7 +529,7 @@ beginning:
 			case LX2_INPAR:
 				d = hgetc();
 				hungetc(d);
-				if (d == ')' || incmdpos)
+				if (d == ')' || (incmdpos && peek != ENVSTRING))
 					goto brk;
 				pct++;
 				c = Inpar;
@@ -752,32 +746,32 @@ char *s,*t;
 	if (interact && isset(SHINSTDIN) && !strin && !incasepat && tok == STRING &&
 		(isset(CORRECTALL) || (isset(CORRECT) && incmdpos)) && !nocorrect)
 			spckword(&tokstr,&s,&t,!incmdpos,1);
-	if (zleparse && !alstackind)
+	if (zleparse && !alstackind) {
+		int zp = zleparse;
 		gotword(s);
+		if (zp && !zleparse) {
+			if (t) *t = HISTSPACE;
+			return 0;
+		}
+	}
 	an = gethnode(s,aliastab);
-	if (t)
-		*t = HISTSPACE;
+	if (t) *t = HISTSPACE;
 	if (alstackind != MAXAL && an && !an->inuse)
-		if (!(an->cmd && !incmdpos && alstat != ALSTAT_MORE))
-			{
-			if (an->cmd < 0)
-				{
+		if (!(an->cmd && !incmdpos && alstat != ALSTAT_MORE)) {
+			if (an->cmd < 0) {
 				tok = DO-an->cmd-1;
 				return 0;
-				}
-			else
-				{
+			} else {
 				an->inuse = 1;
 				hungets(ALPOPS);
 				hungets((alstack[alstackind++] = an)->text);
 				alstat = 0;
 				/* remove from history if it begins with space */
-				if (isset(HISTIGNORESPACE) && an->text[0] == ' ')
-				    remhist();
+				if (isset(HISTIGNORESPACE) && an->text[0] == ' ') remhist();
 				lexstop = 0;
 				return 1;
-				}
 			}
+		}
 	return 0;
 }
 
