@@ -1,33 +1,22 @@
 /*
-
-	loop.c - loop execution
-
-	This file is part of zsh, the Z shell.
-
-   zsh is free software; no one can prevent you from reading the source
-   code, or giving it to someone else.
-   This file is copyrighted under the GNU General Public License, which
-   can be found in the file called COPYING.
-
-   Copyright (C) 1990, 1991 Paul Falstad
-
-   zsh is distributed in the hope that it will be useful, but
-   WITHOUT ANY WARRANTY.  No author or distributor accepts
-   responsibility to anyone for the consequences of using it or for
-   whether it serves any particular purpose or works at all, unless he
-   says so in writing.  Refer to the GNU General Public License
-   for full details.
-
-   Everyone is granted permission to copy, modify and redistribute
-   zsh, but only under the conditions described in the GNU General Public
-   License.   A copy of this license is supposed to have been given to you
-   along with zsh so you can know your rights and responsibilities.
-   It should be in a file named COPYING.
-
-   Among other things, the copyright notice and this notice must be
-   preserved on all copies.
-
-*/
+ *
+ * loop.c - loop execution
+ *
+ * This file is part of zsh, the Z shell.
+ *
+ * This software is Copyright 1992 by Paul Falstad
+ *
+ * Permission is hereby granted to copy, reproduce, redistribute or otherwise
+ * use this software as long as: there is no monetary profit gained
+ * specifically from the use or reproduction of this software, it is not
+ * sold, rented, traded or otherwise marketed, and this copyright notice is
+ * included prominently in any copy made. 
+ *
+ * The author make no claims as to the fitness or correctness of this software
+ * for any use whatsoever, and it is provided as is. Any use of this software
+ * is at the user's own risk. 
+ *
+ */
 
 #include "zsh.h"
 
@@ -74,6 +63,7 @@ int cj = thisjob;
 		}
 	popheap();
 	thisjob = cj;
+	loops--;
 	return lastval;
 }
 
@@ -86,28 +76,38 @@ char *str,*s;
 Lklist args;
 Lknode n;
 int cj = thisjob,t0;
+FILE *inp;
 
-	loops++;
 	node = cmd->u.forcmd;
 	args = cmd->args;
+	if (!node->inflag) {
+		char **x;
+
+		args = newlist();
+		for (x = pparams; *x; x++)
+			addnode(args,ztrdup(*x));
+	}
 	if (!full(args))
 		return 1;
+	loops++;
 	exiting = 0;
 	pushheap();
+	inp = fdopen(dup((SHTTY==-1)?0:SHTTY),"r");
 	for (;;)
 		{
 		do
 			{
+			int pl;
 			selectlist(args);
-			if (interact && SHTTY != -1 && isset(USEZLE))
-				{
-				int pl;
-
-				str = putprompt(prompt3,&pl);
-				str = zleread(str,NULL,pl);
-				}
-			else
-				str = fgets(zalloc(256),256,bshin);
+			str = putprompt(prompt3,&pl);
+			if (full(bufstack)) str = (char *) getnode(bufstack);
+			else if (interact && SHTTY != -1 && isset(USEZLE)) {
+				str = (char *)zleread(str,NULL,pl);
+			} else {
+				fprintf(stderr,"%s",str);
+				fflush(stderr);
+				str = fgets(zalloc(256),256,inp);
+			}
 			if (!str || errflag)
 				{
 				fprintf(stderr,"\n");
@@ -146,7 +146,9 @@ int cj = thisjob,t0;
 		}
 done:
 	popheap();
+	fclose(inp);
 	thisjob = cj;
+	loops--;
 	return lastval;
 }
  
@@ -157,10 +159,10 @@ List list;
 struct whilecmd *node;
 int cj = thisjob; 
 
-	loops++;
 	node = cmd->u.whilecmd;
 	exiting = 0;
 	pushheap();
+	loops++;
 	for(;;)
 		{
 		list = dupstruct(node->cont);
@@ -185,6 +187,7 @@ int cj = thisjob;
 		}
 	popheap();
 	thisjob = cj;
+	loops--;
 	return lastval;
 }
  
@@ -194,7 +197,6 @@ Cmd cmd;
 List list;
 int cj = thisjob,count;
 
-	loops++;
 	exiting = 0;
 	if (!full(cmd->args) || nextnode(firstnode(cmd->args)))
 		{
@@ -203,6 +205,7 @@ int cj = thisjob,count;
 		}
 	count = atoi(peekfirst(cmd->args));
 	pushheap();
+	loops++;
 	while (count--)
 		{
 		list = dupstruct(cmd->u.list);
@@ -225,6 +228,7 @@ int cj = thisjob,count;
 		}
 	popheap();
 	thisjob = cj;
+	loops--;
 	return lastval;
 }
  
