@@ -1,34 +1,22 @@
 /*
-
-	math.c - mathematical expression evaluation
-
-	This file is part of zsh, the Z shell.
-
-	zsh is free software; no one can prevent you from reading the source
-   code, or giving it to someone else.
-
-   This file is copyrighted under the GNU General Public License, which
-   can be found in the file called COPYING.
-
-   Copyright (C) 1990, 1991 Paul Falstad
-
-   zsh is distributed in the hope that it will be useful, but
-   WITHOUT ANY WARRANTY.  No author or distributor accepts
-   responsibility to anyone for the consequences of using it or for
-   whether it serves any particular purpose or works at all, unless he
-   says so in writing.  Refer to the GNU General Public License
-   for full details.
-
-   Everyone is granted permission to copy, modify and redistribute
-   zsh, but only under the conditions described in the GNU General Public
-   License.   A copy of this license is supposed to have been given to you
-   along with zsh so you can know your rights and responsibilities.
-   It should be in a file named COPYING.
-
-   Among other things, the copyright notice and this notice must be
-   preserved on all copies.
-
-*/
+ *
+ * math.c - mathematical expression evaluation
+ *
+ * This file is part of zsh, the Z shell.
+ *
+ * This software is Copyright 1992 by Paul Falstad
+ *
+ * Permission is hereby granted to copy, reproduce, redistribute or otherwise
+ * use this software as long as: there is no monetary profit gained
+ * specifically from the use or reproduction of this software, it is not
+ * sold, rented, traded or otherwise marketed, and this copyright notice is
+ * included prominently in any copy made. 
+ *
+ * The author make no claims as to the fitness or correctness of this software
+ * for any use whatsoever, and it is provided as is. Any use of this software
+ * is at the user's own risk. 
+ *
+ */
 
 #include "zsh.h"
 
@@ -213,7 +201,7 @@ int zzlex() /**/
 			default:
 				if (idigit(*--ptr))
 					{ unary = 0; yyval = zstrtol(ptr,&ptr,10); return NUM; }
-				if (ialpha(*ptr) || *ptr == '$')
+				if (iident(*ptr) || *ptr == '$')
 					{
 					char *p,q;
 
@@ -226,7 +214,7 @@ int zzlex() /**/
 						return EOI;
 						}
 					unary = 0;
-					while(ialpha(*++ptr));
+					while(iident(*++ptr));
 					q = *ptr;
 					*ptr = '\0';
 					lvals[yylval = lvc++] = ztrdup(p);
