@@ -1,34 +1,22 @@
 /*
-
-	params.c - parameters
-
-	This file is part of zsh, the Z shell.
-
-	zsh is free software; no one can prevent you from reading the source
-   code, or giving it to someone else.
-
-   This file is copyrighted under the GNU General Public License, which
-   can be found in the file called COPYING.
-
-   Copyright (C) 1990, 1991 Paul Falstad
-
-   zsh is distributed in the hope that it will be useful, but
-   WITHOUT ANY WARRANTY.  No author or distributor accepts
-   responsibility to anyone for the consequences of using it or for
-   whether it serves any particular purpose or works at all, unless he
-   says so in writing.  Refer to the GNU General Public License
-   for full details.
-
-   Everyone is granted permission to copy, modify and redistribute
-   zsh, but only under the conditions described in the GNU General Public
-   License.   A copy of this license is supposed to have been given to you
-   along with zsh so you can know your rights and responsibilities.
-   It should be in a file named COPYING.
-
-   Among other things, the copyright notice and this notice must be
-   preserved on all copies.
-
-*/
+ *
+ * params.c - parameters
+ *
+ * This file is part of zsh, the Z shell.
+ *
+ * This software is Copyright 1992 by Paul Falstad
+ *
+ * Permission is hereby granted to copy, reproduce, redistribute or otherwise
+ * use this software as long as: there is no monetary profit gained
+ * specifically from the use or reproduction of this software, it is not
+ * sold, rented, traded or otherwise marketed, and this copyright notice is
+ * included prominently in any copy made. 
+ *
+ * The author make no claims as to the fitness or correctness of this software
+ * for any use whatsoever, and it is provided as is. Any use of this software
+ * is at the user's own risk. 
+ *
+ */
 
 #include "zsh.h"
 #include <pwd.h>
@@ -37,238 +25,148 @@
 
 static Param argvparam;
 
+struct iparam {
+	struct hashnode *next; int canfree; char *nam; /* hash data */
+	void *value;
+	int (*func1)(); /* set func */
+	int (*func2)(); /* get func */
+	int ct;				/* output base or field width */
+	int flags;
+	vptr data;			/* used by getfns */
+	char *env;			/* location in environment, if exported */
+	char *ename;		/* name of corresponding environment var */
+	};
+
+#define IFN(X) ((int (*)())(X))
+
 /* put predefined params in hash table */
 
 void setupparams() /**/
 {
-/* special integer params */
-static struct { char *name;
-			long (*get) DCLPROTO((Param));
-			void (*set) DCLPROTO((Param,long)); } x1[] = {
-	"#",poundgetfn,NULL,
-	"ARGC",poundgetfn,NULL,
-	"ERRNO",errnogetfn,NULL,
-	"GID",gidgetfn,NULL,
-	"HISTSIZE",histsizegetfn,histsizesetfn,
-	"LITHISTSIZE",lithistsizegetfn,lithistsizesetfn,
-	"RANDOM",randomgetfn,randomsetfn,
-	"SECONDS",secondsgetfn,secondssetfn,
-	"UID",uidgetfn,NULL,
-	NULL,NULL,NULL
-	}, *p1 = x1;
-/* special string params */
-static struct { char *name;
-			char *(*get) DCLPROTO((Param));
-			void (*set) DCLPROTO((Param,char *)); } x2[] = {
-	"-",dashgetfn,NULL,
-	"HISTCHARS",histcharsgetfn,histcharssetfn,
-	"HOME",homegetfn,homesetfn,
-	"TERM",termgetfn,termsetfn,
-	"WORDCHARS",wordcharsgetfn,wordcharssetfn,
-	"IFS",ifsgetfn,ifssetfn,
-	"_",underscoregetfn,NULL,
-	NULL,NULL,NULL
-	}, *p2 = x2;
-/* constant string params */
-static struct { char *name,*data; } x3[] = {
-	"HOSTTYPE",HOSTTYPE,
-	"VERSION",VERSIONSTR,
-	NULL,NULL
-	}, *p3 = x3;
-/* variable integer params */
-static struct { char *name; long *data; } x4[] = {
-	"!",&lastpid,				/* read only */
-	"$",&mypid,
-	"?",&lastval,
-	"status",&lastval,
-	"LINENO",&lineno,
-	"PPID",&ppid,
-	NULL,NULL,
+static struct iparam pinit[] = {
+#define IPDEF1(A,B,C,D) {NULL,0,A,NULL,IFN(C),IFN(B),10,\
+		PMFLAG_i|PMFLAG_SPECIAL|D,NULL,NULL,NULL}
+	IPDEF1("#",poundgetfn,IFN(nullsetfn),PMFLAG_r),
+	IPDEF1("ARGC",poundgetfn,IFN(nullsetfn),PMFLAG_r),
+	IPDEF1("ERRNO",errnogetfn,IFN(nullsetfn),PMFLAG_r),
+	IPDEF1("GID",gidgetfn,IFN(nullsetfn),PMFLAG_r),
+	IPDEF1("HISTSIZE",histsizegetfn,histsizesetfn,0),
+	IPDEF1("LITHISTSIZE",lithistsizegetfn,lithistsizesetfn,0),
+	IPDEF1("RANDOM",randomgetfn,randomsetfn,0),
+	IPDEF1("SECONDS",secondsgetfn,secondssetfn,0),
+	IPDEF1("UID",uidgetfn,IFN(nullsetfn),PMFLAG_r),
 
-	"BAUD",&baud,				/* read/write */
-	"COLUMNS",&columns,
-	"DIRSTACKSIZE",&dirstacksize,
-	"LINES",&lines,
-	"LISTMAX",&listmax,
-	"LOGCHECK",&logcheck,
-	"MAILCHECK",&mailcheck,
-	"OPTIND",&optind,
-	"PERIOD",&period,
-	"SAVEHIST",&savehist,
-	"SHLVL",&shlvl,
-	"TMOUT",&tmout,
-	NULL,NULL
-	}, *p4 = x4;
-/* variable string params */
-static struct { char *name; char **data; } x5[] = {
-	"LOGNAME",&username,		/* read only */
-	"OLDPWD",&oldpwd,
-	"PWD",&cwd,
-	"TTY",&ttystrname,
-	"USERNAME",&username,
-	NULL,NULL,
+#define IPDEF2(A,B,C,D) {NULL,0,A,NULL,IFN(C),IFN(B),0,\
+		PMFLAG_SPECIAL|D,NULL,NULL,NULL}
+	IPDEF2("-",dashgetfn,IFN(nullsetfn),PMFLAG_r),
+	IPDEF2("HISTCHARS",histcharsgetfn,histcharssetfn,0),
+	IPDEF2("HOME",homegetfn,homesetfn,0),
+	IPDEF2("TERM",termgetfn,termsetfn,0),
+	IPDEF2("WORDCHARS",wordcharsgetfn,wordcharssetfn,0),
+	IPDEF2("IFS",ifsgetfn,ifssetfn,0),
+	IPDEF2("_",underscoregetfn,IFN(nullsetfn),PMFLAG_r),
 
-	"FCEDIT",&fceditparam,	/* read/write */
-	"HOST",&hostnam,
-	"OPTARG",&optarg,
-	"MAIL",&mailfile,
-	"NULLCMD",&nullcmd,
-	"prompt",&prompt,
-	"PROMPT",&prompt,
-	"PROMPT2",&prompt2,
-	"PROMPT3",&prompt3,
-	"PROMPT4",&prompt4,
-	"RPROMPT",&rprompt,
-	"PS1",&prompt,
-	"PS2",&prompt2,
-	"PS3",&prompt3,
-	"PS4",&prompt4,
-	"RPS1",&rprompt,
-	"SPROMPT",&sprompt,
-	"TIMEFMT",&timefmt,
-	"TMPPREFIX",&tmpprefix,
-	"WATCHFMT",&watchfmt,
-	"0",&argzero,
-	NULL,NULL
-	}, *p5 = x5;
-/* colonsplit string params */
-static struct { char *name; } x6[] = {
-	"CDPATH","FIGNORE","FPATH","MAILPATH","WATCH","HOSTS",
-	"HOSTCMDS","OPTCMDS","BINDCMDS","VARCMDS",
-	NULL
-	}, *p6 = x6;
-/* variable array params */
-static struct { char *name; char ***data; } x7[] = {
-	"cdpath",&cdpath,
-	"fignore",&fignore,
-	"fpath",&fpath,
-	"mailpath",&mailpath,
-	"watch",&watch,
-	"hosts",&hosts,
-	"hostcmds",&hostcmds,
-	"optcmds",&optcmds,
-	"bindcmds",&bindcmds,
-	"varcmds",&varcmds,
-	"signals",(char ***) &sigptr,
-	"argv",&pparams,
-	"*",&pparams,
-	"@",&pparams,
-	NULL,NULL
-	}, *p7 = x7;
-/* special array params */
-static struct { char *name;
-					char **(*get) DCLPROTO((Param));
-					void (*set) DCLPROTO((Param,char **)); } x8[] = {
-	"path",pathgetfn,pathsetfn,
-	NULL,NULL,NULL
-	}, *p8 = x8;
-Param pm,pm2;
+#define IPDEF3(A,B) {NULL,0,A,NULL,IFN(nullsetfn),IFN(strconstgetfn),0,PMFLAG_r|\
+		PMFLAG_SPECIAL,(vptr)B,NULL,NULL}
+	IPDEF3("HOSTTYPE",HOSTTYPE),
+	IPDEF3("VERSION",VERSIONSTR),
 
-	for (;p1->name;p1++)
-		{
-		new(pm);
-		pm->gets.ifn = p1->get;
-		pm->sets.ifn = p1->set;
-		pm->flags = (p1->set) ? PMFLAG_i|PMFLAG_SPECIAL :
-			PMFLAG_r|PMFLAG_i|PMFLAG_SPECIAL;
-		pm->ct = 10;
-		addhperm(p1->name,pm,paramtab,NULL);
-		}
-	for (;p2->name;p2++)
-		{
-		new(pm);
-		pm->gets.cfn = p2->get;
-		pm->sets.cfn = p2->set;
-		pm->flags = (p2->set) ? PMFLAG_SPECIAL : PMFLAG_r|PMFLAG_SPECIAL;
-		addhperm(p2->name,pm,paramtab,NULL);
-		}
-	for (;p3->name;p3++)
-		{
-		new(pm);
-		pm->gets.cfn = strconstgetfn;
-		pm->flags = PMFLAG_r|PMFLAG_SPECIAL;
-		pm->data = p3->data;
-		addhperm(p3->name,pm,paramtab,NULL);
-		}
-	for (;p4->name;p4++)
-		{
-		new(pm);
-		pm->gets.ifn = intvargetfn;
-		pm->sets.ifn = NULL;
-		pm->data = p4->data;
-		pm->flags = PMFLAG_r|PMFLAG_i|PMFLAG_SPECIAL;
-		pm->ct = 10;
-		addhperm(p4->name,pm,paramtab,NULL);
-		}
-	for (p4++;p4->name;p4++)
-		{
-		new(pm);
-		pm->gets.ifn = intvargetfn;
-		pm->sets.ifn = intvarsetfn;
-		pm->data = p4->data;
-		pm->flags = PMFLAG_i|PMFLAG_SPECIAL;
-		pm->ct = 10;
-		addhperm(p4->name,pm,paramtab,NULL);
-		}
-	for (;p5->name;p5++)
-		{
-		new(pm);
-		pm->gets.cfn = strvargetfn;
-		pm->sets.cfn = NULL;
-		pm->data = p5->data;
-		pm->flags = PMFLAG_r|PMFLAG_SPECIAL;
-		addhperm(p5->name,pm,paramtab,NULL);
-		}
-	for (p5++;p5->name;p5++)
-		{
-		new(pm);
-		pm->gets.cfn = strvargetfn;
-		pm->sets.cfn = strvarsetfn;
-		pm->data = p5->data;
-		pm->flags = PMFLAG_SPECIAL;
-		addhperm(p5->name,pm,paramtab,NULL);
-		}
-	for (;p6->name;p6++,p7++)
-		{
-		new(pm);
-		new(pm2);
-		pm->gets.cfn = colonarrgetfn;
-		pm->sets.cfn = colonarrsetfn;
-		pm2->gets.afn = arrvargetfn;
-		pm2->sets.afn = arrvarsetfn;
-		pm->data = p7->data;
-		pm2->data = p7->data;
-		pm->flags = PMFLAG_SPECIAL;
-		pm2->flags = PMFLAG_A|PMFLAG_SPECIAL;
-		pm2->ename = p6->name;
-		addhperm(p6->name,pm,paramtab,NULL);
-		addhperm(p7->name,pm2,paramtab,NULL);
-		}
-	new(pm);
-	pm->gets.cfn = colonarrgetfn;
-	pm->sets.cfn = colonarrsetfn;
-	pm->data = NULL;
-	pm->flags = PMFLAG_SPECIAL;
-	addhperm("PATH",pm,paramtab,NULL);
-	for (;p7->name;p7++)
-		{
-		new(pm);
-		pm->gets.afn = arrvargetfn;
-		pm->sets.afn = arrvarsetfn;
-		pm->data = p7->data;
-		pm->flags = PMFLAG_A|PMFLAG_SPECIAL;
-		if (pm->data == &sigptr)
-			pm->flags |= PMFLAG_r;
-		addhperm(p7->name,pm,paramtab,NULL);
-		}
-	for (;p8->name;p8++)
-		{
-		new(pm);
-		pm->gets.afn = p8->get;
-		pm->sets.afn = p8->set;
-		pm->flags = PMFLAG_A|PMFLAG_SPECIAL;
-		addhperm(p8->name,pm,paramtab,NULL);
-		}
+#define IPDEF4(A,B) {NULL,0,A,NULL,IFN(nullsetfn),IFN(intvargetfn),10,\
+		PMFLAG_r|PMFLAG_i|PMFLAG_SPECIAL,(vptr)B,NULL,NULL}
+	IPDEF4("!",&lastpid),
+	IPDEF4("$",&mypid),
+	IPDEF4("?",&lastval),
+	IPDEF4("status",&lastval),
+	IPDEF4("LINENO",&lineno),
+	IPDEF4("PPID",&ppid),
+
+#define IPDEF5(A,B) {NULL,0,A,NULL,IFN(intvarsetfn),IFN(intvargetfn),10,\
+		PMFLAG_i|PMFLAG_SPECIAL,(vptr)B,NULL,NULL}
+	IPDEF5("BAUD",&baud),
+	IPDEF5("COLUMNS",&columns),
+	IPDEF5("DIRSTACKSIZE",&dirstacksize),
+	IPDEF5("LINES",&lines),
+	IPDEF5("LISTMAX",&listmax),
+	IPDEF5("LOGCHECK",&logcheck),
+	IPDEF5("MAILCHECK",&mailcheck),
+	IPDEF5("OPTIND",&zoptind),
+	IPDEF5("PERIOD",&period),
+	IPDEF5("REPORTTIME",&reporttime),
+	IPDEF5("SAVEHIST",&savehist),
+	IPDEF5("SHLVL",&shlvl),
+	IPDEF5("TMOUT",&tmout),
+
+#define IPDEF6(A,B) {NULL,0,A,NULL,IFN(nullsetfn),IFN(strvargetfn),0,\
+		PMFLAG_r|PMFLAG_SPECIAL,(vptr)B,NULL,NULL}
+	IPDEF6("LOGNAME",&logname),
+	IPDEF6("PWD",&pwd),
+	IPDEF6("TTY",&ttystrname),
+	IPDEF6("USERNAME",&username),
+
+#define IPDEF7(A,B) {NULL,0,A,NULL,IFN(strvarsetfn),IFN(strvargetfn),0,\
+		PMFLAG_SPECIAL,(vptr)B,NULL,NULL}
+	IPDEF7("FCEDIT",&fceditparam),
+	IPDEF7("HOST",&hostnam),
+	IPDEF7("OLDPWD",&oldpwd),
+	IPDEF7("OPTARG",&optarg),
+	IPDEF7("MAIL",&mailfile),
+	IPDEF7("NULLCMD",&nullcmd),
+	IPDEF7("POSTEDIT",&postedit),
+	IPDEF7("prompt",&prompt),
+	IPDEF7("PROMPT",&prompt),
+	IPDEF7("PROMPT2",&prompt2),
+	IPDEF7("PROMPT3",&prompt3),
+	IPDEF7("PROMPT4",&prompt4),
+	IPDEF7("READNULLCMD",&readnullcmd),
+	IPDEF7("RPROMPT",&rprompt),
+	IPDEF7("PS1",&prompt),
+	IPDEF7("PS2",&prompt2),
+	IPDEF7("PS3",&prompt3),
+	IPDEF7("PS4",&prompt4),
+	IPDEF7("RPS1",&rprompt),
+	IPDEF7("SPROMPT",&sprompt),
+	IPDEF7("TIMEFMT",&timefmt),
+	IPDEF7("TMPPREFIX",&tmpprefix),
+	IPDEF7("WATCHFMT",&watchfmt),
+	IPDEF7("0",&argzero),
+
+#define IPDEF8(A,B,C) {NULL,0,A,NULL,IFN(colonarrsetfn),IFN(colonarrgetfn),0,\
+		PMFLAG_SPECIAL,(vptr)C,NULL,B}
+	IPDEF8("CDPATH","cdpath",&cdpath),
+	IPDEF8("FIGNORE","fignore",&fignore),
+	IPDEF8("FPATH","fpath",&fpath),
+	IPDEF8("MAILPATH","mailpath",&mailpath),
+	IPDEF8("MANPATH","manpath",&manpath),
+	IPDEF8("WATCH","watch",&watch),
+	IPDEF8("HOSTS","hosts",&hosts),
+	IPDEF8("PATH",NULL,NULL),
+
+#define IPDEF9(A,B,C) {NULL,0,A,NULL,IFN(arrvarsetfn),IFN(arrvargetfn),0,\
+		PMFLAG_A|PMFLAG_SPECIAL|C,(vptr)B,NULL,NULL}
+	IPDEF9("cdpath",&cdpath,0),
+	IPDEF9("fignore",&fignore,0),
+	IPDEF9("fpath",&fpath,0),
+	IPDEF9("mailpath",&mailpath,0),
+	IPDEF9("manpath",&manpath,0),
+	IPDEF9("watch",&watch,0),
+	IPDEF9("hosts",&hosts,0),
+	IPDEF9("signals",&sigptr,PMFLAG_r),
+	IPDEF9("argv",&pparams,0),
+	IPDEF9("*",&pparams,0),
+	IPDEF9("@",&pparams,0),
+
+#define IPDEF10(A,C,D) {NULL,0,A,NULL,IFN(D),IFN(C),0,\
+		PMFLAG_A|PMFLAG_SPECIAL,NULL,NULL,NULL}
+	IPDEF10("path",pathgetfn,pathsetfn),
+	IPDEF10("hostcmds",nullgetfn,hostcmdssetfn),
+	IPDEF10("optcmds",nullgetfn,optcmdssetfn),
+	IPDEF10("bindcmds",nullgetfn,bindcmdssetfn),
+	IPDEF10("varcmds",nullgetfn,varcmdssetfn),
+	{NULL,}
+	};
+struct iparam *ip;
+
+	for (ip = pinit; ip->nam; ip++) addhperm(ip->nam,ip,paramtab,NULL);
 	argvparam = gethnode("argv",paramtab);
 }
 
@@ -284,26 +182,21 @@ char buf[20];
 	if (isset(ALLEXPORT))
 		flags |= PMFLAG_x;
 	pm->flags = flags;
-	if ((flags & PMTYPE) == PMFLAG_s)
-		{
-		pm->u.str = ztrdup(value);
+	if ((flags & PMTYPE) == PMFLAG_s) {
+		pm->u.str = value;
 		pm->sets.cfn = strsetfn;
 		pm->gets.cfn = strgetfn;
-		}
-	else if ((flags & PMTYPE) == PMFLAG_A)
-		{
+	} else if ((flags & PMTYPE) == PMFLAG_A) {
 		pm->u.arr = value;
 		pm->sets.afn = arrsetfn;
 		pm->gets.afn = arrgetfn;
-		}
-	else
-		{
+	} else {
 		pm->u.val = (value) ? matheval(value) : 0;
 		pm->sets.ifn = intsetfn;
 		pm->gets.ifn = intgetfn;
 		sprintf(buf,"%ld",pm->u.val);
 		value = buf;
-		}
+	}
 	if (flags & PMFLAG_x)
 		pm->env = addenv(name,value);
 	addhnode(ztrdup(name),pm,paramtab,freepm);
@@ -350,7 +243,7 @@ Value v;
 	else return NULL;
 	if (sav = *s) *s = '\0';
 	if (idigit(*t) && *t != '0') {
-		v = (Value) alloc(sizeof *v);
+		v = (Value) hcalloc(sizeof *v);
 		v->pm = argvparam;
 		v->a = v->b = atoi(t)-1;
 		if (sav)
@@ -365,7 +258,7 @@ Value v;
 		*pptr = s;
 		if (!pm)
 			return NULL;
-		v = alloc(sizeof *v);
+		v = hcalloc(sizeof *v);
 		if (pmtype(pm) == PMFLAG_A)
 			v->isarr = isvarat ? -1 : 1;
 		v->pm = pm;
@@ -381,7 +274,7 @@ Value v;
 			if (*t == Outbrack)
 				*t = ']';
 			if ((s[0] == '*' || s[0] == '@')  && s[1] == ']') {
-				v->isarr = (s[0] == '*') ? 1 : -1;
+				if (v->isarr) v->isarr = (s[0] == '*') ? 1 : -1;
 				v->a = 0;
 				v->b = -1;
 				s += 2;
@@ -484,8 +377,10 @@ Value v;char *val;
 {
 char *s;
 
-	if (v->pm->flags & PMFLAG_r)
+	if (v->pm->flags & PMFLAG_r) {
+		free(val);
 		return;
+	}
 	if ((s = v->pm->env) && val)
 		v->pm->env = replenv(v->pm->env,val);
 	switch (pmtype(v->pm)) {
@@ -502,14 +397,14 @@ char *s;
 				}
 				if (v->a > strlen(z)) v->a = strlen(z);
 				if (v->b < 0) v->b += strlen(z);
-				if (v->b <= v->a) v->b = v->a-1;
+				if (v->b < v->a) v->b = v->a;
 				z[v->a] = '\0';
 				y = z+v->b+1;
 				x = zalloc(strlen(z)+strlen(y)+strlen(val)+1);
 				strcpy(x,z);
 				strcat(x,val);
 				strcat(x,y);
-				(v->pm->sets.cfn)(v->pm,z);
+				(v->pm->sets.cfn)(v->pm,x);
 			}
 			if (v->pm->flags & (PMFLAG_L|PMFLAG_R|PMFLAG_Z) && !v->pm->ct)
 				v->pm->ct = strlen(val);
@@ -580,6 +475,18 @@ char buf[20];
 		}
 }
 
+void setintenv(s,val) /**/
+char *s; long val;
+{
+Param pm;
+char buf[20];
+
+	if ((pm = gethnode(s,paramtab)) && pm->env) {
+		sprintf(buf,"%ld",val);
+		pm->env = replenv(pm->env,buf);
+	}
+}
+
 void setarrvalue(v,val) /**/
 Value v;char **val;
 {
@@ -587,7 +494,7 @@ Value v;char **val;
 		return;
 	if (pmtype(v->pm) != PMFLAG_A)
 		{
-		zerr("attempt to assign non-array to array",NULL,0);
+		zerr("attempt to assign array value to non-array",NULL,0);
 		return;
 		}
 	(v->pm->sets.afn)(v->pm,val);
@@ -628,17 +535,26 @@ Value v;
 	return getstrvalue(v);
 }
 
+char **getaparam(s) /**/
+char *s;
+{
+Value v;
+
+	if (!(v = getvalue(&s,0))) return NULL;
+	return getarrvalue(v);
+}
+
 Param setsparam(s,val) /**/
 char *s;char *val;
 {
 Value v;
 char *t = s;
 
-	if (!isident(s))
-		{
+	if (!isident(s)) {
 		zerr("not an identifier: %s",s,0);
+		free(val);
 		return NULL;
-		}
+	}
 	if (!(v = getvalue(&s,1)) || *s)
 		return createparam(t,val,PMFLAG_s);
 	if ((v->pm->flags & PMTYPE) != PMFLAG_s &&
@@ -716,8 +632,10 @@ Param pm;
 			(pm->sets.afn)(pm,mkarray(NULL));
 			break;
 		}
-	if (pmtype(pm) == PMFLAG_s && (pm->flags & PMFLAG_x))
+	if (pmtype(pm) == PMFLAG_s && (pm->flags & PMFLAG_x)) {
 		delenv(pm->env);
+		free(pm->env);
+	}
 	if (!(pm->flags & PMFLAG_SPECIAL))
 		freepm(remhnode(s,paramtab));
 	unsetflag = 0;
@@ -752,6 +670,12 @@ Param pm;
 	return pm->u.str;
 }
 
+void nullsetfn(pm,x) /**/
+Param pm; char *x;
+{
+	free(x);
+}
+
 void arrsetfn(pm,x) /**/
 Param pm;char **x;
 {
@@ -788,13 +712,19 @@ Param pm;
 void strvarsetfn(pm,x) /**/
 Param pm;char *x;
 {
-	*((char **) pm->data) = x;
+char **q = ((char **) pm->data);
+
+	if (*q) free(*q);
+	*q = x;
 }
 
 void strvarnonullsetfn(pm,x) /**/
 Param pm;char *x;
 {
-	*((char **) pm->data) = (x) ? x : ztrdup("");
+char **q = ((char **) pm->data);
+
+	if (*q) free(*q);
+	*q = (x) ? x : ztrdup("");
 }
 
 char *strvargetfn(pm) /**/
@@ -803,8 +733,7 @@ Param pm;
 char *s;
 
 	s = *((char **) pm->data);
-	if (!s)
-		return "";
+	if (!s) return "";
 	return s;
 }
 
@@ -817,34 +746,33 @@ Param pm;
 void colonarrsetfn(pm,x) /**/
 Param pm;char *x;
 {
-char **s,**t,*u;
+char **s,**t,*u,*up;
 
 	s = colonsplit(x);
-	for (t = s; *t; t++)
-		{
-		u = *t;
-		if (*u == '~')
-			*u = Tilde;
-		if (*u == '=')
-			*u = Equals;
-		u = strdup(u);
-		filesub(&u);
-		if (!*u)
-			u = ".";
-		*t = ztrdup(u);
+	free(x);
+	if (pm->data != &fignore)
+		for (t = s; *t; t++) {
+			u = *t;
+			if (*u == '~') *u = Tilde;
+			if (*u == '=') *u = Equals;
+			up = hcalloc(strlen(u)+1);
+			strcpy(up,u);
+			u = up;
+			filesub(&u);
+			if (!*u) u = ".";
+			*t = ztrdup(u);
 		}
-	if (pm->data)
-		{
+	if (pm->data) {
+		freearray(*((char ***) pm->data));
 		*((char ***) pm->data) = s;
 		if (pm->ename)
 			arrfixenv(pm->ename,s);
-		}
-	else
-		{
+	} else {
+		freearray(path);
 		path = s;
 		newcmdnamtab();
 		arrfixenv("PATH",s);
-		}
+	}
 }
 
 char *colonarrgetfn(pm) /**/
@@ -881,13 +809,46 @@ Param pm;
 void pathsetfn(pm,x) /**/
 Param pm;char **x;
 {
-	if (path != x)
-		freearray(path);
+	if (path != x) freearray(path);
 	path = x;
 	newcmdnamtab();
 	arrfixenv("PATH",x);
 }
 
+void hostcmdssetfn(pm,x) /**/
+Param pm;char **x;
+{
+	compctl_process(x,CC_HOSTS|CC_FILES,NULL);
+	freearray(x);
+}
+
+void optcmdssetfn(pm,x) /**/
+Param pm;char **x;
+{
+	compctl_process(x,CC_OPTIONS,NULL);
+	freearray(x);
+}
+
+void bindcmdssetfn(pm,x) /**/
+Param pm;char **x;
+{
+	compctl_process(x,CC_BINDINGS,NULL);
+	freearray(x);
+}
+
+void varcmdssetfn(pm,x) /**/
+Param pm;char **x;
+{
+	compctl_process(x,CC_VARS,NULL);
+	freearray(x);
+}
+
+char **nullgetfn(pm) /**/
+Param pm;
+{
+static char *nl = NULL; return &nl;
+}
+
 void unsettablesetfn(pm,x) /**/
 Param pm;char *x;
 { ; }
@@ -919,7 +880,7 @@ Param pm;
 void secondssetfn(pm,x) /**/
 Param pm;long x;
 {
-	shtimer = x+time(NULL);
+	shtimer = time(NULL)-x;
 }
 
 long uidgetfn(pm) /**/
@@ -967,15 +928,15 @@ Param pm;
 void ifssetfn(pm,x) /**/
 Param pm;char *x;
 {
-	if (x) ifs = x;
+	if (x) { free(ifs); ifs = x; }
 	inittyptab();
 }
 
 void histsizesetfn(pm,v) /**/
 Param pm;long v;
 {
-	if ((histsiz = v) <= 2)
-		histsiz = 2;
+	if ((histsiz = v) <= 2) histsiz = 2;
+	resizehistents();
 }
 
 long histsizegetfn(pm) /**/
@@ -987,8 +948,8 @@ Param pm;
 void lithistsizesetfn(pm,v) /**/
 Param pm;long v;
 {
-	if ((lithistsiz = v) <= 2)
-		lithistsiz = 2;
+	if ((lithistsiz = v) <= 2) lithistsiz = 2;
+	resizehistents();
 }
 
 long lithistsizegetfn(pm) /**/
@@ -1037,22 +998,15 @@ int t0;
 	return buf;
 }
 
-char *ttygetfn(pm) /**/
-Param pm;
-{
-	return ttyname(SHTTY);
-}
-
 void histcharssetfn(pm,x) /**/
 Param pm;char *x;
 {
-	if (x)
-		{
+	if (x) {
 		bangchar = x[0];
 		hatchar = (bangchar) ? x[1] : '\0';
 		hashchar = (hatchar) ? x[2] : '\0';
 		free(x);
-		}
+	}
 }
 
 char *histcharsgetfn(pm) /**/
@@ -1076,10 +1030,9 @@ Param pm;
 void homesetfn(pm,x) /**/
 Param pm;char *x;
 {
-	if (isset(CHASELINKS) && (home = xsymlink(x)))
-		free(x);
-	else
-		home = x;
+	free(home);
+	if (isset(CHASELINKS) && (home = xsymlink(x))) free(x);
+	else home = x;
 }
 
 char *wordcharsgetfn(pm) /**/
@@ -1091,10 +1044,9 @@ Param pm;
 void wordcharssetfn(pm,x) /**/
 Param pm;char *x;
 {
-	if (x)
-		wordchars = x;
-	else
-		wordchars = ztrdup(DEFWORDCHARS);
+	free(wordchars);
+	if (x) wordchars = x;
+	else wordchars = ztrdup(DEFWORDCHARS);
 	inittyptab();
 }
 
@@ -1122,8 +1074,7 @@ Param pm;
 void termsetfn(pm,x) /**/
 Param pm;char *x;
 {
-	if (term)
-		free(term);
+	if (term) free(term);
 	term = x;
 	if (!interact || unset(USEZLE))
 		return;
@@ -1174,6 +1125,7 @@ Param pm;char *x;
 		if (tccan(TCDOWN) && tcstr[TCDOWN][0] == '\n')
 			{
 			tclen[TCDOWN] = 0;
+			free(tcstr[TCDOWN]);
 			tcstr[TCDOWN] = NULL;
 			}
 
@@ -1208,28 +1160,31 @@ int ct;
 
 			*str = '\0';
 			if (isident(*envp))
-				pm = setsparam(iname = ztrdup(*envp),ztrdup(str+1));
-			*str = '=';
+				pm = setsparam(iname = *envp,ztrdup(str+1));
 			if (pm) {
 				pm->flags |= PMFLAG_x;
 				pm->env = *envp2;
 				if (pm->flags & PMFLAG_SPECIAL)
 					pm->env = replenv(pm->env,getsparam(iname));
 			}
+			*str = '=';
 		}
 	}
 	pm = gethnode("HOME",paramtab);
-	if (!(pm->flags & PMFLAG_x))
-		{
+	if (!(pm->flags & PMFLAG_x)) {
 		pm->flags |= PMFLAG_x;
 		pm->env = addenv("HOME",home);
-		}
+	}
 	pm = gethnode("PWD",paramtab);
-	if (!(pm->flags & PMFLAG_x))
-		{
+	if (!(pm->flags & PMFLAG_x)) {
 		pm->flags |= PMFLAG_x;
-		pm->env = addenv("PWD",cwd);
-		}
+		pm->env = addenv("PWD",pwd);
+	}
+	pm = gethnode("LOGNAME",paramtab);
+	if (!(pm->flags & PMFLAG_x)) {
+		pm->flags |= PMFLAG_x;
+		pm->env = addenv("LOGNAME",logname);
+	}
 	pm = gethnode("SHLVL",paramtab);
 	if (!(pm->flags & PMFLAG_x))
 		pm->flags |= PMFLAG_x;
@@ -1259,14 +1214,11 @@ char **ep;
 int sl = strlen(s);
 
 	for (ep = environ; *ep; ep++)
-		if (!strncmp(*ep,s,sl) && (*ep)[sl] == '=')
-			{
+		if (!strncmp(*ep,s,sl) && (*ep)[sl] == '=') {
 			char *u = colonjoin(t);
-
 			replenv(*ep,u);
-			free(u);
 			break;
-			}
+		}
 }
 
 char *replenv(e,value) /**/
