@@ -1,34 +1,22 @@
 /*
-
-	parse.c - parser
-
-	This file is part of zsh, the Z shell.
-
-	zsh is free software; no one can prevent you from reading the source
-   code, or giving it to someone else.
-
-   This file is copyrighted under the GNU General Public License, which
-   can be found in the file called COPYING.
-
-   Copyright (C) 1990, 1991 Paul Falstad
-
-   zsh is distributed in the hope that it will be useful, but
-   WITHOUT ANY WARRANTY.  No author or distributor accepts
-   responsibility to anyone for the consequences of using it or for
-   whether it serves any particular purpose or works at all, unless he
-   says so in writing.  Refer to the GNU General Public License
-   for full details.
-
-   Everyone is granted permission to copy, modify and redistribute
-   zsh, but only under the conditions described in the GNU General Public
-   License.   A copy of this license is supposed to have been given to you
-   along with zsh so you can know your rights and responsibilities.
-   It should be in a file named COPYING.
-
-   Among other things, the copyright notice and this notice must be
-   preserved on all copies.
-
-*/
+ *
+ * parse.c - parser
+ *
+ * This file is part of zsh, the Z shell.
+ *
+ * This software is Copyright 1992 by Paul Falstad
+ *
+ * Permission is hereby granted to copy, reproduce, redistribute or otherwise
+ * use this software as long as: there is no monetary profit gained
+ * specifically from the use or reproduction of this software, it is not
+ * sold, rented, traded or otherwise marketed, and this copyright notice is
+ * included prominently in any copy made. 
+ *
+ * The author make no claims as to the fitness or correctness of this software
+ * for any use whatsoever, and it is provided as is. Any use of this software
+ * is at the user's own risk. 
+ *
+ */
 
 #include "zsh.h"
 #include "funcs.h"
@@ -66,7 +54,7 @@ Sublist sl;
 List l = NULL;
 
 	while (tok == SEPER) {
-		if (isnewlin) return NULL;
+		if (isnewlin > 0) return NULL;
 		yylex();
 	}
 	if (tok == ENDINPUT) return NULL;
@@ -77,7 +65,7 @@ List l = NULL;
 		} else if (tok == SEPER) {
 			l = make_list();
 			l->type = SYNC; l->left = sl;
-			if (!isnewlin) yylex();
+			if (isnewlin <= 0) yylex();
 		} else if (tok == AMPER) {
 			l = make_list();
 			l->type = ASYNC; l->left = sl;
@@ -88,7 +76,7 @@ List l = NULL;
 		if (errflag) { yyerror(); return NULL; }
 		yyerror();
 		errflag = 0;
-		if (!isnewlin) {
+		if (isnewlin <= 0) {
 			int c;
 
 			hwbegin();
@@ -180,7 +168,7 @@ Pline p;
 	sl = make_sublist();
 	if (tok == COPROC) { sl->flags |= PFLAG_COPROC; yylex(); }
 	else if (tok == BANG) { sl->flags |= PFLAG_NOT; yylex(); }
-	if (!(p = par_pline()))
+	if (!(p = par_pline()) && !sl->flags)
 		return NULL;
 	sl->left = p;
 	return sl;
@@ -386,7 +374,7 @@ int brflag;
 }
 
 /*
- * if	: { ( IF | ELIF | ELSE IF ) { SEPER } ( INPAR list OUTPAR | list )
+ * if	: { ( IF | ELIF ) { SEPER } ( INPAR list OUTPAR | list )
 			{ SEPER } ( THEN list | INBRACE list OUTBRACE | list1 ) }
 			[ FI | ELSE list FI | ELSE { SEPER } INBRACE list OUTBRACE ]
 			(you get the idea...?)
@@ -402,10 +390,6 @@ int xtok;
 	for (;;) {
 		xtok = tok;
 		yylex();
-		if (xtok == ELSE && tok == IF) {
-			yylex();
-			xtok = ELIF;
-		}
 		if (xtok == FI)
 			break;
 		while (tok == SEPER) yylex();
@@ -545,9 +529,18 @@ Cmd c;
 void par_funcdef(c) /**/
 Cmd c;
 {
+	nocorrect = 1;
+	incmdpos = 0;
 	yylex();
 	c->type = FUNCDEF;
-	c->args = par_wordlist();
+	c->args = newlist();
+	while (tok == STRING) {
+		if (*tokstr == Inbrace && !tokstr[1]) { tok = INBRACE; break; }
+		addnode(c->args,tokstr);
+		yylex();
+	}
+	nocorrect = 0;
+	incmdpos = 1;
 	if (tok == INOUTPAR) yylex();
 	while (tok == SEPER) yylex();
 	if (tok == INBRACE) {
@@ -585,6 +578,7 @@ Cmd c;
 	c->u.cond = par_cond();
 	if (tok != DOUTBRACK) YYERRORV;
 	incond = 0;
+	incmdpos = 1;
 	yylex();
 }
 
@@ -770,16 +764,19 @@ Lklist l;
 char *toks;
 struct redir *fn = allocnode(N_REDIR);
 int mergerror = 0;
-int oldcmdpos;
+int oldcmdpos,oldnc;
 
 	oldcmdpos = incmdpos;
 	incmdpos = 0;
+	oldnc = nocorrect;
+	if (tok != INANG) nocorrect = 1;
 	fn->type = redirtab[tok-OUTANG];
 	fn->fd1 = tokfd;
 	yylex();
 	if (tok != STRING && tok != ENVSTRING) YYERRORV;
 	toks = tokstr;
 	incmdpos = oldcmdpos;
+	nocorrect = oldnc;
 	yylex();
 
 /* assign default fd */
@@ -790,23 +787,43 @@ int oldcmdpos;
 /* > >(...) or < <(...) */
 
 	if ((*toks == Inang || *toks == Outang) && toks[1] == Inpar) {
-		if (fn->type == WRITE)
-			fn->type = OUTPIPE;
-		else if (fn->type == READ)
-			fn->type = INPIPE;
-		else
-			YYERRORV;
+		if ((fn->type & ~1) == WRITE) fn->type = OUTPIPE;
+		else if (fn->type == READ) fn->type = INPIPE;
+		else YYERRORV;
 		fn->name = toks;
 
 /* <<[-] name */
 
 	} else if (fn->type == HEREDOC || fn->type == HEREDOCDASH) {
+		char tbuf[256], *tlin;
+		int tsiz = 0, l;
+		/* Save the rest of the current line for later tokenization */
+		if (!isnewlin) {
+			while (hgets(tbuf, 256) != NULL) {
+				l = strlen(tbuf);
+				if (tsiz == 0) {
+					tlin = ztrdup(tbuf); /* Test for failure? */
+					tsiz = l;
+				} else {
+					tlin = realloc(tlin,tsiz+l+1); /* Test for failure? */
+					strcpy(&tlin[tsiz], tbuf);
+					tsiz += l;
+				}
+				if (tbuf[l-1] == '\n') break;
+			}
+		}
+		/* Now grab the document */
 		fn->name = gethere(toks,fn->type);
 		fn->type = HERESTR;
+		/* Put back the saved line to resume tokenizing */
+		if (tsiz > 0) {
+			hungets(tlin);
+			free(tlin);
+		}
 
 /* >& name or >>& name */
 
-	} else if (IS_ERROR_REDIR(fn->type) && getfdstr(toks) == -1) {
+	} else if (IS_ERROR_REDIR(fn->type) && getfdstr(toks) == FD_WORD) {
 		mergerror = 1;
 		fn->name = toks;
 		fn->type = UN_ERROR_REDIR(fn->type);
@@ -818,25 +835,10 @@ int oldcmdpos;
 
 /* >& # */
 
-	} else if (fn->type == MERGEOUT) {
-		struct redir *fe = allocnode(N_REDIR);
-
-		fn->type = CLOSE;
-		addnode(l,fn);
-		fe->fd1 = fn->fd1;
-		fe->fd2 = getfdstr(toks);
-		if (fe->fd2 == -2)
-			fe->fd2 = coprocout;
-		fe->type = MERGEOUT;
-		fn = fe;
 	} else if (fn->type == MERGE || fn->type == MERGEOUT) {
-		if (*toks == '-')
-			fn->type = CLOSE;
-		else {
-			fn->fd2 = getfdstr(toks);
-			if (fn->fd2 == -2)
-				fn->fd2 = (fn->type == MERGEOUT) ? coprocout : coprocin;
-		}
+		fn->fd2 = getfdstr(toks);
+		if (fn->fd2 == FD_CLOSE) fn->type = CLOSE;
+		else if (fn->fd2 == FD_WORD) fn->fd2 = (fn->type==MERGEOUT) ? 1 : 0;
 	} else
 		fn->name = toks;
 	addnode(l,fn);
@@ -887,13 +889,11 @@ Lklist l;
 int getfdstr(s) /**/
 char *s;
 {
-	if (s[1])
-		return -1;
-	if (idigit(*s))
-		return *s-'0';
-	if (*s == 'p')
-		return -2;
-	return -1;
+	if (s[1]) return FD_WORD;
+	if (idigit(*s)) return *s-'0';
+	if (*s == 'p') return FD_COPROC;
+	if (*s == '-') return FD_CLOSE;
+	return FD_WORD;
 }
 
 Cond par_cond_double(a,b) /**/
