@@ -1,7 +1,7 @@
 /* this file is created automatically by buildzsh */
 /* if all this is wrong, blame csh ;-) */
 
-#define SIGCOUNT      39 
+#define SIGCOUNT       31
 
 #ifdef GLOBALS
 
@@ -12,7 +12,7 @@ char *sigmsg[SIGCOUNT+2] = {
 	"quit",
 	"illegal instruction",
 	"trace trap",
-	"IOT instruction",
+	"abort",
 	"EMT instruction",
 	"floating point exception",
 	"killed",
@@ -48,20 +48,12 @@ char *sigmsg[SIGCOUNT+2] = {
 	"SIGIO",
 	"cpu limit exceeded",
 	"filesize limit exceeded",
-	"SIGMSG",
+	"virtual time alarm",
+	"SIGPROF",
 	"SIGWINCH",
-	"power fail",
+	"SIGLOST",
 	"SIGUSR1",
 	"SIGUSR2",
-	"SIGPROF",
-	"SIGDANGER",
-	"virtual time alarm",
-	"SIGMIGRATE",
-	"SIGPRE",
-	"SIGGRANT",
-	"SIGRETRACT",
-	"SIGSOUND",
-	"SIGSAK",
 	NULL
 };
 
@@ -72,7 +64,7 @@ char *sigs[SIGCOUNT+4] = {
 	"QUIT",
 	"ILL",
 	"TRAP",
-	"IOT",
+	"ABRT",
 	"EMT",
 	"FPE",
 	"KILL",
@@ -92,20 +84,12 @@ char *sigs[SIGCOUNT+4] = {
 	"IO",
 	"XCPU",
 	"XFSZ",
-	"MSG",
+	"VTALRM",
+	"PROF",
 	"WINCH",
-	"PWR",
+	"LOST",
 	"USR1",
 	"USR2",
-	"PROF",
-	"DANGER",
-	"VTALRM",
-	"MIGRATE",
-	"PRE",
-	"GRANT",
-	"RETRACT",
-	"SOUND",
-	"SAK",
 	"ERR",
 	"DEBUG",
 	NULL
