@@ -1,34 +1,22 @@
 /*
-
-	subst.c - various substitutions
-
-	This file is part of zsh, the Z shell.
-
-	zsh is free software; no one can prevent you from reading the source
-   code, or giving it to someone else.
-
-   This file is copyrighted under the GNU General Public License, which
-   can be found in the file called COPYING.
-
-   Copyright (C) 1990, 1991 Paul Falstad
-
-   zsh is distributed in the hope that it will be useful, but
-   WITHOUT ANY WARRANTY.  No author or distributor accepts
-   responsibility to anyone for the consequences of using it or for
-   whether it serves any particular purpose or works at all, unless he
-   says so in writing.  Refer to the GNU General Public License
-   for full details.
-
-   Everyone is granted permission to copy, modify and redistribute
-   zsh, but only under the conditions described in the GNU General Public
-   License.   A copy of this license is supposed to have been given to you
-   along with zsh so you can know your rights and responsibilities.
-   It should be in a file named COPYING.
-
-   Among other things, the copyright notice and this notice must be
-   preserved on all copies.
-
-*/
+ *
+ * subst.c - various substitutions
+ *
+ * This file is part of zsh, the Z shell.
+ *
+ * This software is Copyright 1992 by Paul Falstad
+ *
+ * Permission is hereby granted to copy, reproduce, redistribute or otherwise
+ * use this software as long as: there is no monetary profit gained
+ * specifically from the use or reproduction of this software, it is not
+ * sold, rented, traded or otherwise marketed, and this copyright notice is
+ * included prominently in any copy made. 
+ *
+ * The author make no claims as to the fitness or correctness of this software
+ * for any use whatsoever, and it is provided as is. Any use of this software
+ * is at the user's own risk. 
+ *
+ */
 
 #include "zsh.h"
 #include <pwd.h>
@@ -46,8 +34,8 @@ int qt;
 		char *str,*str3;
 		
 		str = str3 = getdata(node);
-		if (str[1] == Inpar && (*str == Inang ||
-				*str == Outang || *str == Equals))
+		if ((*str == Inang || *str == Outang || *str == Equals) &&
+				str[1] == Inpar)
 			{
 			if (*str == Inang)
 				setdata(node,getoutproc(str+2));		/* <(...) */
@@ -69,6 +57,8 @@ int qt;
 						{
 						arithsubst((vptr*) &str,&str3);	/* $[...] */
 						setdata(node,str3);
+						str = str3;
+						continue;
 						}
 					else
 						{
@@ -204,11 +194,11 @@ char *str = *namptr,*cnam;
 		{
 		if (str[1] == '+' && (str[2] == '/' || str[2] == '\0'))
 			{
-			char *foo = strdup(cwd);	/* ~+ */
+			char *foo = strdup(pwd);	/* ~+ */
 
 			str+=2;
 			modify(&foo,&str);
-			*namptr = dyncat(cwd,str);
+			*namptr = dyncat(pwd,str);
 			return 1;
 			}
 		else if (str[1] == '-' && (str[2] == '/' || str[2] == '\0'))
@@ -218,7 +208,7 @@ char *str = *namptr,*cnam;
 			if (cnam = oldpwd)
 				foo = cnam;
 			else
-				foo = cwd;
+				foo = pwd;
 			str += 2;
 			foo = strdup(foo);
 			modify(&foo,&str);
@@ -318,7 +308,10 @@ struct passwd *pw;
 		user[len] = sav;
 		return str;
 		}
-	if (!(pw = getpwnam(user))) {
+	permalloc(); /* fixes iris bug--getpwnam calls strdup! */
+	pw = getpwnam(user);
+	lastalloc();
+	if (!pw) {
 		user[len] = sav;
 		return NULL;
 	}
@@ -390,13 +383,11 @@ Value v;
 			*s != '!' && *s != '$' && *s != String && *s != Qstring &&
 			*s != '?' && *s != Quest && *s != '_' &&
 			*s != '*' && *s != Star && *s != '@' && *s != '{' &&
-			*s != Inbrace && *s != '=' && *s != Hat && *s != '^')
-		{
+			*s != Inbrace && *s != '=' && *s != Hat && *s != '^') {
 		s[-1] = '$';
 		return;
-		}
-	if (brs = (*s == '{' || *s == Inbrace))
-		s++;
+	}
+	if (brs = (*s == '{' || *s == Inbrace)) s++;
 	for (;;)
 		if (*s == '^' || *s == Hat)
 			plan9 ^= 1,s++;
@@ -408,20 +399,16 @@ Value v;
 			break;
 
 	idbeg = s;
-	if (!(v = getvalue(&s,1)))
-		{
+	if (!(v = getvalue(&s,1))) {
 		vunset = 1;
 		idend = s;
-		}
-	else
+	} else
 		if (isarr = v->isarr)
 			aval = getarrvalue(v);
-		else
-			{
+		else {
 			val = getstrvalue(v);
 			fwidth = v->pm->ct;
-			switch (v->pm->flags & (PMFLAG_L | PMFLAG_R | PMFLAG_Z))
-				{
+			switch (v->pm->flags & (PMFLAG_L | PMFLAG_R | PMFLAG_Z)) {
 				char *t;
 				int t0;
 
@@ -457,8 +444,7 @@ Value v;
 					}
 					break;
 				}
-			switch (v->pm->flags & (PMFLAG_l | PMFLAG_u))
-				{
+			switch (v->pm->flags & (PMFLAG_l | PMFLAG_u)) {
 				char *t;
 
 				case PMFLAG_l:
@@ -471,34 +457,28 @@ Value v;
 					for (;*t;t++)
 						*t = tuupper(*t);
 					break;
-				}
 			}
-	if (colf = *s == ':')
-		s++;
+		}
+	if (colf = *s == ':') s++;
 	
 	/* check for ${..?...} or ${..=..} or one of those.  Only works
 		if the name is in braces. */
 
 	if (brs && (*s == '-' || *s == '=' || *s == '?' || *s == '+' || *s == '#' ||
-			*s == '%' || *s == Quest || *s == Pound))
-		{
-		if (v && v->isarr && (*s == '%' || *s == '#' || *s == Pound))
-			{
+			*s == '%' || *s == Quest || *s == Pound)) {
+		if (v && v->isarr && (*s == '%' || *s == '#' || *s == Pound)) {
 			zerr("operator requires a scalar",NULL,0);
 			return;
-			}
-		if (*s == s[1])
-			{
+		}
+		if (*s == s[1]) {
 			s++;
 			doub = 1;
-			}
+		}
 		u = ++s;
-		if (brs)
-			{
+		if (brs) {
 			int bct = 1;
 
-			for (;;)
-				{
+			for (;;) {
 				if (*s == '{' || *s == Inbrace)
 					bct++;
 				else if (*s == '}' || *s == Outbrace)
@@ -506,42 +486,37 @@ Value v;
 				if (!bct || !*s)
 					break;
 				s++;
-				}
 			}
-		else
-			{
+		} else {
 			while (*s++);
 			s--;
-			}
-		*s++ = '\0';
+		}
+		if (*s) *s++ = '\0';
 		if (colf && !vunset)
 			vunset = (isarr) ? !*aval : !*val;
-		switch (u[-1])
-			{
+		switch ((int)(unsigned char)u[-1]) {
 			case '-':
 				if (vunset)
 					val = strdup(u), isarr = 0;
 				break;
 			case '=':
-				if (vunset)
-					{
+				if (vunset) {
 					char sav = *idend;
 
 					*idend = '\0';
 					setsparam(idbeg,ztrdup(val = strdup(u)));
 					*idend = sav;
 					isarr = 0;
-					}
+				}
 				break;
 			case '?':
-			case Quest:
-				if (vunset)
-					{
+			case (int)(unsigned char)Quest:
+				if (vunset) {
 					zerr("%s",(*u) ? u : "parameter not set",0);
 					if (!interact)
 						exit(1);
 					return;
-					}
+				}
 				break;
 			case '+':
 				if (vunset)
@@ -551,7 +526,7 @@ Value v;
 				isarr = 0;
 				break;
 			case '#':
-			case Pound:
+			case (int)(unsigned char)Pound:
 				if (vunset)
 					val = strdup("");
 				singsub(&u);
@@ -563,100 +538,84 @@ Value v;
 				singsub(&u);
 				getmatch(&val,u,doub+2);
 				break;
-			}
 		}
-	else		/* no ${...=...} or anything, but possible modifiers. */
-		{
-		if (vunset)
-			{
-			if (isset(NOUNSET))
-				{
+	} else {		/* no ${...=...} or anything, but possible modifiers. */
+		if (vunset) {
+			if (isset(NOUNSET)) {
 				zerr("parameter not set",NULL,0);
 				return;
-				}
+			}
 			val = strdup("");
-			}
-		if (colf)
-			{
+		}
+		if (colf) {
 			s--;
-			if (!isarr)
-				modify(&val,&s);
-			/*else
-				modifyarr(&aval,&s);*/
+			if (!isarr) modify(&val,&s);
+			else {
+				char *ss = s;
+				char **ap = aval;
+				while (*ap) {
+					ss = s;
+					modify(ap,&ss);
+				}
 			}
-		if (brs)
-			{
-			if (*s != '}' && *s != Outbrace)
-				{
+		}
+		if (brs) {
+			if (*s != '}' && *s != Outbrace) {
 				zerr("closing brace expected",NULL,0);
 				errflag = 1;
 				return;
-				}
-			s++;
 			}
+			s++;
 		}
+	}
 	if (errflag)
 		return;
-	if (getlen)
-		{
+	if (getlen) {
 		long len = 0;
 		char buf[14];
 
-		if (isarr)
-			{
+		if (isarr) {
 			char **ctr;
 			for (ctr = aval; *ctr; ctr++,len++);
-			}
-		else
+		} else
 			len = strlen(val);
 		sprintf(buf,"%ld",len);
 		val = strdup(buf);
 		isarr = 0;
-		}
+	}
 	if (isarr)
-		if (!aval || !aval[0])
-			{
+		if (!aval || !aval[0]) {
 			if (isarr < 0)
 				wasnularr = 1;
 			val = strdup("");
 			isarr = 0;
-			}
-		else if (!aval[1])
-			{
+		} else if (!aval[1]) {
 			val = aval[0];
 			isarr = 0;
-			}
-	if (qt)
-		{
-		if (isarr > 0)
-			{
+		}
+	if (qt) {
+		if (isarr > 0) {
 			val = spacejoin(aval);
 			isarr = 0;
-			}
 		}
-	else if (spbreak)
-		{
+	} else if (spbreak) {
 		if (isarr)
 			val = spacejoin(aval);
 		isarr = 1;
 		aval = spacesplit(val);
-		if (!aval || !aval[0])
-			{
+		if (!aval || !aval[0]) {
 			val = strdup("");
 			isarr = 0;
-			}
-		else if (!aval[1])
-			{
+		} else if (!aval[1]) {
 			val = aval[0];
 			isarr = 0;
-			}
+		}
 		/* if only one member, not really an array */
 		if (!aval[1])
 			isarr = 0;
-		}
+	}
 	if (isarr)
-		if (plan9)
-			{
+		if (plan9) {
 			int dlen;
 			char *y;
 
@@ -665,18 +624,15 @@ Value v;
 			strcpy(y,ostr);
 			strcat(y,aval[0]);
 			strcat(y,s);
-			while (*++aval)
-				{
+			while (*++aval) {
 				char *x = ncalloc(dlen+strlen(*aval));
 
 				strcpy(x,ostr);
 				strcat(x,*aval);
 				strcat(x,s);
 				insnode(l,n,x), incnode(n);
-				}
 			}
-		else
-			{
+		} else {
 			char *zz;
 
 			zz = ncalloc((char *) aptr-(bptr)+strlen(aval[0])+1);
@@ -689,15 +645,14 @@ Value v;
 			strcpy(zz,*aval);
 			strcat(zz,s);
 			insnode(l,n,zz);
-			}
-	else
-		{
+		}
+	else {
 		bptr = ncalloc((char *) aptr-bptr+strlen(val)+strlen(s)+1);
 		setdata(n,bptr);
 		strcpy(bptr,ostr);
 		strcat(bptr,val);
 		strcat(bptr,s);
-		}
+	}
 }
 
 /* arithmetic substitution */
@@ -781,7 +736,7 @@ int val;
 Lknode node;
  
 	if ((val < 0 && !firstnode(dirstack)) || !val--)
-		return cwd;
+		return pwd;
 	if (val < 0)
 		node = lastnode(dirstack);
 	else
