@@ -1,34 +1,22 @@
 /*
-
-	table.c - linked lists and hash tables
-
-	This file is part of zsh, the Z shell.
-
-	zsh is free software; no one can prevent you from reading the source
-   code, or giving it to someone else.
-
-   This file is copyrighted under the GNU General Public License, which
-   can be found in the file called COPYING.
-
-   Copyright (C) 1990, 1991 Paul Falstad
-
-   zsh is distributed in the hope that it will be useful, but
-   WITHOUT ANY WARRANTY.  No author or distributor accepts
-   responsibility to anyone for the consequences of using it or for
-   whether it serves any particular purpose or works at all, unless he
-   says so in writing.  Refer to the GNU General Public License
-   for full details.
-
-   Everyone is granted permission to copy, modify and redistribute
-   zsh, but only under the conditions described in the GNU General Public
-   License.   A copy of this license is supposed to have been given to you
-   along with zsh so you can know your rights and responsibilities.
-   It should be in a file named COPYING.
-
-   Among other things, the copyright notice and this notice must be
-   preserved on all copies.
-
-*/
+ *
+ * table.c - linked lists and hash tables
+ *
+ * This file is part of zsh, the Z shell.
+ *
+ * This software is Copyright 1992 by Paul Falstad
+ *
+ * Permission is hereby granted to copy, reproduce, redistribute or otherwise
+ * use this software as long as: there is no monetary profit gained
+ * specifically from the use or reproduction of this software, it is not
+ * sold, rented, traded or otherwise marketed, and this copyright notice is
+ * included prominently in any copy made. 
+ *
+ * The author make no claims as to the fitness or correctness of this software
+ * for any use whatsoever, and it is provided as is. Any use of this software
+ * is at the user's own risk. 
+ *
+ */
 
 #define TABLE_C
 #include "zsh.h"
@@ -65,15 +53,13 @@ char *s;
 {
 unsigned hash = 0,g;
  
-	for (; *s; s++)
-		{
+	for (; *s; s++) {
 		hash = (hash << 4) + *s;
-		if (g = hash & 0xf0000000)
-			{
+		if (g = hash & 0xf0000000) {
 			hash ^= g;
 			hash ^= g >> 24;
-			}
 		}
+	}
 	return hash;
 }
 
@@ -83,32 +69,48 @@ void Addhnode(nam,dat,ht,freefunc,canfree) /**/
 char *nam;vptr dat;Hashtab ht;FFunc freefunc;int canfree;
 {
 int hval = hasher(nam) % ht->hsize;
-struct hashnode *hp = ht->nodes[hval],*hn;
+struct hashnode **hp = ht->nodes+hval,*hn;
  
-	for (; hp; hp = hp->next)
-		if (!strcmp(hp->nam,nam))
-			{
-			if (!freefunc)
-				zerr("attempt to call NULL freefunc",NULL,0);
-			else
-				freefunc(hp->dat);
-			hp->dat = dat;
-			if (hp->canfree)
-				{
-				free(hp->nam);
-				hp->nam = nam;
-				}
-			hp->canfree = canfree;
+	for (; *hp; hp = &(*hp)->next)
+		if (!strcmp((*hp)->nam,nam)) {
+			if ((*hp)->canfree) free((*hp)->nam);
+			hn = dat;
+			hn->next = (*hp)->next;
+			if (!freefunc) zerr("attempt to call NULL freefunc",NULL,0);
+			else freefunc(*hp);
+			*hp = hn;
+			hn->nam = nam;
+			hn->canfree = canfree;
 			return;
-			}
-	hn = (Hashnode) zcalloc(sizeof *hn);
+		}
+	hn = (Hashnode) dat;
 	hn->nam = nam;
-	hn->dat = dat;
 	hn->canfree = canfree;
 	hn->next = ht->nodes[hval];
 	ht->nodes[hval] = hn;
-	if (++ht->ct == ht->hsize*4)
-		expandhtab(ht);
+	if (++ht->ct == ht->hsize*4) expandhtab(ht);
+}
+
+/* add a node to command hash table */
+
+void addhcmdnode(nam,pnam) /**/
+char *nam;char **pnam;
+{
+int hval = hasher(nam) % cmdnamtab->hsize;
+struct hashnode *hp = cmdnamtab->nodes[hval],*hn;
+Cmdnam cc;
+ 
+	for (; hp; hp = hp->next) if (!strcmp(hp->nam,nam)) return;
+	cc = (Cmdnam) zcalloc(sizeof *cc);
+	cc->type = EXCMD;
+	cc->u.nam = tricat(*pnam,"/",nam);
+	cc->pcomp = pnam;
+	hn = (Hashnode) cc;
+	hn->nam = ztrdup(nam);
+	hn->canfree = 1;
+	hn->next = cmdnamtab->nodes[hval];
+	cmdnamtab->nodes[hval] = hn;
+	if (++cmdnamtab->ct == cmdnamtab->hsize*4) expandhtab(cmdnamtab);
 }
 
 /* expand hash tables when they get too many entries */
@@ -116,20 +118,18 @@ struct hashnode *hp = ht->nodes[hval],*hn;
 void expandhtab(ht) /**/
 Hashtab ht;
 {
-struct hashnode *hp,**arr,**ha,*hn;
+struct hashnode **arr,**ha,*hn,*hp;
 int osize = ht->hsize,nsize = osize*8;
 
 	ht->hsize = nsize;
 	arr = ht->nodes;
 	ht->nodes = (Hashnode*) zcalloc(nsize*sizeof(struct hashnode *));
 	for (ha = arr; osize; osize--,ha++)
-		for (hn = *ha; hn; )
-			{
-			Addhnode(hn->nam,hn->dat,ht,NULL,hn->canfree);
+		for (hn = *ha; hn; ) {
 			hp = hn->next;
-			free(hn);
+			Addhnode(hn->nam,(vptr)hn,ht,NULL,hn->canfree);
 			hn = hp;
-			}
+		}
 	free(arr);
 }
 
@@ -141,9 +141,7 @@ char *nam;Hashtab ht;
 int hval = hasher(nam) % ht->hsize;
 struct hashnode *hn = ht->nodes[hval];
  
-	for (; hn; hn = hn->next)
-		if (!strcmp(hn->nam,nam))
-			return hn->dat;
+	for (; hn; hn = hn->next) if (!strcmp(hn->nam,nam)) return (vptr)hn;
 	return NULL;
 }
  
@@ -154,14 +152,12 @@ int val;
 struct hashnode *hn,**hp = &ht->nodes[0],*next;
  
 	for (val = ht->hsize; val; val--,hp++)
-		for (hn = *hp; hn; )
-			{
+		for (hn = *hp; hn; ) {
 			next = hn->next;
-			freefunc(hn->dat);
 			if (hn->canfree) free(hn->nam);
-			free(hn);
+			freefunc(hn);
 			hn = next;
-			}
+		}
 	free(ht->nodes);
 	free(ht);
 }
@@ -173,31 +169,21 @@ char *nam;Hashtab ht;
 {
 int hval = hasher(nam) % ht->hsize;
 struct hashnode *hn = ht->nodes[hval],*hp;
-vptr dat;
 
-	if (!hn)
-		return NULL;
-	if (!strcmp(hn->nam,nam))
-		{
+	if (!hn) return NULL;
+	if (!strcmp(hn->nam,nam)) {
 		ht->nodes[hval] = hn->next;
-		dat = hn->dat;
-		if (hn->canfree)
-			free(hn->nam);
-		free(hn);
+		if (hn->canfree) free(hn->nam);
 		ht->ct--;
-		return dat;
-		}
+		return (vptr)hn;
+	}
 	for (hp = hn, hn = hn->next; hn; hn = (hp = hn)->next)
-		if (!strcmp(hn->nam,nam))
-			{
+		if (!strcmp(hn->nam,nam)) {
 			hp->next = hn->next;
-			dat = hn->dat;
-			if (hn->canfree)
-				free(hn->nam);
-			free(hn);
+			if (hn->canfree) free(hn->nam);
 			ht->ct--;
-			return dat;
-			}
+			return (vptr)hn;
+		}
 	return NULL;
 }
 
@@ -213,26 +199,23 @@ Lknode tmp;
 	llast->next->last = llast;
 	llast->next->dat = dat;
 	llast->next->next = tmp;
-	if (tmp)
-		tmp->last = llast->next;
-	else
-		list->last = llast->next;
+	if (tmp) tmp->last = llast->next;
+	else list->last = llast->next;
 }
 
 void addnodeinorder(x,dat) /**/
 Lklist x; char *dat;
 {
 Lknode y, l = NULL;
+int val = 123;
 
 	for (y = firstnode(x); y; incnode(y)) {
-		if (forstrcmp((char **) &y->dat, &dat) >= 0) 
-			break;
+		if ((val = forstrcmp((char **) &y->dat, &dat)) >= 0) break;
 		l = y;
 	}
-	if (l == NULL)
-		insnode(x, (Lknode) x, dat);
-	else
-		insnode(x, l, dat);
+	if (!val) return;
+	if (l == NULL) insnode(x, (Lknode) x, dat);
+	else insnode(x, l, dat);
 }
 
 
@@ -244,12 +227,10 @@ Lklist list;Lknode nd;
 vptr dat;
 
 	nd->last->next = nd->next;
-	if (nd->next)
-		nd->next->last = nd->last;
-	else
-		list->last = nd->last;
-	free(nd);
+	if (nd->next) nd->next->last = nd->last;
+	else list->last = nd->last;
 	dat = nd->dat;
+	free(nd);
 	return dat;
 }
 
@@ -261,10 +242,8 @@ Lklist list;Lknode nd;
 vptr dat;
 
 	nd->last->next = nd->next;
-	if (nd->next)
-		nd->next->last = nd->last;
-	else
-		list->last = nd->last;
+	if (nd->next) nd->next->last = nd->last;
+	else list->last = nd->last;
 	dat = nd->dat;
 	return dat;
 }
@@ -274,8 +253,7 @@ vptr dat;
 void chuck(str) /**/
 char *str;
 {
-	while (str[0] = str[1])
-		str++;
+	while (str[0] = str[1]) str++;
 }
 
 /* get top node in a linked list */
@@ -322,14 +300,12 @@ Lklist tab;FFunc freefunc;
 {
 Lknode node = tab->first,next;
  
-	while (node)
-		{
+	while (node) {
 		next = node->next;
-		if (freefunc)
-			freefunc(node->dat);
+		if (freefunc) freefunc(node->dat);
 		free(node);
 		node = next;
-		}
+	}
 	free(tab);
 }
  
@@ -338,14 +314,11 @@ char *s;char *t;
 {
 char *p1,*p2;
  
-	for (; *s; s++)
-		{
+	for (; *s; s++) {
 		for (p1 = s, p2 = t; *p2; p1++,p2++)
-			if (*p1 != *p2)
-				break;
-		if (!*p2)
-			 return (char *) s;
-		}
+			if (*p1 != *p2) break;
+		if (!*p2) return (char *) s;
+	}
 	return NULL;
 }
 
@@ -356,15 +329,12 @@ Lklist l;Lknode where;Lklist x;
 {
 Lknode nx = where->next;
 
-	if (!l->first)
-		return;
+	if (!l->first) return;
 	where->next = l->first;
 	l->last->next = nx;
 	l->first->last = where;
-	if (nx)
-		nx->last = l->last;
-	else
-		x->last = l->last;
+	if (nx) nx->last = l->last;
+	else x->last = l->last;
 }
 
 int countnodes(x) /**/
