@@ -1,34 +1,22 @@
 /*
-
-	text.c - textual representations of syntax trees
-
-	This file is part of zsh, the Z shell.
-
-	zsh is free software; no one can prevent you from reading the source
-   code, or giving it to someone else.
-
-   This file is copyrighted under the GNU General Public License, which
-   can be found in the file called COPYING.
-
-   Copyright (C) 1990, 1991 Paul Falstad
-
-   zsh is distributed in the hope that it will be useful, but
-   WITHOUT ANY WARRANTY.  No author or distributor accepts
-   responsibility to anyone for the consequences of using it or for
-   whether it serves any particular purpose or works at all, unless he
-   says so in writing.  Refer to the GNU General Public License
-   for full details.
-
-   Everyone is granted permission to copy, modify and redistribute
-   zsh, but only under the conditions described in the GNU General Public
-   License.   A copy of this license is supposed to have been given to you
-   along with zsh so you can know your rights and responsibilities.
-   It should be in a file named COPYING.
-
-   Among other things, the copyright notice and this notice must be
-   preserved on all copies.
-
-*/
+ *
+ * text.c - textual representations of syntax trees
+ *
+ * This file is part of zsh, the Z shell.
+ *
+ * This software is Copyright 1992 by Paul Falstad
+ *
+ * Permission is hereby granted to copy, reproduce, redistribute or otherwise
+ * use this software as long as: there is no monetary profit gained
+ * specifically from the use or reproduction of this software, it is not
+ * sold, rented, traded or otherwise marketed, and this copyright notice is
+ * included prominently in any copy made. 
+ *
+ * The author make no claims as to the fitness or correctness of this software
+ * for any use whatsoever, and it is provided as is. Any use of this software
+ * is at the user's own risk. 
+ *
+ */
 
 #include "zsh.h"
 
@@ -41,12 +29,12 @@ void taddchr(c) /**/
 int c;
 {
 	*tptr++ = c;
-	if (tptr == tlim)
-		{
+	if (tptr == tlim) {
+		if (!tbuf) { tptr--; return; }
 		tbuf = realloc(tbuf,tsiz *= 2);
 		tlim = tbuf+tsiz;
 		tptr = tbuf+tsiz/2;
-		}
+	}
 }
 
 /* add a string to the text buffer */
@@ -56,14 +44,14 @@ char *s;
 {
 int sl = strlen(s);
 
-	while (tptr+sl >= tlim)
-		{
+	while (tptr+sl >= tlim) {
 		int x = tptr-tbuf;
 
+		if (!tbuf) return;
 		tbuf = realloc(tbuf,tsiz *= 2);
 		tlim = tbuf+tsiz;
 		tptr = tbuf+x;
-		}
+	}
 	strcpy(tptr,s);
 	tptr += sl;
 }
@@ -95,12 +83,12 @@ int t0;
 		taddstr("; ");
 }
 
-/* get a textual representation of n */
+/* get a permanent textual representation of n */
 
-char *gettext(n,nls) /**/
-struct node *n;int nls;
+char *getpermtext(n) /**/
+struct node *n;
 {
-	tnewlins = nls;
+	tnewlins = 1;
 	tbuf = zalloc(tsiz = 32);
 	tptr = tbuf;
 	tlim = tbuf+tsiz;
@@ -111,6 +99,24 @@ struct node *n;int nls;
 	return tbuf;
 }
 
+/* get a representation of n in a job text buffer */
+
+char *getjobtext(n) /**/
+struct node *n;
+{
+static char jbuf[JOBTEXTSIZE];
+
+	tnewlins = 0;
+	tbuf = NULL;
+	tptr = jbuf;
+	tlim = tptr+JOBTEXTSIZE-1;
+	tindent = 1;
+	gettext2(n);
+	*tptr = '\0';
+	untokenize(jbuf);
+	return jbuf;
+}
+
 #define gt2(X) gettext2((struct node *) (X))
 
 /*
@@ -400,8 +406,8 @@ Cmd cmd;
 {
 Lknode n;
 static char *fstr[] = {
-	">",">!",">>",">>!",">&",">&!",">>&",">>&!","<","<<","<&",">&",">&-",
-	">&-","..","..","<<<"
+	">",">!",">>",">>!",">&",">&!",">>&",">>&!","<","<<",
+	"<<-","<<<","<&",">&-","..",".."
 	};
 
 	taddchr(' ');
