@@ -1,33 +1,22 @@
 /*
-
-	utils.c - miscellaneous utilities
-
-	This file is part of zsh, the Z shell.
-
-   zsh is free software; no one can prevent you from reading the source
-   code, or giving it to someone else.
-   This file is copyrighted under the GNU General Public License, which
-   can be found in the file called COPYING.
-
-   Copyright (C) 1990, 1991 Paul Falstad
-
-   zsh is distributed in the hope that it will be useful, but
-   WITHOUT ANY WARRANTY.  No author or distributor accepts
-   responsibility to anyone for the consequences of using it or for
-   whether it serves any particular purpose or works at all, unless he
-   says so in writing.  Refer to the GNU General Public License
-   for full details.
-
-   Everyone is granted permission to copy, modify and redistribute
-   zsh, but only under the conditions described in the GNU General Public
-   License.   A copy of this license is supposed to have been given to you
-   along with zsh so you can know your rights and responsibilities.
-   It should be in a file named COPYING.
-
-   Among other things, the copyright notice and this notice must be
-   preserved on all copies.
-
-*/
+ *
+ * utils.c - miscellaneous utilities
+ *
+ * This file is part of zsh, the Z shell.
+ *
+ * This software is Copyright 1992 by Paul Falstad
+ *
+ * Permission is hereby granted to copy, reproduce, redistribute or otherwise
+ * use this software as long as: there is no monetary profit gained
+ * specifically from the use or reproduction of this software, it is not
+ * sold, rented, traded or otherwise marketed, and this copyright notice is
+ * included prominently in any copy made. 
+ *
+ * The author make no claims as to the fitness or correctness of this software
+ * for any use whatsoever, and it is provided as is. Any use of this software
+ * is at the user's own risk. 
+ *
+ */
 
 #include "zsh.h"
 #include <pwd.h>
@@ -183,8 +172,8 @@ int c;FILE *f;
 			putc(ztokens[c-Pound],f);
 		return;
 		}
-	c &= 0x7f;
-	if (c >= ' ' && c < 0x7f)
+	c &= 0xff;
+	if (isprint(c))
 		putc(c,f);
 	else if (c == '\n')
 		{
@@ -241,14 +230,14 @@ char *buf;int len;FILE *in;
 
 /* get a symlink-free pathname for s relative to PWD */
 
-char *findcwd(s) /**/
+char *findpwd(s) /**/
 char *s;
 {
 char *t;
 
 	if (*s == '/')
 		return xsymlink(s);
-	s = tricat((cwd[1]) ? cwd : "","/",s);
+	s = tricat((pwd[1]) ? pwd : "","/",s);
 	t = xsymlink(s);
 	free(s);
 	return t;
@@ -256,8 +245,8 @@ char *t;
 
 static char xbuf[MAXPATHLEN];
 
-/* get rid of this */
-char *fixcwd(s) /**/
+#if 0
+char *fixpwd(s) /**/
 char *s;
 {
 struct stat sbuf,tbuf;
@@ -267,7 +256,7 @@ char *t;
 	if (*s == '/')
 		t = ztrdup(s);
 	else
-		t = tricat((cwd[1]) ? cwd : "","/",s);
+		t = tricat((pwd[1]) ? pwd : "","/",s);
 	(void) xsymlinks(t+1,0); 
 	free(t);
 	if (!*xbuf)
@@ -277,6 +266,7 @@ char *t;
 			chdir(xbuf);
 	return ztrdup(xbuf);
 }
+#endif
 
 int ispwd(s) /**/
 char *s;
@@ -409,8 +399,8 @@ int t0;
 
 /* print a directory */
 
-void printdir(s) /**/
-char *s;
+void fprintdir(s, f) /**/
+char *s; FILE *f;
 {
 int t0;
 
@@ -419,20 +409,33 @@ int t0;
 		{
 		if (!strncmp(s,home,t0 = strlen(home)) && t0 > 1)
 			{
-			putchar('~');
-			fputs(s+t0,stdout);
+			putc('~', f);
+			fputs(s+t0,f);
 			}
 		else
-			fputs(s,stdout);
+			fputs(s,f);
 		}
 	else
 		{
-		putchar('~');
-		fputs(usernames[t0],stdout);
-		fputs(s+strlen(userdirs[t0]),stdout);
+		putc('~', f);
+		fputs(usernames[t0],f);
+		fputs(s+strlen(userdirs[t0]),f);
 		}
 }
 
+void printdir(s) /**/
+char *s;
+{
+	fprintdir(s, stdout);
+}
+
+void printdircr(s) /**/
+char *s;
+{
+	fprintdir(s, stdout);
+	putchar('\n');
+}
+
 /* see if a path has a named directory as its prefix */
 
 int finddir(s) /**/
@@ -541,16 +544,14 @@ struct schedcmd *sch,*schl;
 		scanjobs();
 	if (errflag)
 		return;
-	if (list = getshfunc("precmd"))
-		newrunlist(list);
+	if (list = getshfunc("precmd")) doshfuncnoval(list,NULL,0);
 	if (errflag)
 		return;
 	if (period && (time(NULL) > lastperiod+period) &&
-			(list = getshfunc("periodic")))
-		{
-		newrunlist(list);
+			(list = getshfunc("periodic"))) {
+		doshfuncnoval(list,NULL,0);
 		lastperiod = time(NULL);
-		}
+	}
 	if (errflag)
 		return;
 	if (watch)
@@ -626,7 +627,7 @@ char *v,*u,c;
 		else if (S_ISDIR(st.st_mode))
 			{
 			Lklist l;
-			DIR *lock = opendir(s);
+			DIR *lock = opendir(*s);
 			char buf[MAXPATHLEN*2],**arr,**ap;
 			struct direct *de;
 			int ct = 1;
@@ -709,57 +710,18 @@ Cmdnam cc;
 
 void newcmdnamtab() /**/
 {
-int t0,dot = 0,pathct;
-struct direct *de;
-DIR *dir;
-Cmdnam cc;
 Hashtab oldcnt;
 
 	oldcnt = cmdnamtab;
 	permalloc();
 	cmdnamtab = newhtable(101);
-	if (pathsuppress) {
-		addbuiltins();
-		if (oldcnt) {
-			listhtable(oldcnt,(HFunc) saveoldfuncs);
-			freehtab(oldcnt,freecmdnam);
-		}
-		lastalloc();
-		return;
-	}
-	holdintr();
-	for (t0 = 0; path[t0]; t0++)
-		if (!strcmp(".",path[t0]))
-			{
-			dot = 1;
-			break;
-			}
-	for (pathct = arrlen(path), t0 = pathct-1; t0 >= 0; t0--)
-		if (!strcmp(".",path[t0]))
-			dot = 0;
-		else if (strncmp("/./",path[t0],3))
-			{
-			dir = opendir(path[t0]);
-			if (!dir)
-				continue;
-			readdir(dir); readdir(dir);
-			while (de = readdir(dir))
-				{
-				cc = (Cmdnam) zcalloc(sizeof *cc);
-				cc->type = (dot) ? EXCMD_POSTDOT : EXCMD_PREDOT;
-				cc->u.nam = tricat(path[t0],"/",de->d_name);
-				addhnode(ztrdup(de->d_name),cc,cmdnamtab,freecmdnam);
-				}
-			closedir(dir);
-			}
 	addbuiltins();
-	if (oldcnt)
-		{
+	if (oldcnt) {
 		listhtable(oldcnt,(HFunc) saveoldfuncs);
 		freehtab(oldcnt,freecmdnam);
-		}
-	noholdintr();
+	}
 	lastalloc();
+	pathchecked = path;
 }
 
 void freecmdnam(a) /**/
@@ -775,6 +737,14 @@ struct cmdnam *c = (struct cmdnam *) a;
 	free(c);
 }
 
+void freecompctl(a) /**/
+vptr a;
+{
+Compctl cc = (Compctl) a;
+
+	free(cc);
+}
+
 void freestr(a) /**/
 vptr a;
 {
@@ -809,15 +779,15 @@ struct ttyinfo *ti;
 	if (SHTTY != -1)
 		{
 #ifdef TERMIOS
-#ifdef __hpux
-		if (tcgetattr(SHTTY,&ti->termios) == -1)
+#ifdef HAS_TCCRAP
+		if (tcgetattr(SHTTY,&ti->tio) == -1)
 #else
-		if (ioctl(SHTTY,TCGETS,&ti->termios) == -1)
+		if (ioctl(SHTTY,TCGETS,&ti->tio) == -1)
 #endif
 			zerr("bad tcgets: %e",NULL,errno);
 #else
 #ifdef TERMIO
-		ioctl(SHTTY,TCGETA,&ti->termio);
+		ioctl(SHTTY,TCGETA,&ti->tio);
 #else
 		ioctl(SHTTY,TIOCGETP,&ti->sgttyb);
 		ioctl(SHTTY,TIOCLGET,&ti->lmodes);
@@ -827,7 +797,7 @@ struct ttyinfo *ti;
 #endif
 #ifdef TIOCGWINSZ
 		if (ioctl(SHTTY,TIOCGWINSZ,&ti->winsize) == -1)
-			zerr("bad tiocgwinsz: %e",NULL,errno);
+		/*	zerr("bad tiocgwinsz: %e",NULL,errno)*/;
 #endif
 		}
 }
@@ -838,15 +808,18 @@ struct ttyinfo *ti;
 	if (SHTTY != -1)
 		{
 #ifdef TERMIOS
-#ifdef __hpux
-		if (tcsetattr(SHTTY, TCSADRAIN, &ti->termios) == -1)
-#else
-		if (ioctl(SHTTY,TCSETS,&ti->termios) == -1)
+#ifdef HAS_TCCRAP
+#ifndef TCSADRAIN
+#define TCSADRAIN 1   /* XXX Princeton's include files are screwed up */
 #endif
-			zerr("settyinfo: %e",NULL,errno);
+		if (tcsetattr(SHTTY, TCSADRAIN, &ti->tio) == -1)
+#else
+		if (ioctl(SHTTY,TCSETS,&ti->tio) == -1)
+#endif
+		/*	zerr("settyinfo: %e",NULL,errno)*/;
 #else
 #ifdef TERMIO
-		ioctl(SHTTY,TCSETA,&ti->termio);
+		ioctl(SHTTY,TCSETA,&ti->tio);
 #else
 		ioctl(SHTTY,TIOCSETN,&ti->sgttyb);
 		ioctl(SHTTY,TIOCLSET,&ti->lmodes);
@@ -857,7 +830,7 @@ struct ttyinfo *ti;
 #ifdef TIOCGWINSZ
 		signal(SIGWINCH,SIG_IGN);
 		if (ioctl(SHTTY,TIOCSWINSZ,&ti->winsize) == -1)
-			zerr("settyinfo: %e",NULL,errno);
+		/*	zerr("settyinfo: %e",NULL,errno)*/;
 		signal(SIGWINCH,handler);
 #endif
 		}
@@ -871,26 +844,31 @@ struct ttyinfo *ti;
 {
 int t0;
 
-#ifdef TERMIOS
-	ti->termios.c_lflag |= ICANON|ECHO;
-	for (t0 = 0; t0 != NCCS; t0++)
-		if (!ti->termios.c_cc[t0] && savedttyinfo.termios.c_cc[t0])
-			ti->termios.c_cc[t0] = savedttyinfo.termios.c_cc[t0];
+#ifdef TIO
+	ti->tio.c_lflag |= ICANON|ECHO;
+#ifdef FLUSHO
+	ti->tio.c_lflag &= ~FLUSHO;
+#endif
+	for (t0 = 0; t0 !=
+#ifdef NCCS
+	NCCS
 #else
-#ifdef TERMIO
-	ti->termio.c_lflag |= ICANON|ECHO;
-	for (t0 = 0; t0 != NCC; t0++)
-		if (!ti->termio.c_cc[t0] && savedttyinfo.termio.c_cc[t0])
-			ti->termio.c_cc[t0] = savedttyinfo.termio.c_cc[t0];
+	NCC
+#endif
+	; t0++)
+		if (ti->tio.c_cc[t0] == VDISABLEVAL &&
+				savedttyinfo.tio.c_cc[t0] != VDISABLEVAL)
+			ti->tio.c_cc[t0] = savedttyinfo.tio.c_cc[t0];
 #else
 	ti->sgttyb.sg_flags = (ti->sgttyb.sg_flags & ~CBREAK) | ECHO;
+	ti->lmodes &= ~LFLUSHO;
 	SANEKEY(tchars.t_quitc);
 	SANEKEY(tchars.t_startc);
 	SANEKEY(tchars.t_stopc);
 	SANEKEY(ltchars.t_suspc);
 	SANEKEY(ltchars.t_dsuspc);
 	SANEKEY(ltchars.t_lnextc);
-#endif
+	SANEKEY(ltchars.t_flushc);
 #endif
 }
 
@@ -898,11 +876,11 @@ void adjustwinsize() /**/
 {
 #ifdef TIOCGWINSZ
 	ioctl(SHTTY,TIOCGWINSZ,&shttyinfo.winsize);
-	if (!(columns = shttyinfo.winsize.ws_col))
-		columns = 80;
+	if (!(columns = shttyinfo.winsize.ws_col)) columns = 80;
 	lines = shttyinfo.winsize.ws_row;
-	if (zleactive)
-		refresh();
+	setintenv("COLUMNS",columns);
+	setintenv("LINES",lines);
+	if (zleactive) refresh();
 #endif
 }
 
@@ -919,8 +897,8 @@ int pid,statusp;
 		_exit(1);
 		}
 	while (waitpid(pid,&statusp,WUNTRACED) == -1 && errno == EINTR);
-	if (WIFEXITED(SP(statusp)))
-		return WEXITSTATUS(SP(statusp));
+	if (WIFEXITED(statusp))
+		return WEXITSTATUS(statusp);
 	return 1;
 #else
 	if (!waitfork())
@@ -1038,32 +1016,32 @@ int t0;
 	if (t0 == -1)
 		return;
 	if (jobbing && (t0 == SIGTTOU || t0 == SIGTSTP || t0 == SIGTTIN
-			|| t0 == SIGPIPE))
-		{
-		zerr("can't trap SIG%s in interactive shells",sigs[t0],0);
+			|| t0 == SIGPIPE)) {
 		return;
-		}
+	}
 	sigtrapped[t0] = 0;
-	if (t0 == SIGINT)
-		intr();
-	else if (t0 && t0 < SIGCOUNT && t0 != SIGCHLD)
-		signal(t0,SIG_DFL);
-	if (sigfuncs[t0])
-		freestruct(sigfuncs[t0]);
+	if (t0 == SIGINT) intr();
+	else if (t0 && t0 < SIGCOUNT && t0 != SIGCHLD) signal(t0,SIG_DFL);
+	if (sigfuncs[t0]) freestruct(sigfuncs[t0]);
 }
 
 void dotrap(sig) /**/
 int sig;
 {
-int sav;
+int sav,savval;
 
 	sav = sigtrapped[sig];
+	savval = lastval;
 	if (sav == 2)
 		return;
 	sigtrapped[sig] = 2;
-	if (sigfuncs[sig])
-		newrunlist(sigfuncs[sig]);
+	if (sigfuncs[sig]) {
+		lexsave();
+		doshfuncnoval(sigfuncs[sig],NULL,0);
+		lexrestore();
+	}
 	sigtrapped[sig] = sav;
+	lastval = savval;
 }
 
 /* copy len chars from t into s, and null terminate */
@@ -1097,10 +1075,15 @@ char *u = *s;
 	*u = '\0';
 }
 
-void checkrmall() /**/
+void checkrmall(s) /**/
+char *s;
 {
 	fflush(stdin);
-	fprintf(stderr,"zsh: sure you want to delete all the files in %s? ",cwd);
+	if (*s == '/')
+		fprintf(stderr,"zsh: sure you want to delete all the files in %s? ",s);
+	else
+		fprintf(stderr,"zsh: sure you want to delete all the files in %s/%s? ",
+			(pwd[1]) ? pwd : "",s);
 	fflush(stderr);
 	feep();
 	errflag |= (getquery() != 'y');
@@ -1109,8 +1092,13 @@ void checkrmall() /**/
 int getquery() /**/
 {
 char c;
+long val;
 
 	setcbreak();
+#ifdef FIONREAD
+	ioctl(SHTTY,FIONREAD,&val);
+	if (val) { unsetcbreak(); write(2,"n\n",2); return 'n'; }
+#endif
 	if (read(SHTTY,&c,1) == 1)
 		if (c == 'y' || c == 'Y' || c == '\t') c = 'y';
 	unsetcbreak();
@@ -1135,11 +1123,10 @@ char *s;char *junk;
 int nd;
 
 	nd = spdist(s,guess,strlen(guess)/4+1);
-	if (nd <= d)
-		{
+	if (nd <= d) {
 		best = s;
 		d = nd;
-		}
+	}
 }
 
 /* spellcheck a word */
@@ -1149,53 +1136,71 @@ void spckword(s,s2,tptr,cmd,ask) /**/
 char **s;char **s2;char **tptr;int cmd;int ask;
 {
 char *t,*u;
+char firstchar;
 int x;
+int pram = 0;
 
 	if (**s == '-' || **s == '%')
 		return;
 	if (!strcmp(*s,"in"))
 		return;
-	if (gethnode(*s,cmdnamtab) || gethnode(*s,aliastab) || strlen(*s) == 1)
-		return;
-	for (t = *s; *t; t++)
-		if (itok(*t))
-			return;
-	if (access(*s,F_OK) == 0)
-		return;
+	if (!(*s)[0] || !(*s)[1]) return;
+	if (gethnode(*s,cmdnamtab) || gethnode(*s,aliastab)) return;
+	t = *s;
+	if (*t == Tilde || *t == Equals || *t == String) t++;
+	for (; *t; t++) if (itok(*t)) return;
 	best = NULL;
-	for (t = *s; *t; t++)
-		if (*t == '/')
-			break;
-	if ((u = spname(*s)) != *s)
-		best = u;
-	else if (!*t && !cmd) {
-		guess = *s;
+	for (t = *s; *t; t++) if (*t == '/') break;
+	if (**s == String) {
+		if (*t) return;
+		pram = 1;
+		guess = *s+1;
 		d = 100;
-		listhtable(aliastab,spscan);
-		listhtable(cmdnamtab,spscan);
+		listhtable(paramtab,spscan);
+	} else {
+		if ((u = spname(guess = *s)) != *s)
+			best = u;
+		if (!*t && !cmd) {
+			if (access(*s,F_OK) == 0) return;
+			if (hashcmd(*s,pathchecked)) return;
+			guess = *s;
+			d = 100;
+			listhtable(aliastab,spscan);
+			listhtable(cmdnamtab,spscan);
+		}
 	}
-	if (best && strlen(best) > 1 && strcmp(best,*s)) {
+	if (errflag) return;
+	if (best && strlen(best) > 1 && strcmp(best,guess)) {
 		if (ask) {
-			char *p = sprompt;
-			while (*p) {
-				if (*p == '%') {
-					if (p[1] == 'r') fprintf(stderr,"%s",best);
-					if (p[1] == 's') fprintf(stderr,"%s",*s);
-					if (p[1]) p++;
-				} else fputc(*p,stderr);
-				p++;
-			}
+			char *pp;
+			int junk;
+
+			rstring = best; Rstring = guess;
+			firstchar = *guess;
+			if (*guess == Tilde) *guess = '~';
+			else if (*guess == String) *guess = '$';
+			else if (*guess == Equals) *guess = '=';
+			pp = putprompt(sprompt,&junk);
+			*guess = firstchar;
+			fprintf(stderr,"%s",pp);
 			fflush(stderr);
 			feep();
 			x = getquery();
 		} else
 			x = 'y';
 		if (x == 'y') {
-			*s = strdup(best);
+			if (!pram) {
+				*s = strdup(best);
+			} else {
+				*s = alloc(strlen(best)+2);
+				strcpy(*s+1,best);
+				**s = String;
+			}
 			if (s2) {
-				if (*tptr && !strcmp(hlastw,*s2)) {
+				if (*tptr && !strcmp(hlastw,*s2) && hlastw < hptr) {
 					char *z;
 					hptr = hlastw;
+					if (pram) hwaddc('$');
 					for (z = best; *z; z++) hwaddc(*z);
 					hwaddc(HISTSPACE);
 					*tptr = hptr-1;
@@ -1206,8 +1211,7 @@ int x;
 		} else if (x == 'a') {
 			histdone |= HISTFLAG_NOEXEC;
 		} else if (x == 'e') {
-			histdone |= HISTFLAG_NOEXEC;
-			zrecall = 1;
+			histdone |= HISTFLAG_NOEXEC|HISTFLAG_RECALL;
 		}
 	}
 }
@@ -1220,7 +1224,12 @@ static char *estr[] = {"Jan","Feb","Mar","Apr","May","Jun","Jul",
 	"Aug","Sep","Oct","Nov","Dec"};
 static char *lstr[] = {"12"," 1"," 2"," 3"," 4"," 5"," 6"," 7"," 8"," 9",
 	"10","11"};
+char tmp[3];
+#ifdef HAS_STRFTIME
+char *origbuf = buf;
+#endif
 
+	tmp[0] = '%'; tmp[2] = '\0';
 	while (*fmt)
 		if (*fmt == '%')
 			{
@@ -1271,7 +1280,16 @@ static char *lstr[] = {"12"," 1"," 2"," 3"," 4"," 5"," 6"," 7"," 8"," 9",
 					*buf++ = '0'+tm->tm_year%10;
 					break;
 				default:
-					exit(20);
+#ifdef HAS_STRFTIME
+					*buf = '\0';
+					tmp[1] = fmt[-1];
+					strftime(buf,bufsize-strlen(origbuf),tmp,tm);
+					buf += strlen(buf);
+#else
+					*buf++ = '%';
+					*buf++ = fmt[-1];
+#endif
+					break;
 				}
 			}
 		else
@@ -1285,17 +1303,17 @@ char **arr;int delim;
 {
 int len = 0;
 char **s,*ret,*ptr;
+static char *lastmem = NULL;
 
 	for (s = arr; *s; s++)
 		len += strlen(*s)+1;
-	if (!len)
-		return ztrdup("");
-	ptr = ret = zalloc(len);
-	for (s = arr; *s; s++)
-		{
+	if (!len) return "";
+	if (lastmem) free(lastmem);
+	lastmem = ptr = ret = zalloc(len);
+	for (s = arr; *s; s++) {
 		strucpy(&ptr,*s);
 		*ptr++ = delim;
-		}
+	}
 	ptr[-1] = '\0';
 	return ret;
 }
@@ -1317,18 +1335,17 @@ char *s;
 {
 int ct;
 char *t,**ret,**ptr;
+char **lastmem = NULL;
 
-	for (t = s, ct = 0; *t; t++)
-		if (*t == ':')
-			ct++;
-	ptr = ret = (char **) zalloc(sizeof(char **)*(ct+2));
+	for (t = s, ct = 0; *t; t++) if (*t == ':') ct++;
+	if (lastmem) freearray(lastmem);
+	lastmem = ptr = ret = (char **) zalloc(sizeof(char **)*(ct+2));
 	t = s;
-	do
-		{
+	do {
 		for (s = t; *t && *t != ':'; t++);
 		*ptr = zalloc((t-s)+1);
 		ztrncpy(*ptr++,s,t-s);
-		}
+	}
 	while (*t++);
 	*ptr = NULL;
 	return ret;
@@ -1341,17 +1358,14 @@ int ct;
 char *t,**ret,**ptr;
 
 	for (t = s, ct = 0; *t; t++)
-		if (iblank(*t))
-			ct++;
+		if (isep(*t)) ct++;
 	ptr = ret = (char **) zalloc(sizeof(char **)*(ct+2));
 	t = s;
-	do
-		{
-		for (s = t; *t && !inblank(*t); t++);
+	do {
+		for (s = t; *t && !isep(*t); t++);
 		*ptr = zalloc((t-s)+1);
 		ztrncpy(*ptr++,s,t-s);
-		}
-	while (*t++);
+	} while (*t++);
 	*ptr = NULL;
 	return ret;
 }
@@ -1399,6 +1413,7 @@ vptr a;
 struct node *n = a,*m;
 int t0;
 
+	if (!a) return NULL;
 	m = alloc(sizeof *m);
 	*m = *n;
 	for (t0 = 0; t0 != 4; t0++)
@@ -1455,8 +1470,7 @@ char *s;
 {
 char **t = (char **) zalloc((s) ? (2*sizeof s) : (sizeof s));
 
-	if (*t = s)
-		t[1] = NULL;
+	if (*t = s) t[1] = NULL;
 	return t;
 }
 
@@ -1536,8 +1550,7 @@ char *s;
 			t0++)
 		typtab[t0] |= ITOK;
 	for (s = ifs; *s; s++)
-		typtab[(int) (unsigned char) *s] |=
-			(*s == '\n') ? ISEP|INBLANK : ISEP|IBLANK|INBLANK;
+		typtab[(int) (unsigned char) *s] |= ISEP;
 	for (s = wordchars; *s; s++)
 		typtab[(int) (unsigned char) *s] |= IWORD;
 	for (s = SPECCHARS; *s; s++)
@@ -1562,24 +1575,24 @@ char *oldname;
 {
 	char *p,guess[MAXPATHLEN+1],best[MAXPATHLEN+1];
 	static char newname[MAXPATHLEN+1];
-	char *new = newname, *old = oldname;
+	char *new = newname, *old;
 
-	for (;;)
-	{
-		while (*old == '/')
-			*new++ = *old++;
+	if (itok(*oldname)) {
+		singsub(&oldname);
+		if (!oldname) return NULL;
+	}
+	if (access(oldname,F_OK) == 0) return NULL;
+	old = oldname;
+	for (;;) {
+		while (*old == '/') *new++ = *old++;
 		*new = '\0';
-		if (*old == '\0')
-			return newname;
+		if (*old == '\0') return newname;
 		p = guess;
 		for (; *old != '/' && *old != '\0'; old++)
-			if (p < guess+MAXPATHLEN)
-				*p++ = *old;
+			if (p < guess+MAXPATHLEN) *p++ = *old;
 		*p = '\0';
-		if (mindist(newname,guess,best) >= 3)
-			return oldname;
-		for (p = best; *new = *p++; )
-			new++;
+		if (mindist(newname,guess,best) >= 3) return NULL;
+		for (p = best; *new = *p++; ) new++;
 	}
 }
 
@@ -1589,22 +1602,21 @@ char *dir;char *guess;char *best;
 	int d,nd;
 	DIR *dd;
 	struct direct *de;
+	char buf[MAXPATHLEN];
 
 	if (dir[0] == '\0')
 		dir = ".";
 	d = 100;
-	if (!(dd = opendir(dir)))
-		return d;
-	while (de = readdir(dd))
-	{
+	sprintf(buf,"%s/%s",dir,guess);
+	if (access(buf,F_OK) == 0) { strcpy(best,guess); return 0; }
+	if (!(dd = opendir(dir))) return d;
+	while (de = readdir(dd)) {
 		nd = spdist(de->d_name,guess,strlen(guess)/4+1);
-		if (nd <= d)
-			{
+		if (nd <= d) {
 			strcpy(best,de->d_name);
 			d = nd;
-			if (d == 0)
-				break;
-			}
+			if (d == 0) break;
+		}
 	}
 	closedir(dd);
 	return d;
@@ -1688,13 +1700,15 @@ char **av,*p,*q;
 int tulower(c) /**/
 int c;
 {
-	return (c >= 'A' && c <= 'Z') ? c-'A'+'a' : c;
+	c &= 0xff;
+	return (isupper(c) ? tolower(c) : c);
 }
 
 int tuupper(c) /**/
 int c;
 {
-	return (c >= 'a' && c <= 'z') ? c-'a'+'A' : c;
+	c &= 0xff;
+	return (islower(c) ? toupper(c) : c);
 }
 
 #ifdef SYSV
@@ -1714,3 +1728,76 @@ int result;
 	} else return -1;
 }
 #endif
+
+/* set cbreak mode, or the equivalent */
+
+void setcbreak() /**/
+{
+struct ttyinfo ti;
+
+	ti = shttyinfo;
+#ifdef TIO
+	ti.tio.c_lflag &= ~ICANON;
+	ti.tio.c_cc[VMIN] = 1;
+	ti.tio.c_cc[VTIME] = 0;
+#else
+	ti.sgttyb.sg_flags |= CBREAK;
+#endif
+	settyinfo(&ti);
+}
+
+int getlineleng() /**/
+{
+int z;
+
+#ifdef TIOCSWINSZ
+	z = shttyinfo.winsize.ws_col;
+	return (z) ? z : 80;
+#else
+	return 80;
+#endif
+}
+
+void unsetcbreak() /**/
+{
+	settyinfo(&shttyinfo);
+}
+
+/* give the tty to some process */
+
+void attachtty(pgrp) /**/
+long pgrp;
+{
+static int ep = 0;
+
+	if (jobbing) {
+#ifdef HAS_TCSETPGRP
+		if (SHTTY != -1 && tcsetpgrp(SHTTY,pgrp) == -1 && !ep)
+#else
+		int arg = pgrp;
+		if (SHTTY != -1 && ioctl(SHTTY,TIOCSPGRP,&arg) == -1 && !ep)
+#endif
+			{
+			zerr("can't set tty pgrp: %e",NULL,errno);
+			fflush(stderr);
+			opts[MONITOR] = OPT_UNSET;
+			ep =1;
+			errflag = 0;
+			}
+	}
+}
+
+/* get the tty pgrp */
+
+long gettygrp() /**/
+{
+int arg = -1;
+
+	if (SHTTY == -1) return -1;
+#ifdef HAS_TCSETPGRP
+	arg = tcgetpgrp(SHTTY);
+#else
+	ioctl(SHTTY,TIOCGPGRP,&arg);
+#endif
+	return arg;
+}
