@@ -1,34 +1,22 @@
 /*
-
-	zle.h - header file for line editor
-
-	This file is part of zsh, the Z shell.
-
-	zsh is free software; no one can prevent you from reading the source
-   code, or giving it to someone else.
-
-   This file is copyrighted under the GNU General Public License, which
-   can be found in the file called COPYING.
-
-   Copyright (C) 1990, 1991 Paul Falstad
-
-   zsh is distributed in the hope that it will be useful, but
-   WITHOUT ANY WARRANTY.  No author or distributor accepts
-   responsibility to anyone for the consequences of using it or for
-   whether it serves any particular purpose or works at all, unless he
-   says so in writing.  Refer to the GNU General Public License
-   for full details.
-
-   Everyone is granted permission to copy, modify and redistribute
-   zsh, but only under the conditions described in the GNU General Public
-   License.   A copy of this license is supposed to have been given to you
-   along with zsh so you can know your rights and responsibilities.
-   It should be in a file named COPYING.
-
-   Among other things, the copyright notice and this notice must be
-   preserved on all copies.
-
-*/
+ *
+ * zle.h - header file for line editor
+ *
+ * This file is part of zsh, the Z shell.
+ *
+ * This software is Copyright 1992 by Paul Falstad
+ *
+ * Permission is hereby granted to copy, reproduce, redistribute or otherwise
+ * use this software as long as: there is no monetary profit gained
+ * specifically from the use or reproduction of this software, it is not
+ * sold, rented, traded or otherwise marketed, and this copyright notice is
+ * included prominently in any copy made. 
+ *
+ * The author make no claims as to the fitness or correctness of this software
+ * for any use whatsoever, and it is provided as is. Any use of this software
+ * is at the user's own risk. 
+ *
+ */
 
 #ifdef ZLEGLOBALS
 #define ZLEXTERN
@@ -74,6 +62,9 @@ ZLEXTERN int lastcol;
 /* != 0 if we're getting a vi range */
 ZLEXTERN int virangeflag;
 
+/* kludge to get cw and dw to work right */
+ZLEXTERN int wordflag;
+
 #endif
 
 /* last named command done */
@@ -97,7 +88,7 @@ ZLEXTERN int resetneeded;
 ZLEXTERN int zleactive;
 
 /* the line buffer */
-ZLEXTERN char *line;
+ZLEXTERN unsigned char *line;
 
 /* the cut buffer */
 ZLEXTERN char *cutbuf;
@@ -111,6 +102,9 @@ ZLEXTERN char *curhistline;
 /* the status line */
 ZLEXTERN char *statusline;
 
+/* 1 if a complete added a slash at the end of a directory name */
+ZLEXTERN int addedslash;
+
 /*
 	the current history line and cursor position for the top line
 	on the buffer stack
@@ -134,6 +128,7 @@ typedef void bindfunc DCLPROTO((void));
 typedef bindfunc *F;
 
 struct key {
+	struct hashnode *next; int canfree; char *nam; /* hash data */
 	int func;			/* function code for this key */
 	char *str;			/* string corresponding to this key,
 								if func = z_sequenceleadin				 */
@@ -160,7 +155,7 @@ struct undoent {
 struct undoent undos[UNDOCT];
 
 /* the line before last mod (for undo purposes) */
-ZLEXTERN char *lastline;
+ZLEXTERN unsigned char *lastline;
 
 /* buffer specified with "x */
 ZLEXTERN int vibufspec;
@@ -180,6 +175,7 @@ ZLEXTERN int visrchsense;
 #define ZLE_KILL        (64|ZLE_NAMEDBUFFER)
 #define ZLE_HISTSEARCH  256
 #define ZLE_NEGARG      512
+#define ZLE_INSERT     1024
 
 typedef struct key *Key;
 
@@ -192,9 +188,7 @@ ZLEXTERN int vimarkcs[27],vimarkline[27];
 #define KRINGCT 8
 ZLEXTERN char *kring[KRINGCT];
 ZLEXTERN int kringnum;
-ZLEXTERN char *vibuf[35];
-
-ZLEXTERN int zrecall;
+ZLEXTERN char *vibuf[36];
 
 #define z_acceptandhold 0
 #define z_acceptandinfernexthistory 1
@@ -340,7 +334,10 @@ ZLEXTERN int zrecall;
 #define z_vibackwardkillword 141
 #define z_expandcmdpath 142
 #define z_negargument 143
-#define ZLECMDCOUNT 144
+#define z_poundinsert 144
+#define z_viforwardword 145
+#define z_vibackwardword 146
+#define ZLECMDCOUNT 147
 
 extern struct zlecmd zlecmds[];
 
