@@ -1,34 +1,22 @@
 /*
-
-	zle_bindings.c - commands and keymaps
-
-	This file is part of zsh, the Z shell.
-
-	zsh is free software; no one can prevent you from reading the source
-   code, or giving it to someone else.
-
-   This file is copyrighted under the GNU General Public License, which
-   can be found in the file called COPYING.
-
-   Copyright (C) 1990, 1991 Paul Falstad
-
-   zsh is distributed in the hope that it will be useful, but
-   WITHOUT ANY WARRANTY.  No author or distributor accepts
-   responsibility to anyone for the consequences of using it or for
-   whether it serves any particular purpose or works at all, unless he
-   says so in writing.  Refer to the GNU General Public License
-   for full details.
-
-   Everyone is granted permission to copy, modify and redistribute
-   zsh, but only under the conditions described in the GNU General Public
-   License.   A copy of this license is supposed to have been given to you
-   along with zsh so you can know your rights and responsibilities.
-   It should be in a file named COPYING.
-
-   Among other things, the copyright notice and this notice must be
-   preserved on all copies.
-
-*/
+ *
+ * zle_bindings.c - commands and keymaps
+ *
+ * This file is part of zsh, the Z shell.
+ *
+ * This software is Copyright 1992 by Paul Falstad
+ *
+ * Permission is hereby granted to copy, reproduce, redistribute or otherwise
+ * use this software as long as: there is no monetary profit gained
+ * specifically from the use or reproduction of this software, it is not
+ * sold, rented, traded or otherwise marketed, and this copyright notice is
+ * included prominently in any copy made. 
+ *
+ * The author make no claims as to the fitness or correctness of this software
+ * for any use whatsoever, and it is provided as is. Any use of this software
+ * is at the user's own risk. 
+ *
+ */
 
 #define ZLE
 #include "zsh.h"
@@ -81,7 +69,7 @@ struct zlecmd zlecmds[] = {
 "history-search-backward",historysearchbackward,ZLE_HISTSEARCH,
 "history-search-forward",historysearchforward,ZLE_HISTSEARCH,
 "infer-next-history",infernexthistory,0,
-"insert-last-word",insertlastword,0,
+"insert-last-word",insertlastword,ZLE_INSERT,
 "kill-buffer",killbuffer,ZLE_KILL,
 "kill-line",killline,ZLE_KILL,
 "kill-region",killregion,ZLE_KILL,
@@ -93,14 +81,14 @@ struct zlecmd zlecmds[] = {
 "menu-expand-or-complete",menuexpandorcomplete,ZLE_MENUCMP,
 "overwrite-mode",overwritemode,0,
 "push-line",pushline,0,
-"quoted-insert",quotedinsert,0,
+"quoted-insert",quotedinsert,ZLE_INSERT,
 "quote-line",quoteline,0,
 "quote-region",quoteregion,0,
 "redisplay",redisplay,0,
 "reverse-menu-complete",reversemenucomplete,ZLE_MENUCMP,
 "run-help",processcmd,0,
-"self-insert",selfinsert,0,
-"self-insert-unmeta",selfinsertunmeta,0,
+"self-insert",selfinsert,ZLE_INSERT,
+"self-insert-unmeta",selfinsertunmeta,ZLE_INSERT,
 "send-break",sendbreak,0,
 "send-string",sendstring,0,
 "",NULL,0,
@@ -176,9 +164,12 @@ struct zlecmd zlecmds[] = {
 "emacs-backward-word",emacsbackwardword,ZLE_MOVEMENT,
 "kill-word",killword,ZLE_KILL,
 "vi-kill-line",vikillline,0,
-"vi-backward-delete-word",vibackwardkillword,0,
+"vi-backward-kill-word",vibackwardkillword,0,
 "expand-cmd-path",expandcmdpath,0,
 "neg-argument",negargument,ZLE_NEGARG,
+"pound-insert",poundinsert,0,
+"vi-forward-word",viforwardword,ZLE_MOVEMENT,
+"vi-backward-word",vibackwardword,ZLE_MOVEMENT,
 "",NULL,0
 };
 
@@ -512,7 +503,7 @@ int vicmdbind[128] = {
 /*   */ z_viforwardchar,
 /* ! */ z_undefinedkey,
 /* " */ z_visetbuffer,
-/* # */ z_undefinedkey,
+/* # */ z_poundinsert,
 /* $ */ z_viendofline,
 /* % */ z_vimatchbracket,
 /* & */ z_undefinedkey,
@@ -575,7 +566,7 @@ int vicmdbind[128] = {
 /* _ */ z_undefinedkey,
 /* ` */ z_vigotomark,
 /* a */ z_viaddnext,
-/* b */ z_backwardword,
+/* b */ z_vibackwardword,
 /* c */ z_vichange,
 /* d */ z_videlete,
 /* e */ z_viforwardwordend,
@@ -592,11 +583,11 @@ int vicmdbind[128] = {
 /* p */ z_viputafter,
 /* q */ z_undefinedkey,
 /* r */ z_vireplacechars,
-/* s */ z_vichangewholeline,
+/* s */ z_visubstitute,
 /* t */ z_vifindnextcharskip,
-/* u */ z_undo,
+/* u */ z_viundochange,
 /* v */ z_undefinedkey,
-/* w */ z_forwardword,
+/* w */ z_viforwardword,
 /* x */ z_videletechar,
 /* y */ z_viyank,
 /* z */ z_undefinedkey,
