@@ -1,41 +1,31 @@
 /*
-
-	zle_main.c - main routines for line editor
-
-	This file is part of zsh, the Z shell.
-
-	zsh is free software; no one can prevent you from reading the source
-   code, or giving it to someone else.
-
-   This file is copyrighted under the GNU General Public License, which
-   can be found in the file called COPYING.
-
-   Copyright (C) 1990, 1991 Paul Falstad
-
-   zsh is distributed in the hope that it will be useful, but
-   WITHOUT ANY WARRANTY.  No author or distributor accepts
-   responsibility to anyone for the consequences of using it or for
-   whether it serves any particular purpose or works at all, unless he
-   says so in writing.  Refer to the GNU General Public License
-   for full details.
-
-   Everyone is granted permission to copy, modify and redistribute
-   zsh, but only under the conditions described in the GNU General Public
-   License.   A copy of this license is supposed to have been given to you
-   along with zsh so you can know your rights and responsibilities.
-   It should be in a file named COPYING.
-
-   Among other things, the copyright notice and this notice must be
-   preserved on all copies.
-
-*/
+ *
+ * zle_main.c - main routines for line editor
+ *
+ * This file is part of zsh, the Z shell.
+ *
+ * This software is Copyright 1992 by Paul Falstad
+ *
+ * Permission is hereby granted to copy, reproduce, redistribute or otherwise
+ * use this software as long as: there is no monetary profit gained
+ * specifically from the use or reproduction of this software, it is not
+ * sold, rented, traded or otherwise marketed, and this copyright notice is
+ * included prominently in any copy made. 
+ *
+ * The author make no claims as to the fitness or correctness of this software
+ * for any use whatsoever, and it is provided as is. Any use of this software
+ * is at the user's own risk. 
+ *
+ */
 
 #define ZLEGLOBALS
 #define ZLE
 #include "zsh.h"
 #include <sys/types.h>
-#include <sys/ioctl.h>
 #include <sys/errno.h>
+#ifdef HAS_SYS_SELECT
+#include <sys/select.h>
+#endif
 
 static Key cky;
 
@@ -44,7 +34,9 @@ static Key cky;
 void setterm() /**/
 {
 struct ttyinfo ti;
+#ifdef FIONREAD
 long val;
+#endif
 
 #ifdef CLOBBERS_TYPEAHEAD
 #ifdef FIONREAD
@@ -54,37 +46,37 @@ long val;
 #endif
 	inittty();
 	ti = shttyinfo;
-#ifdef TERMIOS
-	ti.termios.c_lflag &= ~(ICANON|ECHO);
-	ti.termios.c_cc[VQUIT] =
+#ifdef TIO
+	ti.tio.c_lflag &= ~(ICANON|ECHO
+#ifdef FLUSHO
+		|FLUSHO
+#endif
+		);
+	ti.tio.c_cc[VQUIT] =
 #ifdef VDISCARD
-		ti.termios.c_cc[VDISCARD] = 
+		ti.tio.c_cc[VDISCARD] = 
 #endif
 #ifdef VSUSP
-		ti.termios.c_cc[VSUSP] =
+		ti.tio.c_cc[VSUSP] =
 #endif
 #ifdef VDSUSP
-		ti.termios.c_cc[VDSUSP] =
+		ti.tio.c_cc[VDSUSP] =
 #endif
-		0;
-	ti.termios.c_cc[VMIN] = 1;
-	ti.termios.c_cc[VTIME] = 0;
-	ti.termios.c_iflag &= ~(INLCR|ICRNL);
-#else
-#ifdef TERMIO
-	ti.termio.c_lflag &= ~(ICANON|ECHO);
-	ti.termio.c_cc[VQUIT] =
-		0;
-	ti.termio.c_cc[VMIN] = 1;
-	ti.termio.c_cc[VTIME] = 0;
-	ti.termio.c_iflag &= ~(INLCR|ICRNL);
+#ifdef VSWTCH
+		ti.tio.c_cc[VSWTCH] =
+#endif
+		VDISABLEVAL;
+	ti.tio.c_cc[VMIN] = 1;
+	ti.tio.c_cc[VTIME] = 0;
+	ti.tio.c_iflag &= ~(INLCR|ICRNL);
 #else
 	ti.sgttyb.sg_flags = (ti.sgttyb.sg_flags | CBREAK) & ~ECHO;
+	ti.lmodes &= ~LFLUSHO;
 	ti.tchars.t_quitc =
 		ti.ltchars.t_suspc =
+		ti.ltchars.t_flushc =
 		ti.ltchars.t_dsuspc = ti.ltchars.t_lnextc = -1;
 #endif
-#endif
 #ifdef TTY_NEEDS_DRAINING
 	drainoutput();
 #endif
@@ -119,12 +111,20 @@ unsigned int getkey(tmok) /**/
 int tmok;
 {
 char cc;
-unsigned int ret;
+unsigned int ret = 0;
+int die = 0;
+#ifdef HAS_SELECT
+fd_set foofd;
+#endif
 
 	if (kungetct)
 		ret = (unsigned int) (unsigned char) kungetbuf[--kungetct];
 	else {
-		while (read(0,&cc,1) != 1)
+		while (
+#ifdef HAS_SELECT
+		FD_SET(0,&foofd), select(1,&foofd,NULL,NULL,NULL) != 1 ||
+#endif
+		read(0,&cc,1) != 1)
 			if (errno == EINTR) {
 				if (!errflag)
 					continue;
@@ -134,6 +134,13 @@ unsigned int ret;
 				return 3;
 			} else if (errno == EWOULDBLOCK) {
 				fcntl(0,F_SETFL,0);
+			} else if (errno == EIO && !die) {
+				ret = jobbing;
+				jobbing = 1;
+				attachtty(mypgrp);
+				refresh(); /* kludge! */
+				jobbing = ret;
+				die = 1;
 			} else {
 				zerr("error on TTY read: %e",NULL,errno);
 				exit(1);
@@ -150,29 +157,32 @@ unsigned int ret;
 
 /* read a line */
 
-char *zleread(ppt,ppt2,plen) /**/
-char *ppt;char *ppt2;int plen;
+unsigned char *zleread(ppt,ppt2,plen) /**/
+unsigned char *ppt;unsigned char *ppt2;int plen;
 {
 int z;
 long costmult;
+unsigned char *s;
+#ifdef HAS_SELECT
 struct timeval tv;
 fd_set foofd;
-char *s;
 
 	tv.tv_sec = 0;
+#endif
 	fflush(stdout);
 	fflush(stderr);
 	intr();
 	costmult = 3840000L/((baud) ? baud : 2400);
-	insmode = 1; eofsent = 0; resetneeded =0 ;
-	pmpt = ppt;
-	pmpt2 = ppt2;
+	insmode = unset(OVERSTRIKE); eofsent = 0; resetneeded =0 ;
+	pmpt = (char *)ppt;
+	pmpt2 = (char *)ppt2;
 	permalloc();
 	histline = curhist;
 	pptlen = plen;
 	resetneeded = 1;
+#ifdef HAS_SELECT
 	FD_ZERO(&foofd);
-	zleactive = 1;
+#endif
 	undoing = 1;
 	line = zalloc(linesz = 256);
 	*line = '\0';
@@ -181,12 +191,12 @@ char *s;
 	mult = 1;
 	vibufspec = 0;
 	bindtab = mainbindtab;
-	vichgflag = 0;
+	addedslash = vichgflag = 0;
 	viinsbegin = 0;
 	statusline = NULL;
 	if (s = getnode(bufstack))
 		{
-		setline(s);
+		setline((char *) s);
 		free(s);
 		if (stackcs != -1)
 			{
@@ -202,14 +212,10 @@ char *s;
 			}
 		}
 	initundo();
-	if (unset(NOPROMPTCLOBBER))
+	if (unset(NOPROMPTCR))
 		putchar('\r');
 	if (tmout)
 		alarm(tmout);
-	if (zrecall) {
-		zrecall = 0;
-		uphistory();
-	}
 	refresh();
 	errflag = 0;
 	while (!done && !errflag)
@@ -230,24 +236,32 @@ char *s;
 				mult = 1;
 			if ((lastcmd & ZLE_UNDO) != (zc->flags & ZLE_UNDO) && undoing)
 				addundo();
-			if (menucmp && !(zc->flags & ZLE_MENUCMP))
-				freemenu();
+			if (!(zc->flags & ZLE_MENUCMP)) {
+				if (menucmp) freemenu();
+				if (addedslash &&
+						!((zc->flags & ZLE_INSERT) && c != ' ')) {
+					backdel(1);
+				}
+				addedslash = 0;
+			}
 			if (zc->func)
-				(zc->func)();
+				(*zc->func)();
 			lastcmd = zc->flags;
-			if (!(lastcmd & ZLE_UNDO)) addundo();
+			if (!(lastcmd & ZLE_UNDO) && undoing) addundo();
 			}
 		else
 			{
 			errflag = 1;
 			break;
 			}
+#ifdef HAS_SELECT
 		FD_SET(0,&foofd);
 		if ((tv.tv_usec = cost*costmult) > 500000)
 			tv.tv_usec = 500000;
+#endif
 		if (!kungetct
 #ifdef HAS_SELECT
-			&& !select(1,&foofd,NULL,NULL,&tv)
+			&& select(1,&foofd,NULL,NULL,&tv) <= 0
 #endif
 			)
 			refresh();
@@ -257,7 +271,7 @@ char *s;
 	statusline = NULL;
 	trashzle();
 	alarm(0);
-	z = strlen(line);
+	z = strlen((char *)line);
 	line[z] = '\n';
 	line[z+1] = 0;
 	heapalloc();
@@ -338,28 +352,24 @@ int t0,vi = 0;
 char buf[3],*s;
 
 	lastnamed = z_undefinedkey;
-	if (s = zgetenv("VISUAL"))
-		{
+	if (s = zgetenv("VISUAL")) {
 		if (ztrstr(s,"vi"))
 			vi = 1;
-		}
+	}
 	else if ((s = zgetenv("EDITOR")) && ztrstr(s,"vi"))
 		vi = 1;
-	if (vi)
-		{
+	if (vi) {
 		for (t0 = 0; t0 != 32; t0++)
 			mainbindtab[t0] = viinsbind[t0];
 		for (t0 = 32; t0 != 256; t0++)
 			mainbindtab[t0] = z_selfinsert;
 		mainbindtab[127] = z_backwarddeletechar;
-		}
-	else
-		{
+	} else {
 		for (t0 = 0; t0 != 128; t0++)
 			mainbindtab[t0] = emacsbind[t0];
 		for (t0 = 128; t0 != 256; t0++)
 			mainbindtab[t0] = z_selfinsert;
-		}
+	}
 	for (t0 = 0200; t0 != 0240; t0++)
 		mainbindtab[t0] = z_undefinedkey;
 	for (t0 = 0; t0 != 128; t0++)
@@ -395,16 +405,13 @@ char buf[3],*s;
 	addhperm("run-help",mkanode(ztrdup("man"),1),aliastab,NULL);
 	addhperm("which-command",mkanode(ztrdup("whence"),1),aliastab,NULL);
 	strcpy(buf,"\33q");
-	if (!vi)
-		for (t0 = 128; t0 != 256; t0++)
-			if (emacsbind[t0] != z_undefinedkey) {
-				buf[1] = t0 & 0x7f;
-				addhnode(ztrdup(buf),makefunckey(emacsbind[t0]),xbindtab,NULL);
-			}
-	for (t0 = 0; t0 != 36; t0++)
-		vibuf[t0] = NULL;
-	for (t0 = 0; t0 != 26; t0++)
-		vimarkline[t0] = 0;
+	for (t0 = 128; t0 != 256; t0++)
+		if (emacsbind[t0] != z_undefinedkey) {
+			buf[1] = t0 & 0x7f;
+			addhnode(ztrdup(buf),makefunckey(emacsbind[t0]),xbindtab,NULL);
+		}
+	for (t0 = 0; t0 != 36; t0++) vibuf[t0] = NULL;
+	for (t0 = 0; t0 != 26; t0++) vimarkline[t0] = 0;
 	stackhist = stackcs = -1;
 	vichgbufsz = 0;
 	vichgbuf = NULL;
