@@ -1,34 +1,22 @@
 /*
-
-	zle_tricky.c - expansion and completion
-
-	This file is part of zsh, the Z shell.
-
-	zsh is free software; no one can prevent you from reading the source
-   code, or giving it to someone else.
-
-   This file is copyrighted under the GNU General Public License, which
-   can be found in the file called COPYING.
-
-   Copyright (C) 1990, 1991 Paul Falstad
-
-   zsh is distributed in the hope that it will be useful, but
-   WITHOUT ANY WARRANTY.  No author or distributor accepts
-   responsibility to anyone for the consequences of using it or for
-   whether it serves any particular purpose or works at all, unless he
-   says so in writing.  Refer to the GNU General Public License
-   for full details.
-
-   Everyone is granted permission to copy, modify and redistribute
-   zsh, but only under the conditions described in the GNU General Public
-   License.   A copy of this license is supposed to have been given to you
-   along with zsh so you can know your rights and responsibilities.
-   It should be in a file named COPYING.
-
-   Among other things, the copyright notice and this notice must be
-   preserved on all copies.
-
-*/
+ *
+ * zle_tricky.c - expansion and completion
+ *
+ * This file is part of zsh, the Z shell.
+ *
+ * This software is Copyright 1992 by Paul Falstad
+ *
+ * Permission is hereby granted to copy, reproduce, redistribute or otherwise
+ * use this software as long as: there is no monetary profit gained
+ * specifically from the use or reproduction of this software, it is not
+ * sold, rented, traded or otherwise marketed, and this copyright notice is
+ * included prominently in any copy made. 
+ *
+ * The author make no claims as to the fitness or correctness of this software
+ * for any use whatsoever, and it is provided as is. Any use of this software
+ * is at the user's own risk. 
+ *
+ */
 
 #define ZLE
 #include "zsh.h"
@@ -53,7 +41,7 @@ static Lknode menunode;
 
 int usetab() /**/
 {
-char *s = line+cs-1;
+unsigned char *s = line+cs-1;
 
 	for (; s >= line && *s != '\n'; s--)
 		if (*s != '\t' && *s != ' ')
@@ -184,7 +172,7 @@ int t0,t1;
 
 static char *lastmenu = NULL;
 static int lastmenupos = -1;
-static int lincmd,lastambig;
+static int lincmd,linredir,lastambig;
 static char *cmdstr;
 
 void docomplete(lst) /**/
@@ -203,15 +191,12 @@ char *s;
 
 			if (*q == Tilde) q++;
 			else if (*q == Equals) {
-				for (q++; *q; q++)
-					if (*q == '/')
-						break;
-				if (!*q)
-					lst = COMP_EXPAND;
 				q = s+1;
+				if (gethnode(q,cmdnamtab) || hashcmd(q,pathchecked))
+					lst = COMP_EXPAND;
 			} else {
 				for (; *q && *q != String; q++);
-				if (*q == String) {
+				if (*q == String && q[1] != Inpar) {
 					if (getsparam(q+1)) lst = COMP_EXPAND;
 					else lst = COMP_COMPLETE;
 				}
@@ -231,7 +216,7 @@ char *s;
 			cs = wb;
 			foredel(we-wb);
 			/* call the real spell checker, ash@aaii.oz.zu */
-			spckword(x, NULL, NULL, lincmd, 0);
+			spckword(x, NULL, NULL, !lincmd, 0);
 			inststr(*x);
 		} else if (COMP_ISEXPAND(lst))
 			doexpansion(s,lst,lincmd);
@@ -251,7 +236,7 @@ char *s;
 
 	if (isset(LASTMENU) && lastmenu) {
 		if (COMP_ISEXPAND(lst) || cs != lastmenupos ||
-				strcmp(line, lastmenu) != 0) {
+				strcmp((char *) line, lastmenu) != 0) {
 			free(lastmenu);
 			lastmenu = NULL;
 			lastmenupos = -1;
@@ -284,11 +269,12 @@ char *s;
 char *get_comp_string() /**/
 {
 int t0;
-char *s,*linptr;
+unsigned char *s,*linptr;
 
 	linptr = line;
 start:
 	lincmd = incmdpos;
+	linredir = inredir;
 	cmdstr = NULL;
 	zleparse = 1;
 	lexsave();
@@ -298,20 +284,21 @@ start:
 	pushheap();
 	do {
 		lincmd = incmdpos;
+		linredir = inredir;
 		ctxtlex();
 		if (tok == ENDINPUT) break;
-		if (lincmd) cmdstr = strdup(tokstr);
+		if (lincmd && tok == STRING) cmdstr = strdup(tokstr);
 	} while (tok != ENDINPUT && zleparse);
 	t0 = tok;
 	if (t0 == ENDINPUT) {
-		s = ztrdup("");
+		s = (unsigned char *)ztrdup("");
 		we = wb = cs;
 		t0 = STRING;
 	} else if (t0 == STRING) {
-		s = ztrdup(tokstr);
+		s = (unsigned char *)ztrdup(tokstr);
 	} else if (t0 == ENVSTRING) {
-		for (s = tokstr; *s && *s != '='; s++, wb++);
-		if (*s) { s++; wb++; t0 = STRING; s = ztrdup(s); }
+		for (s = (unsigned char *)tokstr; *s && *s != (unsigned char)'='; s++, wb++);
+		if (*s) { s++; wb++; t0 = STRING; s = (unsigned char *)ztrdup(s); }
 		lincmd = 1;
 	}
 	hflush();
@@ -325,7 +312,7 @@ start:
 		goto start;
 	}
 	if (t0 != STRING) { feep(); return NULL; }
-	return s;
+	return (char *)s;
 }
 
 void doexpansion(s,lst,lincmd) /**/
@@ -361,10 +348,14 @@ char *ss;
 	while (ss = ugetnode(vl)) {
 		untokenize(ss);
 		inststr(ss);
+#if 0
 		if (full(vl)) {
 			spaceinline(1);
 			line[cs++] = ' ';
 		}
+#endif
+		spaceinline(1);
+		line[cs++] = ' ';
 	}
 end:
 	popheap();
@@ -411,13 +402,13 @@ char *t,*u,*v;
 		}
 	*v = '\0';
 	spaceinline(strlen(t));
-	strncpy(line+cs,t,strlen(t));
+	strncpy((char *) line+cs,t,strlen(t));
 	cs += strlen(t);
 }
 
 static int ambig,haspath,exact;
 static Lklist matches;
-static char *pat;
+static char *pat,*exactstr;
 static int typechar;
 
 void addmatch(s) /**/
@@ -432,8 +423,7 @@ char *s;
 		}
 	else
 		ambig = strlen(s);
-	if (!strcmp(pat,s))
-		exact = 1;
+	if (!strcmp(pat,s)) { exact = 1; exactstr = pat; }
 	addnodeinorder(matches,strdup(s));
 }
 
@@ -441,15 +431,24 @@ char *s;
 void addcmdmatch(s,t) /**/
 char *s;char *t;
 {
-	if (strpfx(pat,s))
-		addmatch(s);
+	if (strpfx(pat,s)) addmatch(s);
+}
+
+void addcmddirparam(s,t) /**/
+char *s;char *t;
+{
+Param pm = (Param) t;
+
+	if (strpfx(pat,s) && pmtype(pm) == PMFLAG_s) {
+		t = pm->gets.cfn(pm);
+		if (t && *t == '/') addmatch(s);
+	}
 }
 
 void addcmdnodis(s,t) /**/
 char *s;char *t;
 {
-	if (strpfx(pat,s) && ((Cmdnam) t)->type != DISABLED)
-		addmatch(s);
+	if (strpfx(pat,s) && ((Cmdnam) t)->type != DISABLED) addmatch(s);
 }
 
 void maketildelist(s) /**/
@@ -502,8 +501,8 @@ char *s;
 {
 struct stat sbuf;
 
-   if (stat(s,&sbuf) == -1)
-      return 0;
+	if (!*s) return 0;
+   if (stat(s,&sbuf) == -1) return 0;
    return S_ISDIR(sbuf.st_mode);
 }
 
@@ -541,6 +540,8 @@ char *s;int lst;int incmd;
 {
 char *tokorigs;
 char *origs;
+Compctl cc;
+char *pfx = s;
 
 	slashflag = SLASH_MAYBE;
 	addedstar = 0;
@@ -553,32 +554,18 @@ char *origs;
 	untokenize(s);
 	origs = strdup(s);
 	matches = newlist();
-	if (cmdstr && inarray(cmdstr,hostcmds)) {
-		char **x;
-		haspath = exact = 0;
+	if (incmd)
+		cc = &cc_compos;
+	else if (linredir || !(cmdstr && (cc = gethnode(cmdstr,compctltab))))
+		cc = &cc_default;
+	exact = 0;
+	if (cc->mask & CC_COMMPATH) gen_matches_reg(s,1,(cc->mask & CC_FILES));
+	else if (cc->mask & CC_FILES) gen_matches_reg(s,0,1);
+	else {
+		haspath = 0;
 		slashflag = SLASH_NO;
-		pat = s;
-		for (x = hosts; *x; x++) addcmdmatch(*x,NULL);
-	} else if (cmdstr && inarray(cmdstr,optcmds)) {
-		struct option *o;
-		haspath = exact = 0;
-		slashflag = SLASH_NO;
-		pat = s;
-		for (o = optns; o->name; o++) addcmdmatch(o->name,NULL);
-	} else if (cmdstr && inarray(cmdstr,varcmds)) {
-		haspath = exact = 0;
-		slashflag = SLASH_NO;
-		pat = s;
-		listhtable(paramtab,addcmdmatch);
-	} else if (cmdstr && inarray(cmdstr,bindcmds)) {
-		int t0;
-		haspath = exact = 0;
-		slashflag = SLASH_NO;
-		pat = s;
-		for (t0 = 0; t0 != ZLECMDCOUNT; t0++)
-			if (*zlecmds[t0].name) addcmdmatch(zlecmds[t0].name,NULL);
-	} else {
-		gen_matches_reg(s,incmd);
+	}
+	if (cc->mask & (CC_FILES|CC_COMMPATH)) {
 		/* only do "globbed" completion if regular completion fails.
 		   pem, 7Oct91 */
 		if ((!full(matches) || errflag) && useglob) {
@@ -590,6 +577,25 @@ char *origs;
 		    *s = 0;
 		}
 	}
+	pat = s;
+	if ((cc->mask & CC_HOSTS) && !haspath) {
+		char **x;
+		for (x = hosts; *x; x++) addcmdmatch(*x,NULL);
+	}
+	if ((cc->mask & CC_OPTIONS) && !haspath) {
+		struct option *o;
+		for (o = optns; o->name; o++) addcmdmatch(o->name,NULL);
+	}
+	if ((cc->mask & CC_VARS) && !haspath) listhtable(paramtab,addcmdmatch);
+	if ((cc->mask & CC_BINDINGS) && !haspath) {
+		int t0;
+		for (t0 = 0; t0 != ZLECMDCOUNT; t0++)
+			if (*zlecmds[t0].name) addcmdmatch(zlecmds[t0].name,NULL);
+	}
+	if (cc->mask & CC_USRKEYS) {
+		char **usr = get_user_var(cc->keyvar);
+		if (usr) while (*usr) addcmdmatch(*usr++,NULL);
+	}
 	if (lst != COMP_LIST_COMPLETE) do_fignore(origs);
 	if (!full(matches) || errflag) {
 		feep();
@@ -598,16 +604,22 @@ char *origs;
 			unset(LISTTYPES) ? NULL :
 				(haspath) ? pathprefix : "./");
 	} else if (nextnode(firstnode(matches))) {
-		do_ambiguous(s);
+		do_ambiguous(pfx);
 	} else {
-		do_single(s);
+		do_single(pfx);
 	}
-	ll = strlen(line);
+	ll = strlen((char *) line);
 	setterm();
 	popheap();
 	permalloc();
 }
 
+char **get_user_var(nam) /**/
+char *nam;
+{
+	return (nam) ? getaparam(nam) : NULL;
+}
+
 void gen_matches_glob(s,incmd) /**/
 char *s;int incmd;
 {
@@ -641,7 +653,9 @@ struct direct *de;
 		slashflag = SLASH_NO;
 		pat = s;
 		listhtable(aliastab ,addcmdmatch);
+		if (isset(HASHLISTALL)) fullhash();
 		listhtable(cmdnamtab,addcmdnodis);
+		if (isset(AUTOCD)) listhtable(paramtab ,addcmddirparam);
 		if (d = opendir(".")) {
 			char *q;
 
@@ -710,25 +724,33 @@ struct direct *de;
 	}
 }
 
-void gen_matches_reg(s,incmd) /**/
-char *s;int incmd;
+void gen_matches_reg(s,incmd,regfiles) /**/
+char *s;int incmd;int regfiles;
 {
 char *u;
 DIR *d;
 struct direct *de;
 
-	haspath = exact = 0;
+	haspath = 0;
 	for (u = s+strlen(s); u >= s; u--)
 		if (*u == '/' || *u == '@' || *u == '$') break;
-	typechar = *u;
 	if (u >= s) {
+		typechar = *u;
 		*u++ = '\0';
 		haspath = 1;
+	} else if (*s == '=') {
+		typechar = '=';
+		*s = '\0'; u = s+1;
+		haspath = 1;
 	} else u = s;
 	pat = u;
 	if (typechar == '$' && haspath) {
 		/* slashflag = SLASH_NO; */
 		listhtable(paramtab,addcmdmatch);
+	} else if (typechar == '=' && haspath) {
+		slashflag = SLASH_NO;
+		if (isset(HASHLISTALL)) fullhash();
+		listhtable(cmdnamtab,addcmdnodis);
 	} else if (typechar == '@' && haspath) {
 		char **x;
 		slashflag = SLASH_NO;
@@ -740,18 +762,33 @@ struct direct *de;
 	} else if (incmd && !haspath) {
 		slashflag = SLASH_NO;
 		listhtable(aliastab ,addcmdmatch);
+		if (isset(HASHLISTALL)) fullhash();
 		listhtable(cmdnamtab,addcmdnodis);
+		if (isset(AUTOCD) && isset(CDABLEVARS))
+			listhtable(paramtab ,addcmddirparam);
 		if (d = opendir(".")) {
 			char *q;
 			struct stat buf;
 
 			readdir(d); readdir(d);
-			while ((de = readdir(d)) && !errflag)
-				if (strpfx(pat,q = de->d_name) &&
+			if (regfiles) {
+				while ((de = readdir(d)) && !errflag)
+					if (strpfx(pat,q = de->d_name) &&
+						(*q != '.' || *u == '.' || isset(GLOBDOTS))) addmatch(q);
+			} else if (isset(AUTOCD)) {
+				while ((de = readdir(d)) && !errflag)
+					if (strpfx(pat,q = de->d_name) &&
+						(*q != '.' || *u == '.' || isset(GLOBDOTS)) &&
+						stat(q,&buf) >= 0 &&
+						(buf.st_mode & S_IEXEC) == S_IEXEC) addmatch(q);
+			} else {
+				while ((de = readdir(d)) && !errflag)
+					if (strpfx(pat,q = de->d_name) &&
 						(*q != '.' || *u == '.' || isset(GLOBDOTS)) &&
 						stat(q,&buf) >= 0 &&
 						(buf.st_mode & (S_IFMT|S_IEXEC)) == (S_IFREG|S_IEXEC))
-					addmatch(q);
+							addmatch(q);
+			}
 			closedir(d);
 		}
 	} else if (d = OPENDIR(pathprefix =
@@ -785,10 +822,13 @@ char *origstr;
 	if (full(matches) && nextnode(firstnode(matches))) {
 		Lknode z,zn;
 
+		z = firstnode(matches);
+		ambig = 1000;
 		for (z = firstnode(matches); z; z = zn) {
 			char *q = getdata(z);
 			int namlen = strlen(q);
 			int	slen = strlen(origstr);
+			int	slpt;
 			char **pt = fignore;
 	
 			zn = nextnode(z);
@@ -796,13 +836,19 @@ char *origstr;
 				/* We try to be smart here and override the
 				   fignore variable if the user has explicity
 				   used the ignored prefix, pem, 7 May 1991 */
-				if (!addedstar && strcmp(origstr+slen-strlen(*pt), *pt) == 0)
+				slpt = strlen(*pt);
+				if (!addedstar && slen > slpt &&
+						strcmp(origstr+slen-slpt, *pt) == 0)
 					continue;
-				if (strlen(*pt) < namlen && !strcmp(q+namlen-strlen(*pt),*pt)) {
+				if (slpt < namlen && !strcmp(q+namlen-slpt,*pt)) {
 					uremnode(matches,z);
 					break;
 				}
 			}
+			if (!*pt) {
+				int y = pfxlen(peekfirst(matches),q);
+				if (y < ambig) ambig = y;
+			}
 		}
 	}
 }
@@ -836,12 +882,13 @@ char *s;
 			spaceinline(1);
 			line[cs++] = *s++;
 		}
-		inststr(pat);
+		inststr(exactstr);
 		spaceinline(1);
 		switch (slashflag) {
 			case SLASH_YES: line[cs++] = '/'; break;
 			case SLASH_NO : line[cs++] = ' '; break;
-			case SLASH_MAYBE: line[cs++] = isdir(pat,pathprefix)?'/':' '; break;
+			case SLASH_MAYBE: line[cs++] =
+				isdir(exactstr,pathprefix) ? '/' : ' '; break;
 		}
 		return;
 	}
@@ -886,6 +933,7 @@ char *s;
 		case SLASH_NO : line[cs++] = ' '; break;
 		case SLASH_MAYBE: line[cs++] = isdir(s,pathprefix) ? '/' : ' '; break;
 	}
+	if (isset(AUTOREMOVESLASH) && line[cs-1] == '/') addedslash = 1;
 }
 
 void do_ambig_menu(s) /**/
@@ -997,8 +1045,7 @@ char **arr,**ap;
 
 				printf("%s",*ap);
 				sprintf(pbuf,"%s/%s",apps,*ap);
-				if (lstat(pbuf,&buf))
-					putchar(' ');
+				if (lstat(pbuf,&buf)) putchar(' ');
 				else switch (buf.st_mode & S_IFMT) /* screw POSIX */
 					{
 					case S_IFDIR: putchar('/'); break;
@@ -1062,45 +1109,52 @@ char **arr,**ap;
 	longest++;
 	while (t0)
 		t0 /= 10, longest++;
-	fct = (columns-1)/(longest+2);
+	fct = (columns-1)/(longest+3); /* to compensate for added ')' */
 	if (fct == 0)
 		fct = 1;
 	else
 		fw = (columns-1)/fct;
 	colsz = (ct+fct-1)/fct;
-	for (t1 = 0; t1 != colsz; t1++)
-		{
+	for (t1 = 0; t1 != colsz; t1++) {
 		ap = arr+t1;
-		do
-			{
-			int t2 = strlen(*ap)+1,t3;
+		do {
+			int t2 = strlen(*ap)+2,t3;
 
-			fprintf(stderr,"%d %s",t3 = ap-arr+1,*ap);
-			while (t3)
-				t2++,t3 /= 10;
+			fprintf(stderr,"%d) %s",t3 = ap-arr+1,*ap);
+			while (t3) t2++,t3 /= 10;
 			for (; t2 < fw; t2++) fputc(' ',stderr);
 			for (t0 = colsz; t0 && *ap; t0--,ap++);
-			}
-		while (*ap);
+		} while (*ap);
 		fputc('\n',stderr);
+	} 
+
+/* Below is a simple attempt at doing it the Korn Way.. 
+	ap = arr;
+	t0 = 0;
+	do
+		{
+		t0++;
+		fprintf(stderr,"%d) %s\n",t0,*ap);
+		ap++;
 		}
+	while (*ap);*/
 	resetneeded = 1;
 	fflush(stderr);
 }
 
 int doexpandhist() /**/
 {
-char *cc,*ce;
+unsigned char *cc,*ce;
 int t0,oldcs,oldll;
 
 	for (cc = line, ce = line+ll; cc < ce; cc++)
 		if (*cc == '\\' && cc[1])
 			cc++;
-		else if (*cc == bangchar)
+		else if (*cc == bangchar ||
+				(*cc == hatchar && *line == hatchar && cc != line))
 			break;
 	if (*cc == bangchar && cc[1] == '"') return 0;
-	if (cc == ce && *line != hatchar)
-		return 0;
+	if (cc == ce) return 0;
 	oldcs = cs;
 	oldll = ll;
 	zleparse = 1;
@@ -1130,9 +1184,9 @@ int t0,oldcs,oldll;
 
 void magicspace() /**/
 {
-	doexpandhist();
 	c = ' ';
 	selfinsert();
+	doexpandhist();
 }
 
 void expandhistory() /**/
@@ -1183,9 +1237,9 @@ char *s,*t;
 	mult = 1;
 	pushline();
 	sizeline(strlen(s)+strlen(t)+1);
-	strcpy(line,t);
-	strcat(line," ");
-	cs = ll = strlen(line);
+	strcpy((char *) line,t);
+	strcat((char *) line," ");
+	cs = ll = strlen((char *) line);
 	inststr(s);
 	free(s);
 	done = 1;
@@ -1204,7 +1258,7 @@ char *s,*str;
 	cs = cmdwb;
 	foredel(cmdwe-cmdwb);
 	spaceinline(strlen(str));
-	strncpy(line+cs,str,strlen(str));
+	strncpy((char *) line+cs,str,strlen(str));
 	cs = oldcs;
 	if (cs >= cmdwe) cs += cmdwe-cmdwb+strlen(str);
 	if (cs > ll) cs = ll;
