@@ -1,34 +1,22 @@
 /*
-
-	zle_utils.c - miscellaneous line editor utilities
-
-	This file is part of zsh, the Z shell.
-
-	zsh is free software; no one can prevent you from reading the source
-   code, or giving it to someone else.
-
-   This file is copyrighted under the GNU General Public License, which
-   can be found in the file called COPYING.
-
-   Copyright (C) 1990, 1991 Paul Falstad
-
-   zsh is distributed in the hope that it will be useful, but
-   WITHOUT ANY WARRANTY.  No author or distributor accepts
-   responsibility to anyone for the consequences of using it or for
-   whether it serves any particular purpose or works at all, unless he
-   says so in writing.  Refer to the GNU General Public License
-   for full details.
-
-   Everyone is granted permission to copy, modify and redistribute
-   zsh, but only under the conditions described in the GNU General Public
-   License.   A copy of this license is supposed to have been given to you
-   along with zsh so you can know your rights and responsibilities.
-   It should be in a file named COPYING.
-
-   Among other things, the copyright notice and this notice must be
-   preserved on all copies.
-
-*/
+ *
+ * zle_utils.c - miscellaneous line editor utilities
+ *
+ * This file is part of zsh, the Z shell.
+ *
+ * This software is Copyright 1992 by Paul Falstad
+ *
+ * Permission is hereby granted to copy, reproduce, redistribute or otherwise
+ * use this software as long as: there is no monetary profit gained
+ * specifically from the use or reproduction of this software, it is not
+ * sold, rented, traded or otherwise marketed, and this copyright notice is
+ * included prominently in any copy made. 
+ *
+ * The author make no claims as to the fitness or correctness of this software
+ * for any use whatsoever, and it is provided as is. Any use of this software
+ * is at the user's own risk. 
+ *
+ */
 
 #define ZLE
 #include "zsh.h"
@@ -39,7 +27,7 @@ void sizeline(sz) /**/
 int sz;
 {
 	while (sz > linesz)
-		line = realloc(line,(linesz *= 4)+1);
+		line = (unsigned char *)realloc(line,(linesz *= 4)+1);
 }
 
 /* insert space for ct chars at cursor position */
@@ -50,7 +38,7 @@ int ct;
 int i;
 
 	while (ct+ll > linesz)
-		line = realloc(line,(linesz *= 4)+1);
+		line = (unsigned char *)realloc(line,(linesz *= 4)+1);
 	for (i = ll; i >= cs; i--)
 		line[i+ct] = line[i];
 	ll += ct;
@@ -111,7 +99,7 @@ int i;int ct;int dir;
 	}
 	if (dir) {
 		char *s = zalloc(strlen(cutbuf)+ct+1);
-		strncpy(s,line+i,ct);
+		strncpy(s,(char *) line+i,ct);
 		strcpy(s+ct,cutbuf);
 		free(cutbuf);
 		cutbuf = s;
@@ -147,17 +135,17 @@ void setline(s) /**/
 char *s;
 {
 	sizeline(strlen(s));
-	strcpy(line,s);
+	strcpy((char *) line,s);
 	cs = ll = strlen(s);
 	if (cs && bindtab == altbindtab) cs--;
 }
 
 void sethistline(s) /**/
-char *s;
+unsigned char *s;
 {
 	setline(s);
 	for (s = line; *s; s++)
-		if (*s == HISTSPACE)
+		if (*s == (unsigned char)HISTSPACE)
 			*s = ' ';
 }
 
@@ -194,35 +182,36 @@ int t0;
 		undos[t0].change = NULL;
 	undoct = 0;
 	lastline = zalloc(lastlinelen = (ll+1 < 32) ? 32 : ll+1);
-	strcpy(lastline,line);
+	strcpy((char *) lastline,(char *) line);
 	lastcs = cs;
 }
 
 void addundo() /**/
 {
 int pf,sf;
-char *s,*s2,*t,*t2;
+unsigned char *s,*s2,*t,*t2;
 struct undoent *ue;
 
 	for (s = line, t = lastline; *s && *s==*t; s++,t++);
 	if (!*s && !*t)
 		return;
 	pf = s-line;
-	for (s2 = line+strlen(line), t2 = lastline+strlen(lastline);
+	for (s2 = (unsigned char *)line+strlen((char *) line),
+			t2 = lastline+strlen((char *) lastline);
 		s2 > s && t > t2 && s2[-1] == t2[-1]; s2--,t2--);
-	sf = strlen(s2);
+	sf = strlen((char *) s2);
 	ue = undos+(undoct = (UNDOCT-1) & (undoct+1));
 	ue->pref = pf;
 	ue->suff = sf;
 	ue->len = t2-t;
 	ue->cs = lastcs;
-	strncpy(ue->change = halloc(ue->len),t,ue->len);
+	strncpy(ue->change = halloc(ue->len),(char *) t,ue->len);
 	while (ll+1 > lastlinelen)
 		{
 		free(lastline);
 		lastline = zalloc(lastlinelen *= 2);
 		}
-	strcpy(lastline,line);
+	strcpy((char *) lastline,(char *) line);
 	lastcs = cs;
 }
 
