@@ -1,34 +1,22 @@
 /*
-
-	zle_vi.c - vi-specific functions
-
-	This file is part of zsh, the Z shell.
-
-	zsh is free software; no one can prevent you from reading the source
-   code, or giving it to someone else.
-
-   This file is copyrighted under the GNU General Public License, which
-   can be found in the file called COPYING.
-
-   Copyright (C) 1990, 1991 Paul Falstad
-
-   zsh is distributed in the hope that it will be useful, but
-   WITHOUT ANY WARRANTY.  No author or distributor accepts
-   responsibility to anyone for the consequences of using it or for
-   whether it serves any particular purpose or works at all, unless he
-   says so in writing.  Refer to the GNU General Public License
-   for full details.
-
-   Everyone is granted permission to copy, modify and redistribute
-   zsh, but only under the conditions described in the GNU General Public
-   License.   A copy of this license is supposed to have been given to you
-   along with zsh so you can know your rights and responsibilities.
-   It should be in a file named COPYING.
-
-   Among other things, the copyright notice and this notice must be
-   preserved on all copies.
-
-*/
+ *
+ * zle_vi.c - vi-specific functions
+ *
+ * This file is part of zsh, the Z shell.
+ *
+ * This software is Copyright 1992 by Paul Falstad
+ *
+ * Permission is hereby granted to copy, reproduce, redistribute or otherwise
+ * use this software as long as: there is no monetary profit gained
+ * specifically from the use or reproduction of this software, it is not
+ * sold, rented, traded or otherwise marketed, and this copyright notice is
+ * included prominently in any copy made. 
+ *
+ * The author make no claims as to the fitness or correctness of this software
+ * for any use whatsoever, and it is provided as is. Any use of this software
+ * is at the user's own risk. 
+ *
+ */
 
 #define ZLE
 #include "zsh.h"
@@ -70,11 +58,13 @@ int ch;
 	return ch;
 }
 
-int getvirange() /**/
+int getvirange(wf) /**/
+int wf;
 {
-int k2,t0,startline;
+int k2,t0,startline,endline;
 
 	startline = findbol();
+	endline = findeol();
 	for (;;) {
 		k2 = getkeycmd();
 		if (k2 == -1) {
@@ -82,7 +72,7 @@ int k2,t0,startline;
 			return -1;
 		}
 		if (zlecmds[k2].flags & ZLE_ARG)
-			zlecmds[k2].func();
+			(*zlecmds[k2].func)();
 		else
 			break;
 	}
@@ -97,19 +87,25 @@ int k2,t0,startline;
 	t0 = cs;
 
 	virangeflag = 1;
-	zlecmds[k2].func();
-	virangeflag = 0;
+	wordflag = wf;
+	(*zlecmds[k2].func)();
+	wordflag = virangeflag = 0;
 	if (cs == t0) {
 		feep();
 		return -1;
 	}
 	if (startline != findbol()) {
-		if (cs < t0) {
-			cs = startline;
-			t0 = findeol()+1;
+		if (zlecmds[k2].flags & ZLE_LINEMOVE) {
+			if (cs < t0) {
+				cs = startline;
+				t0 = findeol()+1;
+			} else {
+				t0 = startline;
+				cs = findeol()+1;
+			}
 		} else {
-			t0 = startline;
-			cs = findeol()+1;
+			if (cs < startline) cs = startline;
+			else if (cs >= endline) cs = endline-1;
 		}
 	}
 	if (cs > t0) {
@@ -149,7 +145,7 @@ void videlete() /**/
 int c2;
 
 	startvichange(1);
-	if ((c2 = getvirange()) == -1)
+	if ((c2 = getvirange(0)) == -1)
 		{ vichgflag = 0; return; }
 	forekill(c2-cs,0);
 	vichgflag = 0;
@@ -160,7 +156,7 @@ void vichange() /**/
 int c2;
 
 	startvichange(1);
-	if ((c2 = getvirange()) == -1)
+	if ((c2 = getvirange(1)) == -1)
 		{ vichgflag = 0; return; }
 	forekill(c2-cs,0);
 	bindtab = mainbindtab;
@@ -196,8 +192,7 @@ void viyank() /**/
 {
 int c2;
 
-	if ((c2 = getvirange()) == -1)
-		return;
+	if ((c2 = getvirange(0)) == -1) return;
 	cut(cs,c2-cs,0);
 }
 
@@ -226,10 +221,9 @@ int ch;
 		return;
 	}
 	startvichange(1);
-	if (ch = vigetkey())
-		while (mult--)
-			line[cs++] = ch;
+	if (ch = vigetkey()) while (mult--) line[cs++] = ch;
 	vichgflag = 0;
+	cs--;
 }
 
 void vicmdmode() /**/
@@ -260,15 +254,15 @@ void vioperswapcase() /**/
 {
 int c2;
 
-	if ((c2 = getvirange()) == -1)
+	if ((c2 = getvirange(0)) == -1)
 		return;
 	while (cs < c2)
 		{
 		int ch = line[cs];
 
-		if (ch >= 'a' && ch <= 'z')
+		if (islower(ch))
 			ch = tuupper(ch);
-		else if (ch >= 'A' && ch <= 'Z')
+		else if (isupper(ch))
 			ch = tulower(ch);
 		line[cs++] = ch;
 		}
@@ -286,7 +280,7 @@ int c2,endcs,t0,rmult;
 
 	if (mult < 0) { mult = -mult; viunindent(); return; }
 	rmult = mult;
-	if ((c2 = getvirange()) == -1)
+	if ((c2 = getvirange(0)) == -1)
 		return;
 	if (cs != findbol()) { feep(); return; }
 	endcs = cs+rmult;
@@ -304,7 +298,7 @@ int c2,endcs,t0,rmult;
 
 	rmult = mult;
 	if (mult < 0) { mult = -mult; viindent(); return; }
-	if ((c2 = getvirange()) == -1)
+	if ((c2 = getvirange(0)) == -1)
 		return;
 	if (cs != findbol()) { feep(); return; }
 	endcs = cs;
