@@ -1,41 +1,42 @@
 /*
-
-	zsh.h - standard header file
-
-	This file is part of zsh, the Z shell.
-
-	zsh is free software; no one can prevent you from reading the source
-   code, or giving it to someone else.
-
-   This file is copyrighted under the GNU General Public License, which
-   can be found in the file called COPYING.
-
-   Copyright (C) 1990, 1991 Paul Falstad
-
-   zsh is distributed in the hope that it will be useful, but
-   WITHOUT ANY WARRANTY.  No author or distributor accepts
-   responsibility to anyone for the consequences of using it or for
-   whether it serves any particular purpose or works at all, unless he
-   says so in writing.  Refer to the GNU General Public License
-   for full details.
-
-   Everyone is granted permission to copy, modify and redistribute
-   zsh, but only under the conditions described in the GNU General Public
-   License.   A copy of this license is supposed to have been given to you
-   along with zsh so you can know your rights and responsibilities.
-   It should be in a file named COPYING.
-
-   Among other things, the copyright notice and this notice must be
-   preserved on all copies.
-
-*/
+ *
+ * zsh.h - standard header file
+ *
+ * This file is part of zsh, the Z shell.
+ *
+ * This software is Copyright 1992 by Paul Falstad
+ *
+ * Permission is hereby granted to copy, reproduce, redistribute or otherwise
+ * use this software as long as: there is no monetary profit gained
+ * specifically from the use or reproduction of this software, it is not
+ * sold, rented, traded or otherwise marketed, and this copyright notice is
+ * included prominently in any copy made. 
+ *
+ * The author make no claims as to the fitness or correctness of this software
+ * for any use whatsoever, and it is provided as is. Any use of this software
+ * is at the user's own risk. 
+ *
+ */
 
 #include "config.h"
 
 #include <stdio.h>
-#include <string.h>
 #include <ctype.h>
 
+#ifdef HAS_STRING
+#include <string.h>
+#else
+#include <strings.h>
+#endif
+
+#ifdef HAS_LOCALE
+#include <locale.h>
+#endif
+
+#ifdef HAS_STDLIB
+#include <stdlib.h>
+#endif
+
 #ifdef SYSV
 #include <sys/bsdtypes.h>
 #define _POSIX_SOURCE
@@ -46,23 +47,36 @@
 extern int gethostname();
 #define sigmask(m) m
 #include <sys/dirent.h>
-struct rusage { int foo; }
 #else
 #include <sys/types.h>		/* this is the key to the whole thing */
 #endif
 
+#ifdef _IBMR2
+#undef _BSD   /* union wait SUCKS! */
 #include <sys/wait.h>
+#define _BSD
+#else
+#include <sys/wait.h>
+#endif
+
 #include <sys/time.h>
+
 #ifndef SYSV
 #include <sys/resource.h>
 #endif
+
 #include <sys/file.h>
 #include <signal.h>
+
 #ifdef TERMIO
+#define VDISABLEVAL -1
+#define TIO 1
 #include <sys/termio.h>
 #else
 #ifdef TERMIOS
-#include <sys/termios.h>
+#define VDISABLEVAL 0
+#define TIO 1
+#include <termios.h>
 #else
 #include <sgtty.h>
 #endif
@@ -73,6 +87,7 @@ struct rusage { int foo; }
 #endif
 
 #include <sys/param.h>
+#include <sys/times.h>
 
 #ifdef SYSV
 #undef _POSIX_SOURCE
@@ -82,11 +97,16 @@ struct rusage { int foo; }
 #include <sys/bsdtty.h>
 #endif
 
-#define VERSIONSTR "zsh v2.1.0"
+#ifndef sun
+#include <sys/ioctl.h>
+#else
+#include <sys/filio.h>
+#endif
+
+#define VERSIONSTR "zsh v2.2.0"
 
 #if 0 /* __STDC__ */
 #include <unistd.h>
-#include <stdlib.h>
 #include <fcntl.h>
 #include <stat.h>
 #define DCLPROTO(X) X
@@ -100,8 +120,12 @@ struct rusage { int foo; }
 #endif
 #endif /* __STDC__ */
 
+#ifdef QDEBUG
+#define DCLPROTO(X) X
+#endif
+
 #define DEFWORDCHARS "*?_-.[]~=/&;!#$%^(){}<>"
-#define DEFTIMEFMT "%E real  %U user  %S system  %P"
+#define DEFTIMEFMT "%E real  %U user  %S system  %P %J"
 #ifdef UTMP_HOST
 #define DEFWATCHFMT "%n has %a %l from %m."
 #else
@@ -114,6 +138,10 @@ struct rusage { int foo; }
 #define EXTERN extern
 #endif
 
+#ifdef HAS_STRING
+#define killpg(pgrp,sig) kill(-(pgrp),sig)
+#endif
+
 #ifndef F_OK
 #define F_OK 00
 #define R_OK 04
@@ -240,13 +268,13 @@ vptr (*ncalloc)DCLPROTO((int));
 #define ERRAPPNOW 7
 #define READ 8
 #define HEREDOC 9
-#define MERGE 10
-#define CLOSE 11
-#define INPIPE 12
-#define OUTPIPE 13
-#define HERESTR 14
-#define NONE 15
-#define HEREDOCDASH 16
+#define HEREDOCDASH 10
+#define HERESTR 11
+#define MERGE 12
+#define CLOSE 13
+#define INPIPE 14
+#define OUTPIPE 15
+#define NONE 16
 
 #ifdef GLOBALS
 int redirtab[TRINANG-OUTANG+1] = {
@@ -268,11 +296,15 @@ int redirtab[TRINANG-OUTANG+1] = {
 int redirtab[TRINANG-OUTANG+1];
 #endif
 
-#define IS_READFD(X) (((X)>=READ && (X)<=MERGE)||(X)==HEREDOCDASH||(X)==HERESTR)
+#define IS_READFD(X) ((X)>=READ && (X)<=MERGE)
 #define IS_REDIROP(X) ((X)>=OUTANG && (X)<=TRINANG)
 #define IS_ERROR_REDIR(X) ((X)>=MERGEOUT && (X)<=ERRAPPNOW)
 #define UN_ERROR_REDIR(X) ((X)-MERGEOUT+WRITE)
 
+#define FD_WORD   -1
+#define FD_COPROC -2
+#define FD_CLOSE  -3
+
 EXTERN char **environ;
 
 typedef struct hashtab *Hashtab;
@@ -297,6 +329,9 @@ typedef struct comp *Comp;
 typedef struct redir *Redir;
 typedef struct complist *Complist;
 typedef struct heap *Heap;
+typedef struct histent *Histent;
+typedef struct hp *Hp;
+typedef struct compctl *Compctl;
 typedef void (*FFunc)DCLPROTO((vptr));
 typedef vptr (*VFunc)DCLPROTO((vptr));
 typedef void (*HFunc)DCLPROTO((char *,char *));
@@ -326,6 +361,7 @@ struct lklist {
 #define peekfirst(X) ((X)->first->dat)
 #define pushnode(X,Y) insnode(X,(Lknode) X,Y)
 #define incnode(X) (X = nextnode(X))
+#define gethistent(X) (histentarr+((X)%histentct))
 
 /* node structure for syntax trees */
 
@@ -424,8 +460,8 @@ struct cmd {
 
 #define SIMPLE 0
 #define SUBSH 1
-#define ZCTIME 2
-#define CURSH 3
+#define CURSH 2
+#define ZCTIME 3
 #define FUNCDEF 4
 #define CFOR 5
 #define CWHILE 6
@@ -527,25 +563,26 @@ struct multio {
 
 struct cmdnam 
 {
+	struct hashnode *next; int canfree; char *nam; /* hash data */
 	int type,flags;
-	int ifil[2];
 	union {
 		char *nam;		/* full pathname if type != BUILTIN */
 		int binnum;		/* func to exec if type == BUILTIN */
 		List list;		/* list to exec if type == SHFUNC */
 		} u;
+	char **pcomp;		/* location in path for EXCMD's */
 	};
 
-#define EXCMD_PREDOT 0
-#define EXCMD_POSTDOT 1
+#define EXCMD 0
 #define BUILTIN 2
 #define SHFUNC 3
 #define DISABLED 4
-#define ISEXCMD(X) ((X)==EXCMD_PREDOT||(X)==EXCMD_POSTDOT)
+#define ISEXCMD(X) ((X)==EXCMD)
 
 /* node used in parameter hash table (paramtab) */
 
 struct param {
+	struct hashnode *next; int canfree; char *nam; /* hash data */
 	union {
 		char **arr;		/* value if declared array */
 		char *str;		/* value if declared string (scalar) */
@@ -610,10 +647,10 @@ struct fdpair {
 
 struct ttyinfo {
 #ifdef TERMIOS
-	struct termios termios;
+	struct termios tio;
 #else
 #ifdef TERMIO
-	struct termio termio;
+	struct termio tio;
 #else
 	struct sgttyb sgttyb;
 	int lmodes;
@@ -633,7 +670,7 @@ EXTERN struct ttyinfo savedttyinfo;
 struct job {
 	long gleader;					/* process group leader of this job */
 	int stat;
-	char *cwd;						/* current working dir of shell when
+	char *pwd;						/* current working dir of shell when
 											this job was spawned */
 	struct process *procs;		/* list of processes */
 	Lklist filelist;				/* list of files to delete when done */
@@ -649,28 +686,28 @@ struct job {
 
 #define SP_RUNNING -1		/* fake statusp for running jobs */
 
+#ifndef RUSAGE_CHILDREN
+#undef HAS_RUSAGE
+#endif
+
+struct timeinfo {
+#ifdef HAS_RUSAGE
+	struct rusage ru;
+#else
+	long ut, st;
+#endif
+};
+
 /* node in job process lists */
 
+#define JOBTEXTSIZE 80
+
 struct process {
 	struct process *next;
 	long pid;
-	char *text;						/* text to print when 'jobs' is run */
+	char text[JOBTEXTSIZE];		/* text to print when 'jobs' is run */
 	int statusp;					/* return code from wait3() */
-	int lastfg;						/* set if this process represents a
-											fragment of a pipeline run in a subshell
-											for commands like:
-
-											foo | bar | ble
-
-											where foo is a current shell function
-											or control structure.  The command
-											actually executed is:
-
-											foo | (bar | ble)
-
-											That's two procnodes in the parent
-											shell, the latter having this flag set. */
-	struct rusage ru;
+	struct timeinfo ti;
 	time_t bgtime;					/* time job was spawned */
 	time_t endtime;				/* time job exited */
 	};
@@ -678,6 +715,7 @@ struct process {
 /* node in alias hash table */
 
 struct alias {
+	struct hashnode *next; int canfree; char *nam; /* hash data */
 	char *text;			/* expansion of alias */
 	int cmd;				/* one for regular aliases,
 								zero for global aliases,
@@ -699,9 +737,8 @@ struct schedcmd {
 
 struct hashnode {
 	struct hashnode *next;
-	char *nam;
-	vptr dat;
 	int canfree;		/* nam is free()able */
+	char *nam;
 	};
 
 /* hash table */
@@ -712,6 +749,31 @@ struct hashtab {
 	struct hashnode **nodes;		/* array of size hsize */
 	};
 
+/* history entry */
+
+struct histent {
+	char *lex;			/* lexical history line */
+	char *lit;			/* literal history line */
+	time_t stim;		/* command started time (datestamp) */
+	time_t ftim;		/* command finished time */
+	};
+
+/* completion control */
+
+struct compctl {
+	struct hashnode *next; int canfree; char *nam; /* hash data */
+	int mask;
+	char *keyvar;
+	};
+
+#define CC_FILES		1
+#define CC_COMMPATH	2
+#define CC_HOSTS		4
+#define CC_OPTIONS	8
+#define CC_VARS		16
+#define CC_BINDINGS	32
+#define CC_USRKEYS   64
+
 extern char *sys_errlist[];
 extern int errno;
 
@@ -758,7 +820,7 @@ struct option {
 #define RECEXACT 'S'
 #define CDABLEVARS 'T'
 #define MAILWARNING 'U'
-#define NOPROMPTCLOBBER 'V'
+#define NOPROMPTCR 'V'
 #define AUTORESUME 'W'
 #define LISTTYPES 'X'
 #define MENUCOMPLETE 'Y'
@@ -797,6 +859,15 @@ struct option {
 #define NOHUP '\21'
 #define NOEQUALS '\22'
 #define CSHNULLGLOB '\23'
+#define HASHCMDS '\24'
+#define HASHDIRS '\25'
+#define NUMERICGLOBSORT '\26'
+#define BRACECCL '\27'
+#define HASHLISTALL '\30'
+#define OVERSTRIKE '\31'
+#define NOHISTBEEP '\32'
+#define PUSHDIGNOREDUPS '\33'
+#define AUTOREMOVESLASH '\34'
 
 #ifndef GLOBALS
 extern struct option optns[];
@@ -820,6 +891,7 @@ struct option optns[] = {
 	"nullglob",NULLGLOB,
 	"rmstarsilent",RMSTARSILENT,
 	"ignorebraces",IGNOREBRACES,
+	"braceccl",BRACECCL,
 	"autocd",AUTOCD,
 	"nobanghist",NOBANGHIST,
 	"sunkeyboardhack",SUNKEYBOARDHACK,
@@ -832,7 +904,7 @@ struct option optns[] = {
 	"recexact",RECEXACT,
 	"cdablevars",CDABLEVARS,
 	"mailwarning",MAILWARNING,
-	"nopromptclobber",NOPROMPTCLOBBER,
+	"nopromptcr",NOPROMPTCR,
 	"autoresume",AUTORESUME,
 	"listtypes",LISTTYPES,
 	"menucomplete",MENUCOMPLETE,
@@ -871,6 +943,14 @@ struct option optns[] = {
 	"noequals",NOEQUALS,
 	"kshoptionprint",KSHOPTIONPRINT,
 	"cshnullglob",CSHNULLGLOB,
+	"hashcmds",HASHCMDS,
+	"hashdirs",HASHDIRS,
+	"numericglobsort",NUMERICGLOBSORT,
+	"hashlistall",HASHLISTALL,
+	"overstrike",OVERSTRIKE,
+	"nohistbeep",NOHISTBEEP,
+	"pushdignoredups",PUSHDIGNOREDUPS,
+	"autoremoveslash",AUTOREMOVESLASH,
 	NULL,0
 };
 #endif
@@ -886,6 +966,7 @@ struct option optns[] = {
 #define nointr() signal(SIGINT,SIG_IGN)
 #define islogin (isset(LOGINSHELL))
 
+#ifndef _IBMR2
 #undef WIFSTOPPED
 #undef WIFSIGNALED
 #undef WIFEXITED
@@ -894,13 +975,13 @@ struct option optns[] = {
 #undef WSTOPSIG
 #undef WCOREDUMPED
 
-#define SP(X) (X)
 #define WIFSTOPPED(X) (((X)&0377)==0177)
 #define WIFSIGNALED(X) (((X)&0377)!=0&&((X)&0377)!=0177)
 #define WIFEXITED(X) (((X)&0377)==0)
 #define WEXITSTATUS(X) (((X)>>8)&0377)
 #define WTERMSIG(X) ((X)&0177)
 #define WSTOPSIG(X) (((X)>>8)&0377)
+#endif
 #define WCOREDUMPED(X) ((X)&0200)
 
 #ifndef S_ISBLK
@@ -931,14 +1012,19 @@ struct option optns[] = {
 #undef S_IFIFO
 #endif
 
+#ifndef S_IFIFO
+#define NO_FIFOS
+#endif
+
 /* buffered shell input for non-interactive shells */
 
 EXTERN FILE *bshin;
 
 /* NULL-terminated arrays containing path, cdpath, etc. */
 
-EXTERN char **path,**cdpath,**fpath,**watch,**mailpath,**tildedirs,**fignore;
-EXTERN char **hosts,**hostcmds,**optcmds,**bindcmds,**varcmds;
+EXTERN char **path,**cdpath,**fpath,**watch,**mailpath;
+EXTERN char **manpath,**tildedirs,**fignore;
+EXTERN char **hosts;
 
 /* named directories */
 
@@ -980,9 +1066,13 @@ EXTERN int isfirstln;
 
 EXTERN int isfirstch;
 
-/* first event number in the history lists */
+/* number of history entries */
 
-EXTERN int firsthist,firstlithist;
+EXTERN int histentct;
+
+/* array of history entries */
+
+EXTERN Histent histentarr;
 
 /* capacity of history lists */
 
@@ -1005,6 +1095,10 @@ EXTERN int incmdpos;
 
 EXTERN int incond;
 
+/* != 0 if we are after a redirection (for ctxtlex only) */
+
+EXTERN int inredir;
+
 /* != 0 if we are about to read a case pattern */
 
 EXTERN int incasepat;
@@ -1059,9 +1153,9 @@ EXTERN int breaks;
 
 EXTERN int retflag;
 
-/* != 0 means don't hash the PATH */
+/* how far we've hashed the PATH so far */
 
-EXTERN int pathsuppress;
+EXTERN char **pathchecked;
 
 /* # of nested loops we are in */
 
@@ -1099,10 +1193,24 @@ EXTERN Hashtab cmdnamtab;
 
 EXTERN Hashtab xbindtab;
 
+/* hash table for completion info for commands */
+
+EXTERN Hashtab compctltab;
+
+/* default completion infos */
+
+EXTERN struct compctl cc_compos, cc_default;
+
 /* the job table */
 
 EXTERN struct job jobtab[MAXJOB];
 
+/* shell timings */
+
+#ifndef HAS_RUSAGE
+EXTERN struct tms shtms;
+#endif
+
 /* the list of sched jobs pending */
 
 EXTERN struct schedcmd *schedcmds;
@@ -1116,10 +1224,12 @@ EXTERN char *hsubl;
 EXTERN char *hsubr;
 
 EXTERN char *username;	/* $USERNAME */
+EXTERN char *logname;	/* $LOGNAME */
 EXTERN long lastval;		/* $? */
 EXTERN long baud;			/* $BAUD */
 EXTERN long columns;		/* $COLUMNS */
 EXTERN long lines;		/* $LINES */
+EXTERN long reporttime; /* $REPORTTIME */
 
 /* input fd from the coprocess */
 
@@ -1156,9 +1266,9 @@ EXTERN long ppid;			/* $PPID */
 
 EXTERN long mypgrp;
 
-EXTERN char *cwd;			/* $PWD */
+EXTERN char *pwd;			/* $PWD */
 EXTERN char *optarg;		/* $OPTARG */
-EXTERN long optind;		/* $OPTIND */
+EXTERN long zoptind;		/* $OPTIND */
 EXTERN char *prompt;		/* $PROMPT */
 EXTERN char *rprompt;	/* $RPROMPT */
 EXTERN char *prompt2;	/* etc. */
@@ -1170,6 +1280,8 @@ EXTERN char *watchfmt;
 EXTERN char *wordchars;
 EXTERN char *fceditparam;
 EXTERN char *tmpprefix;
+EXTERN char *rstring, *Rstring;
+EXTERN char *postedit;
 
 EXTERN char *argzero;	/* $0 */
 
@@ -1185,6 +1297,7 @@ EXTERN char **pparams;	/* $argv */
 /* the default command for null commands */
 
 EXTERN char *nullcmd;
+EXTERN char *readnullcmd;
 
 /* the List of local variables we have to destroy */
 
@@ -1306,6 +1419,10 @@ EXTERN struct ttyinfo shttyinfo;
 
 EXTERN char *ttystrname;
 
+/* 1 if ttyctl -f has been executed */
+
+EXTERN int ttyfrozen;
+
 /* list of memory heaps */
 
 EXTERN Lklist heaplist;
@@ -1399,11 +1516,12 @@ extern char *tccapnams[TC_COUNT];
 
 #define HISTFLAG_DONE   1
 #define HISTFLAG_NOEXEC 2
+#define HISTFLAG_RECALL 4
 
 #include "ztype.h"
 #include "funcs.h"
 
-#ifdef __hpux
+#ifdef HAS_SETPGID
 #define setpgrp setpgid
 #endif
 
@@ -1415,7 +1533,7 @@ extern char *tccapnams[TC_COUNT];
 #define SV_INTERRUPT SV_BSDSIG
 #endif
 
-#ifdef SYSV
+#ifdef HAS_SIGRELSE
 #define blockchld() sighold(SIGCHLD)
 #define unblockchld() sigrelse(SIGCHLD)
 #define chldsuspend() sigpause(SIGCHLD)
